// vfdemodlg.h : header file
//
#include "VFDemoOptions.h"
#include "resource.h"
#include "afxwin.h"
#include "stdafx.h"
#include <VFinger.h>
#include <ScanMan.h>
#include "VFImage.h"
#include "VFFeatures.h"
#include "math.h"
#include "VFDDatabase.h"
#include "afxcmn.h"

#pragma once

inline int Round(double value) { return (int)(value < 0 ? value - 0.5 : value + 0.5); }

class CMyStatic : public CStatic {
public:
	CMyStatic();
	~CMyStatic();
	bool notDrawn;
	DECLARE_MESSAGE_MAP()
	afx_msg void OnPaint();
	afx_msg BOOL OnEraseBkgnd(CDC* pDC);
};

class CVFDemoDlg : public CDialog
{
public:
	CVFDemoDlg(CWnd* pParent = NULL) ;
	~CVFDemoDlg();

 	enum { IDD = IDD_VFDEMO_DIALOG };

	VFDemoOptions* edOptions;
	VFingerOptions* vfOptions;
	
	VFDDatabase* database;

	HBITMAP bitmap;
	HBITMAP bitmapEx;
	CVFFeatures * featuresEx;
	int zoom;

	CStatic s_Image;
	CMyStatic s_ImageExt;
	CStringArray* scannerIds;
	CEdit e_MainLog;
	CEdit e_ScannerLog;
	CString log;
	CButton b_StartStop;
	CButton b_Extract;
	bool extracted;
	CString fingerID;
	afx_msg void OnToolsOptions();
	afx_msg void OnToolsVerifingeroptions();
	afx_msg void OnFileExit();
	afx_msg void OnViewZoomin();
	afx_msg void OnViewZoomOut();
	afx_msg void OnViewZoom1();
	afx_msg void OnToolsClearmainlog();
	afx_msg void OnToolsClearscannerlog();
	afx_msg void OnBnClickedButtonStart();

	HVFCONTEXT context;
	CVFImage* image;
	CVFImage* imageEx;

	int FeatToLocal(int value)
	{
		return value * (VF_IMAGE_RESOLUTION/VF_FEATURES_RESOLUTION) * zoom;
	}

	void PaintBO(CPaintDC* dc, int x, int y, BYTE dir);
	void PaintDir(CPaintDC* dc, int x, int y, BYTE dir, int length);

protected:

	virtual void DoDataExchange(CDataExchange* pDX);
	virtual BOOL OnInitDialog();
	afx_msg void OnSysCommand(UINT nID, LPARAM lParam);
	afx_msg void OnPaint();
	afx_msg HCURSOR OnQueryDragIcon();
	DECLARE_MESSAGE_MAP()

	HICON m_hIcon;

	int bitmapWidth;
	int bitmapHeight;
	int bitmapWidthEx;
	int bitmapHeightEx;

	COLORREF minColor;
	COLORREF maxColor;
	COLORREF minColorEx;
	COLORREF maxColorEx;

	int resolution;
	int resolutionEx;
	bool isBitmap;
	CString scannerId;
	bool started;
	bool canExtract;
	bool registered;
	DWORD tc;
	DWORD time;
	int result;
	BYTE features[VF_MAX_FEATURES_SIZE];
	DWORD featuresSize;
	CMenu* pMenu;

	CVFFeatures * getFeatures() { return featuresEx; }
	void setFeatures(CVFFeatures * value)
	{
		featuresEx = value;
	}

private:
	static int CALLBACK ResultCompare (LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
	{
	   // lParamSort contains a pointer to the list view control.
		
		CListCtrl* pListCtrl = (CListCtrl*) lParamSort;
		return (int)lParam2 - (int)lParam1;
	}

	static void SM_CALLBACK SMMonitorProc(const CHAR * scannerId, DWORD event, VOID * param)
	{
		((CVFDemoDlg*)param)->OnSMMonitor(CString(scannerId), event);
	}
	static void SM_CALLBACK SMImageProc(const CHAR* scannerId, INT width, INT height, const BYTE * image, INT resolution, VOID * param)
	{
		((CVFDemoDlg*)param)->OnScannerImage(CString(scannerId), width, height, image, resolution);
	}
	static void SM_CALLBACK SMStateProc(const CHAR* scannerId, DWORD state, VOID * param)
	{
		((CVFDemoDlg*)param)->OnScannerState(CString(scannerId), state);
	}

	void OnScannerImage(const CString scanner_id, int width, int height, const BYTE * image, int resolution);
	void OnScannerState(const CString scannerId, DWORD state);
	void OnSMMonitor(const CString & scannerId, DWORD event);

	CString SMEventToString(DWORD event, const CString &scannerId);
	CString SMStateToString(DWORD state, const CString &scannerId);

	void SMStopAllCapturing();
	void SMStopAllCapturingAndExit();
	void FreeBitmap();
	void UpdateImage();
	void OnImagePaint();
	void FreeBitmapEx();
	void UpdateImageEx();
	void OnImagePaintEx();
	void StartAllCapturing();
	BOOL Extract();
	void OnExtract();
	int SearchForDuplicates();
	int ListResults();
	void SaveImage(CVFImage* image);

public:
	afx_msg void OnBnClickedButtonExtract();
	CListCtrl l_IdentResults;
	afx_msg void OnFileSave32772();
	afx_msg void OnSaveprocessedimage();
	afx_msg void OnToolsCleardatabase();
};

int GetComboIndexByDataInt(CComboBox &combo, int data);
int GetComboIndexByDataDouble(CComboBox &combo, double data);
BOOL SaveToolsOptions(CVFDemoDlg* enrollmentDemo);
BOOL SaveVFingerOptions(CVFDemoDlg* enrollmentDemo);

CString SMErrorToString(INT error);
CString SMResultToString(INT result);
void WriteLine(CEdit* editControl,CString text);

double VFMatchingThresholdToFAR(INT th);
INT VFFARToMatchingThreshold(double f);
