#pragma once

class CApplication
{
private:
	HINSTANCE handle;
	bool uninitCom;
	string exeName;

	DWORD osMajorVersion;
	DWORD osMinorVersion;
	DWORD osPlatformId;
	DWORD osBuild;
	string osUpdateName;
	WORD osSPMajorVersion;
	WORD osSPMinorVersion;
	WORD osSuite;
	WORD osType;

	string name;
	DWORD versionHigh;
	DWORD versionLow;
	string copyright;
	string company;
	string registryKey;

	string osName;

	LCID locale;

	vector<LCID> locales;

	bool running;

protected:
	void OnStartRunning();
	void OnEndRunning();
	void CheckRunning();

	void SetName(const string &value);
	void SetVersion(DWORD versionHigh, DWORD versionLow = MAKEDWORD(0, 0));
	void SetCopyright(const string &value);
	void SetCompany(const string &value);
	void SetRegistryKey(const string &value);
	void AddLocale(LCID value);

public:
	CApplication(HINSTANCE handle, bool initCom = false);
	virtual ~CApplication();

	HINSTANCE Handle() { return handle; }
	const string & ExeName() { return exeName; }
	const string & Name() { return name; }
	DWORD VersionHigh() { return versionHigh; }
	DWORD VersionLow() { return versionLow; }
	const string & Copyright() { return copyright; }
	const string & Company() { return company; }
	const string & RegistryKey() { return registryKey; }

	DWORD OSMajorVersion() { return osMajorVersion; }
	DWORD OSMinorVersion() { return osMinorVersion; }
	DWORD OSPlatformId() { return osPlatformId; }
	DWORD OSBuild() { return osBuild; }
	const string & OSName() { return osName; }
	const string & OSUpdateName() { return osUpdateName; }
	WORD OSSPMajorVersion() { return osSPMajorVersion; }
	WORD OSSPMinorVersion() { return osSPMinorVersion; }
	WORD OSSuite() { return osSuite; }
	WORD OSType() { return osType; }

	LCID Locale() { return locale; }
	void SetLocale(LCID value);
	int LocaleCount();
	LCID Locale(int index);

	string LoadString(DWORD id, ...);
	string LoadStringV(DWORD id, va_list list);
	HICON LoadIcon(LPCTSTR iconName);
	HBITMAP LoadBitmap(LPCTSTR bitmapName);

	bool Running() { return running; }
};

CApplication * Application();
bool IsApplication();
void CheckApplication();
