#pragma once

#include "Menus.h"

const UINT CM_START = WM_USER + 10000;
const UINT CM_FIRSTSHOW = CM_START;
const UINT CM_USER = CM_START + 1000;

// Forward declarations
class CForm;

//////////////
//          //
// CControl //
//          //
//////////////

class CControl
{
private:
	bool is_subclass;
	WNDPROC old_wnd_proc;
	string wnd_class;
	HWND handle;
	bool tab_stop, group;
	int x, y, width, height;
	bool enabled, visible;
	bool client_edge;
	bool window_edge;
	bool static_edge;
	HFONT font;

	HWND ParentHandle() { return parent ? parent->Handle() : NULL; }

	static LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

	int control_count;
	CControl **controls;

	void DoEnabledChanged();
	void DoVisibleChanged();

	CControl *parent;

protected:
	string text;

	virtual string GetWndClassName();
	virtual HBRUSH GetBackgroundBrush() { return (HBRUSH)(COLOR_3DFACE + 1); }
	virtual HCURSOR GetCursor() { return LoadCursor(NULL, IDC_ARROW); }

	virtual DWORD GetWndStyle();
	virtual DWORD GetWndExStyle();

	virtual void GetWndClass(WNDCLASSEX &wcex);
	void RegisterClass();

	void CheckHandle() { if(!HandleAllocated()) throw exception("Control handle is not allocated"); }

	LRESULT DefWndProc(UINT message, WPARAM wParam, LPARAM lParam);
	void RestoreMessage(MSG &msg, UINT message, WPARAM wParam, LPARAM lParam);
	virtual LRESULT OnMessage(UINT message, WPARAM wParam, LPARAM lParam);

	LRESULT SendCommand(WORD event);
	LRESULT SendNotify(NMHDR * nmh);
	LRESULT SendNotify(UINT code);

	virtual void DoTextChanged() { OnTextChanged(); }

	virtual void OnCreate() {}
	virtual void OnDestroy() {}
	virtual void OnActiveChanged();
	virtual bool OnDefault();
	virtual bool OnCancel();
	virtual void OnActiveControlChanged();
	virtual void OnControlParentChanged(CControl *control);
	virtual void OnCommandSent(WORD event);
	virtual void OnNotifySent(LPNMHDR nmh);
	virtual void OnHScrollSent(WORD code, WORD pos);
	virtual void OnVScrollSent(WORD code, WORD pos);
	virtual void OnEnabledChanged() {}
	virtual void OnVisibleChanged() {}
	virtual void OnTextChanged() {}
	virtual void OnFontChanged() {}
	virtual LRESULT OnSize(DWORD resizing_type, WORD client_width, WORD client_height)
		{ return DefWndProc(WM_SIZE, (WPARAM)resizing_type, MAKELPARAM(client_width, client_height)); }
	virtual LRESULT OnPaint(WPARAM wParam, LPARAM lParam) { return DefWndProc(WM_PAINT, wParam, lParam); }
	virtual LRESULT OnEraseBackground(HDC dc, LPARAM lParam) { return DefWndProc(WM_ERASEBKGND, (WPARAM)dc, lParam); }
	virtual LRESULT OnCancelMode(WPARAM wParam, LPARAM lParam) { return DefWndProc(WM_CANCELMODE, wParam, lParam); }
	virtual LRESULT OnLButtonDblClk(DWORD keys, int x, int y) { return DefWndProc(WM_LBUTTONDBLCLK, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnLButtonDown(DWORD keys, int x, int y) { return DefWndProc(WM_LBUTTONDOWN, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnLButtonUp(DWORD keys, int x, int y) { return DefWndProc(WM_LBUTTONUP, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnMButtonDblClk(DWORD keys, int x, int y) { return DefWndProc(WM_MBUTTONDBLCLK, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnMButtonDown(DWORD keys, int x, int y) { return DefWndProc(WM_MBUTTONDOWN, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnMButtonUp(DWORD keys, int x, int y) { return DefWndProc(WM_MBUTTONUP, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnRButtonDblClk(DWORD keys, int x, int y) { return DefWndProc(WM_RBUTTONDBLCLK, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnRButtonDown(DWORD keys, int x, int y) { return DefWndProc(WM_RBUTTONDOWN, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnRButtonUp(DWORD keys, int x, int y) { return DefWndProc(WM_RBUTTONUP, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnMouseMove(DWORD keys, int x, int y) { return DefWndProc(WM_MOUSEMOVE, (WPARAM)keys, MAKELPARAM(x, y)); }
	virtual LRESULT OnGetDlgCode(WPARAM wParam, MSG *message) { return DefWndProc(WM_GETDLGCODE, wParam, (LPARAM)message); }
	virtual LRESULT OnChar(DWORD ch, DWORD key_data) { return DefWndProc(WM_CHAR, (WPARAM)ch, (LPARAM)key_data); }
	virtual LRESULT OnKeyDown(DWORD key, DWORD key_data) { return DefWndProc(WM_KEYDOWN, (WPARAM)key, (LPARAM)key_data); }
	virtual LRESULT OnKeyUp(DWORD key, DWORD key_data) { return DefWndProc(WM_KEYUP, (WPARAM)key, (LPARAM)key_data); }
	virtual LRESULT OnKillFocus(HWND hwnd, LPARAM lParam) { return DefWndProc(WM_KILLFOCUS, (WPARAM)hwnd, lParam); }
	virtual LRESULT OnSetFocus(HWND hwnd, LPARAM lParam) { return DefWndProc(WM_SETFOCUS, (WPARAM)hwnd, lParam); }
	virtual LRESULT OnSysChar(DWORD ch, DWORD key_data) { return DefWndProc(WM_SYSCHAR, (WPARAM)ch, (LPARAM)key_data); }
	virtual LRESULT OnSysKeyDown(DWORD key, DWORD key_data) { return DefWndProc(WM_SYSKEYDOWN, (WPARAM)key, (LPARAM)key_data); }
	virtual LRESULT OnSysKeyUp(DWORD key, DWORD key_data) { return DefWndProc(WM_SYSKEYUP, (WPARAM)key, (LPARAM)key_data); }

public:
	CControl(bool subclass = false);
	virtual ~CControl();

	virtual void CreateWnd();
	virtual void DestroyWnd();
	void RecreateWnd();

	LRESULT SendMessage(UINT message, WPARAM wParam, LPARAM lParam);
	bool PostMessage(UINT message, WPARAM wParam, LPARAM lParam);

	bool IsSubclass() { return is_subclass; }
	HWND Handle() { return handle; }
	bool HandleAllocated() { return handle != NULL; }
	CControl* Parent() { return parent; }
	bool TabStop() { return tab_stop; }
	bool Group() { return group; }
	CForm* ParentForm();

	int X() { return x; }
	int Y() { return y; }
	int Width() { return width; }
	int Height() { return height; }
	void GetClientRect(RECT &rect);
	int ClientWidth();
	int ClientHeight();

	bool Enabled() { return enabled; }
	bool Visible() { return visible; }
	const string &Text() { return text; }
	bool Capture();
	bool ClientEdge() { return client_edge; }
	bool WindowEdge() { return window_edge; }
	bool StaticEdge() { return static_edge; }
	HFONT Font() { return font; }
	bool Focused();
	bool CanFocus();

	void SetBounds(int x, int y, int width, int height);
	void SetX(int value) { SetBounds(value, y, width, height); }
	void SetY(int value) { SetBounds(x, value, width, height); }
	void SetWidth(int value) { SetBounds(x, y, value, height); }
	void SetHeight(int value) { SetBounds(x, y, width, value); }
	void SetLocation(int x, int y) { SetBounds(x, y, width, height); }
	void SetSize(int width, int height) { SetBounds(x, y, width, height); }
	void SetClientSize(int width, int height);
	void SetClientWidth(int value) { SetClientSize(value, ClientHeight()); }
	void SetClientHeight(int value) { SetClientSize(ClientWidth(), value); }
	virtual void AdjustRect(RECT &rect);
	void SetEnabled(bool value);
	void SetVisible(bool value);
	virtual void SetParent(CControl *value);
	virtual void SetText(const string &value);
	void SetCapture(bool value);
	void SetClientEdge(bool value);
	void SetWindowEdge(bool value);
	void SetStaticEdge(bool value);
	void SetFont(HFONT value);
	void SetFocus();
	virtual void SetTabStop(bool value);
	virtual void SetGroup(bool value);

	virtual void Show(int cmd_show = SW_SHOW);
	void Hide();
	void Update();
	void Invalidate(RECT *rect = NULL, bool erase = true);
	void SetRedrawState(bool value);

	int ControlCount() { return control_count; }
	CControl* GetControl(int index) { return controls[index]; }
	CControl* GetControl(HWND handle);
	void AddControl(CControl *control);
	void RemoveControl(CControl *control);
	int IndexOfControl(CControl *control);
	bool ContainsControl(CControl *control);
	CControl* FindControl(HWND handle);

	string WndClass() { return wnd_class.IsEmpty() ? GetWndClassName() : wnd_class; }
	LONG ClassLong(int index);
	DWORD ClassStyle();
	HBRUSH BackgroundBrush();
	HCURSOR Cursor();
	LONG WndLong(int index);
	DWORD WndStyle();
	DWORD WndExStyle();
	DWORD GetDlgCode(MSG *message = NULL);
	void SetWndLong(int index, LONG value);
	void SetWndStyle(DWORD value);
	void SetWndExStyle(DWORD value);

	int Left() { return x; }
	int Top() { return y; }
	int Right() { return x + width; }
	int Bottom() { return y + height; }
	int Center() { return  x + width / 2; }
	int VCenter() { return y + height / 2; }
	void SetLeft(int value) { SetX(value); }
	void SetTop(int value) { SetY(value); }
	void SetRight(int value) { SetX(value - width); }
	void SetBottom(int value) { SetY(value - height); }
	void SetCenter(int value) { SetX(value - width / 2); }
	void SetVCenter(int value) { SetY(value - height / 2); }

	void ClientToScreen(POINT &pt);
	void ScreenToClient(POINT &pt);

	friend class CForm;
};

///////////////
//           //
// CMainMenu //
//           //
///////////////

class CMainMenu: public CMenu
{
private:
	CForm * form;

protected:
	void OnChanged();

public:
	CMainMenu();
	CMainMenu(HMENU hMenu);
	CMainMenu(LPCTSTR resourceName);

	CForm * get_Form() { return form; }
	__declspec(property(get = get_Form)) CForm * Form;

	bool TranslateMessage(MSG &msg);

	friend class CForm;
};

///////////
//       //
// CForm //
//       //
///////////

enum FormBorder { fbNone = 0, fbThin = WS_BORDER, fbThick = WS_THICKFRAME, fbDialog = WS_DLGFRAME };
enum FormPosition { fpDefault, fpDefaultPosition, fpDefaultSize, fpCurrent, fpOwnerCenter };
enum FormState { fsNormal = 0, fsMaximized = WS_MAXIMIZE, fsMinimized = WS_MINIMIZE };

class CForm: public CControl
{
private:
	bool showed_once;
	bool in_modal_loop;
	bool exiting_modal_loop;
	int modal_result;
	bool active;

	bool has_caption;
	bool maximize_box;
	bool minimize_box;
	bool system_menu;
	bool is_popup;
	FormBorder form_border;
	bool double_border;
	bool tool_window;
	FormPosition form_position;
	FormState form_state;

	CControl *active_control;

	int form_count;
	CForm **forms;

	CMainMenu * mainMenu;

	bool IsMenuMessage(MSG &msg);
	bool IsDialogMessage(MSG &msg);

	void OnControlKillFocus(HWND hwnd, CControl *control);
	void OnControlSetFocus(HWND hwnd, CControl *control);
	void OnControlEnabledChanged(CControl *control);
	void OnControlVisibleChanged(CControl *control);
	void OnControlParentChanged(CControl *control);

	void DisableForms(vector<CForm*> &formList, CForm * owner, CForm * &foreground);
	void EnableForms(vector<CForm*> &formList, CForm * foreground);

	void UpdateMainMenu();

protected:
	virtual LPCTSTR GetMenuName() { return NULL; }
	virtual HICON GetIcon() { return NULL; }
	virtual HICON GetIconSmall() { return NULL; }
	void GetWndClass(WNDCLASSEX &wcex);

	DWORD GetWndStyle();
	DWORD GetWndExStyle();

	LRESULT OnMessage(UINT message, WPARAM wParam, LPARAM lParam);

	void OnCreate();
	void OnDestroy();

	virtual LPCTSTR GetAccelName() { return NULL; }

	virtual void OnClose() {}
	virtual bool OnCloseQuery() { return true; }
	virtual void OnCommand(WORD id, WORD event, CControl * control);
	virtual LRESULT OnNotify(LPNMHDR nmh, CControl * control);
	virtual void OnHScroll(WORD code, WORD pos, CControl * control);
	virtual void OnVScroll(WORD code, WORD pos, CControl * control);
	virtual void OnFirstShow() {}
	virtual void OnHelp() {}

public:
	CForm();
	~CForm();

	void CreateWnd();
	void DestroyWnd();

	void SetParent(CControl *value);

	CForm * Owner() { return (CForm*)parent; }
	void SetOwner(CForm * value) { SetParent(value); }

	CControl* ActiveControl() { return active_control; }
	void SetActiveControl(CControl *value);
	bool Active() { return active; }

	int FormCount() { return form_count; }
	CForm * GetForm(int index)  { return forms[index]; }
	void AddForm(CForm * form);
	void RemoveForm(CForm *form);
	CForm * FindForm(HWND handle);

	void AdjustRect(RECT &rect);

	virtual void Show(int cmd_show = SW_SHOW);
	int ShowModal(int cmd_show = SW_SHOW);
	void ExitModal(int modal_result);
	void Close();
	int ModalResult() { return modal_result; }

	bool HasCaption() { return has_caption; }
	bool MinimizeBox() { return minimize_box; }
	bool MaximizeBox() { return maximize_box; }
	bool SystemMenu() { return system_menu; }
	bool IsPopup() { return is_popup; }
	FormBorder FormBorder() { return form_border; }
	bool DoubleBorder() { return double_border; }
	bool ToolWindow() { return tool_window; }
	FormPosition FormPosition() { return form_position; }
	FormState FormState() { return form_state; }

	void SetTabStop(bool /*value*/) {}
	void SetGroup(bool /*value*/) {}
	void SetHasCaption(bool value);
	void SetMinimizeBox(bool value);
	void SetMaximizeBox(bool value);
	void SetSystemMenu(bool value);
	void SetIsPopup(bool value);
	void SetFormBorder(::FormBorder value);
	void SetDoubleBorder(bool value);
	void SetToolWindow(bool value);
	void SetFormPosition(::FormPosition value);
	void SetFormState(::FormState value);

	void DrawMenuBar();
	void EnableMenuItem(WORD id, bool value);
	void CheckMenuItem(WORD id, bool value, bool radio);
	void SetMenuItemText(WORD id, const string &value);

	bool Minimized();
	bool Maximized();
	bool Restored();
	void Minimize();
	void Maximize();
	void Restore();

	HICON Icon();
	HICON SmallIcon();
	LPCTSTR MenuName();

	CMainMenu * get_MainMenu() { return mainMenu; }
	void set_MainMenu(CMainMenu * value);
	__declspec(property(get = get_MainMenu, put = set_MainMenu)) CMainMenu * MainMenu;

	friend class CControl;
};

// Functions
int MessageBox(CForm *owner, const string &text, const string &caption, UINT type = MB_OK);
int MessageBox(CForm *owner, const string &text, UINT type = MB_OK);
int ErrorBox(CForm *owner, exception &e, UINT type = MB_OK | MB_ICONERROR);
int ErrorBox(CForm *owner, exception &e, const string &caption, UINT type = MB_OK | MB_ICONERROR);
int ErrorBox(CForm *owner, _com_error &e, UINT type = MB_OK | MB_ICONERROR);
int ErrorBox(CForm *owner, _com_error &e, const string &caption, UINT type = MB_OK | MB_ICONERROR);

/////////////////
//             //
// CWaitCursor //
//             //
/////////////////

class CWaitCursor
{
private:
	HCURSOR oldCursor;

public:
	CWaitCursor(LPCTSTR cursorName = IDC_WAIT)
	{
		oldCursor = SetCursor(LoadCursor(NULL, cursorName));
	}

	CWaitCursor(HCURSOR cursor)
	{
		oldCursor = SetCursor(cursor);
	}

	~CWaitCursor()
	{
		SetCursor(oldCursor);
	}
};
