#pragma once

#include "Controls.h"
#include "StdControls.h"

class CDialog
{
public:
	virtual int Show(CForm *owner = NULL) = 0;
	virtual void Reset() {}
};

class CFileDialog: public CDialog
{
protected:
	OPENFILENAME ofn;

	void BeforeShow(CForm *owner);
	void AfterShow();

public:
	CFileDialog();

	string DefaultExtension;
	string FileName;
	string Filter;
	int FilterIndex;
	string InitialDirectory;
	string Title;
	strings FileNames;

	bool AllowMultiSelect;
	bool CreatePrompt;
	bool EnableSizing;
	bool FileMustExist;
	bool NoChangeDirectory;
	bool PathMustExist;
	bool OverwritePrompt;

	virtual void Reset();
};

class COpenFileDialog: public CFileDialog
{
public:
	void Reset();
	int Show(CForm *owner = NULL);
};

class CSaveFileDialog: public CFileDialog
{
public:
	void Reset();
	int Show(CForm *owner = NULL);
};

bool BrowseForFolderBox(CForm *owner, const string &title, string &path, bool new_folder = false, bool type_folder = false);

class CColorDialog: public CDialog
{
private:
	CHOOSECOLOR cl;

public:
	CColorDialog();

	COLORREF Color;
	COLORREF CustomColors[16];

	void Reset();
	int Show(CForm *owner = NULL);
};

class CInputForm: public CForm
{
private:
	CLabel * labelPrompt;
	CEdit * editValue;
	CButton * buttonOK, * buttonCancel;

public:
	CInputForm();

	const string & Prompt();
	void SetPrompt(const string &value);

	const string & Value();
	void SetValue(const string &value);
};

bool InputBox(CForm * owner, const string &title, const string &prompt, string &value);
