#pragma once

TCHAR MnemonicFromMessage(MSG &msg);
inline bool IsKeyPressed(BYTE key) { return (GetKeyState(key) & 0x8000) != 0; }
inline bool IsKeyToggled(BYTE key) { return (GetKeyState(key) & 0x0001) != 0; }
string VirtualKeyToString(BYTE key, bool extended);

//////////////
//          //
// Shortcut //
//          //
//////////////

struct Shortcut
{
private:
	WORD data;

	static const WORD ControlMask;
	static const WORD ShiftMask;
	static const WORD AltMask;
	static const WORD KeyMask;

	void Create(BYTE key, bool control, bool shift, bool alt);

public:
	Shortcut(BYTE key = 0, bool control = false, bool shift = false, bool alt = false);
	Shortcut(MSG &msg);
	Shortcut(const Shortcut &value);

	bool get_Control() { return (data & ControlMask) != 0; }
	__declspec(property(get = get_Control)) bool Control;
	
	bool get_Shift() { return (data & ShiftMask) != 0; }
	__declspec(property(get = get_Shift)) bool Shift;
	
	bool get_Alt() { return (data & AltMask) != 0; }
	__declspec(property(get = get_Alt)) bool Alt;
	
	BYTE get_Key() { return (BYTE)((data & KeyMask)); }
	__declspec(property(get = get_Key)) BYTE Key;
	
	bool get_IsNone() { return Key == 0; }
	__declspec(property(get = get_IsNone)) bool IsNone;

	string ToString();

	static const Shortcut None;
};

inline bool operator != (Shortcut value1, Shortcut value2)
{
	return value1.Key != value2.Key || value1.Control != value2.Control
		|| value1.Shift != value2.Shift || value1.Alt != value2.Alt;
}

inline bool operator == (Shortcut value1, Shortcut value2)
{
	return value1.Key == value2.Key && value1.Control == value2.Control
		&& value1.Shift == value2.Shift && value1.Alt == value2.Alt;
}

// Forward declarations
class CMenu;

///////////////
//           //
// CMenuItem //
//           //
///////////////

enum MenuItemType {mitText = 0, mitSeparator = MFT_SEPARATOR, mitOwnerDraw = MFT_OWNERDRAW};
enum MenuBreak {mbNone = 0, mbBreak = MFT_MENUBREAK, mbBarBreak = MFT_MENUBARBREAK};

#ifndef MIIM_STRING
#define MIIM_STRING 0x00000040
#endif
#ifndef MIIM_BITMAP
#define MIIM_BITMAP 0x00000080
#endif
#ifndef MIIM_FTYPE
#define MIIM_FTYPE 0x00000100
#endif

class CMenuItem
{
private:
	CMenu * menu;

	MenuItemType type;
	MenuBreak _break;
	bool radioCheck;
	bool rightJustify;
	bool rightOrder;

	bool checked;
	bool _default;
	bool enabled;

	WORD id;

	CMenu * subMenu;

	string text;
	HBITMAP bitmap;
	void * value;

	HBITMAP checkedBitmap;
	HBITMAP uncheckedBitmap;

	Shortcut shortcut;

	void Create(MenuItemType type = mitText, bool radioCheck = false, WORD id = 0, CMenu * subMenu = NULL,
		const string &text = emptyString, void * value = NULL, Shortcut shortcut = Shortcut::None);

	void OnChanged(DWORD mask);

public:
	CMenuItem();
	CMenuItem(MenuItemType type, WORD id = 0, void * value = NULL);
	CMenuItem(const string &text, WORD id, Shortcut shortcut = Shortcut::None);
	CMenuItem(const string &text, WORD id, void * value, Shortcut shortcut = Shortcut::None);
	CMenuItem(const string &text, bool radioCheck, WORD id, Shortcut shortcut = Shortcut::None);
	CMenuItem(const string &text, bool radioCheck, WORD id, void * value, Shortcut shortcut = Shortcut::None);
	CMenuItem(const string &text, CMenu * subMenu);
	CMenuItem(HMENU hMenu, int index);
	~CMenuItem();

	CMenu * get_Menu() { return menu; }
	__declspec(property(get = get_Menu)) CMenu * Menu;

	int get_Index();
	__declspec(property(get = get_Index)) int Index;

	MenuItemType get_Type() { return type; }
	void set_Type(MenuItemType value);
	__declspec(property(get = get_Type, put = set_Type)) MenuItemType Type;

	MenuBreak get_Break() { return _break; }
	void set_Break(MenuBreak value);
	__declspec(property(get = get_Break, put = set_Break)) MenuBreak Break;

	bool get_RadioCheck() { return radioCheck; }
	void set_RadioCheck(bool value);
	__declspec(property(get = get_RadioCheck, put = set_RadioCheck)) bool RadioCheck;

	bool get_RightJustify() { return rightJustify; }
	void set_RightJustify(bool value);
	__declspec(property(get = get_RightJustify, put = set_RightJustify)) bool RightJustify;

	bool get_RightOrder() { return rightOrder; }
	void set_RightOrder(bool value);
	__declspec(property(get = get_RightOrder, put = set_RightOrder)) bool RightOrder;

	bool get_Checked() { return checked; }
	void set_Checked(bool value);
	__declspec(property(get = get_Checked, put = set_Checked)) bool Checked;

	bool get_Default() { return _default; }
	void set_Default(bool value);
	__declspec(property(get = get_Default, put = set_Default)) bool Default;

	bool get_Enabled() { return enabled; }
	void set_Enabled(bool value);
	__declspec(property(get = get_Enabled, put = set_Enabled)) bool Enabled;

	WORD get_Id() { return id; }
	void set_Id(WORD value);
	__declspec(property(get = get_Id, put = set_Id)) WORD Id;

	CMenu * get_SubMenu() { return subMenu; }
	void set_SubMenu(CMenu * value);
	__declspec(property(get = get_SubMenu)) CMenu * SubMenu;

	bool get_HasSubMenu() { return subMenu != NULL; }
	__declspec(property(get = get_HasSubMenu)) bool HasSubMenu;

	const string & get_Text() { return text; }
	void set_Text(const string & value);
	__declspec(property(get = get_Text, put = set_Text)) const string & Text;

	HBITMAP get_Bitmap() { return bitmap; }
	void set_Bitmap(HBITMAP value);
	__declspec(property(get = get_Bitmap, put = set_Bitmap)) HBITMAP Bitmap;

	void * get_Value() { return value; }
	void set_Value(void * value);
	__declspec(property(get = get_Value, put = set_Value)) void * Value;

	HBITMAP get_CheckedBitmap() { return checkedBitmap; }
	void set_CheckedBitmap(HBITMAP value);
	__declspec(property(get = get_CheckedBitmap, put = set_CheckedBitmap)) HBITMAP CheckedBitmap;

	HBITMAP get_UncheckedBitmap() { return uncheckedBitmap; }
	void set_UncheckedBitmap(HBITMAP value);
	__declspec(property(get = get_UncheckedBitmap, put = set_UncheckedBitmap)) HBITMAP UncheckedBitmap;

	Shortcut get_Shortcut() { return shortcut; }
	void set_Shortcut(Shortcut value);
	__declspec(property(get = get_Shortcut, put = set_Shortcut)) Shortcut Shortcut;

	static const DWORD InfoAllMask;
	void Update(DWORD mask = InfoAllMask);
	
	void Assign(HMENU hMenu, int index);

	bool IsMnemonic(TCHAR ch);

	friend class CMenuItemCollection;
};

/////////////////////////
//                     //
// CMenuItemCollection //
//                     //
/////////////////////////

class CMenuItemCollection: public CCollection<CMenuItem>
{
private:
	CMenu * owner;

protected:
	void OnInsertComplete(int index, CMenuItem * value);
	void OnRemoveComplete(int index, CMenuItem * value);
	
	void OnAdded(CMenuItem * value);
	void OnRemoved(CMenuItem * value);
	void OnValidate(CMenuItem * value);

public:
	CMenuItemCollection(CMenu * owner);
};

///////////
//       //
// CMenu //
//       //
///////////

class CMenu
{
private:
	HMENU handle;
	CMenuItemCollection * items;

	void Create();
	void Destroy();

	void OnMenuItemAdded(int index, CMenuItem * value);
	void OnMenuItemRemoved(int index, CMenuItem * value);

protected:
	virtual void OnChanged();

public:
	CMenu();
	CMenu(HMENU hMenu);
	CMenu(LPCTSTR resourceName);
	~CMenu();

	HMENU get_Handle() { return handle; }
	__declspec(property(get = get_Handle)) HMENU Handle;

	CMenuItemCollection * get_Items() { return items; }
	__declspec(property(get = get_Items)) CMenuItemCollection * Items;

	CMenuItem * GetMenuItemWithMnemonic(TCHAR ch);
	bool IsMnemonic(TCHAR ch);
	CMenuItem * GetMenuItemWithShortcut(Shortcut value);
	CMenuItem * GetMenuItemWithId(WORD value);

	void Assign(HMENU value);
	void Assign(LPCTSTR resourceName);

	void AssignAccelerators(HACCEL value);
	void AssignAccelerators(LPCTSTR resourceName);

	friend class CMenuItem;
	friend class CMenuItemCollection;
};
