#include "System.h"
#include "StdControls.h"
#include "Utils.h"

int GetDefaultControlHeight(LPCTSTR controlClass, DWORD style, DWORD exStyle)
{
	CForm * form = new CForm();
	form->CreateWnd();
	HWND handle = CreateWindowEx(exStyle, controlClass, TEXT(""), WS_CHILD | style,
		CW_USEDEFAULT, 0, CW_USEDEFAULT, 0, form->Handle(), NULL, Application()->Handle(), NULL);
	RECT rect;
	GetWindowRect(handle, &rect);
	int height = rect.bottom - rect.top;
	DestroyWindow(handle);
	delete form;
	return height;
}

////////////
//        //
// CPanel //
//        //
////////////

CPanel::CPanel()
	: CControl()
{
	SetSize(200, 100);
}

///////////////////
//               //
// CCaptionPanel //
//               //
///////////////////

CCaptionPanel::CCaptionPanel()
	: CPanel(), active(false), active_control(NULL)
{
}

#ifndef DC_GRADIENT
#define DC_GRADIENT 0x0020
#endif

LRESULT CCaptionPanel::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
		case WM_NCCALCSIZE:
		{
			LRESULT result = CControl::OnMessage(message, wParam, lParam);
			if(wParam)
			{
				LPNCCALCSIZE_PARAMS p = (LPNCCALCSIZE_PARAMS)lParam;
				p->rgrc[0].top += GetSystemMetrics(SM_CYSMCAPTION) + 2;
			}
			return result;
		}
		case WM_NCPAINT:
		{
			HDC dc = GetWindowDC(Handle());
			RECT rect;
			rect.left = 0;
			rect.top = 0;
			rect.right = Width();
			rect.bottom = GetSystemMetrics(SM_CYSMCAPTION);
			DrawCaption(Handle(), dc, &rect, (active ? DC_ACTIVE : 0) | DC_GRADIENT | DC_SMALLCAP | DC_TEXT);
			rect.top = rect.bottom;
			rect.bottom += 2;
			FillRect(dc, &rect, BackgroundBrush());
			ReleaseDC(Handle(), dc);
			return 0;
		}
		case WM_NCHITTEST: return HTCLIENT;
		case WM_LBUTTONDOWN: case WM_MBUTTONDOWN: case WM_RBUTTONDOWN:
		case WM_NCLBUTTONDOWN: case WM_NCMBUTTONDOWN: case WM_NCRBUTTONDOWN:
			if(!ContainsControl(ParentForm()->ActiveControl()))
				SelectTheControl();
			return 0;
		case WM_SETFOCUS:
			SelectTheControl();
			return 0;
		default: return CControl::OnMessage(message, wParam, lParam);
	}
}

void CCaptionPanel::OnActiveChanged()
{
	UpdateActive();
}

void CCaptionPanel::OnActiveControlChanged()
{
	if(ContainsControl(ParentForm()->ActiveControl()))
		active_control = ParentForm()->ActiveControl();
	UpdateActive();
}

void CCaptionPanel::OnControlParentChanged(CControl *control)
{
	if(control == active_control && !ContainsControl(control))
	{
		active_control = NULL;
		UpdateActive();
	}
}

void CCaptionPanel::UpdateActive()
{
	bool new_active = ParentForm()->Active() && (Focused() || ContainsControl(ParentForm()->ActiveControl()));
	if(new_active != active)
	{
		active = new_active;
		if(HandleAllocated()) SendMessage(WM_NCPAINT, 1, 0);
	}
}

void CCaptionPanel::SelectTheControl()
{
	if(active_control && active_control->CanFocus()) active_control->SetFocus();
	else
	{
		active_control = NULL;
		SelectAnyControl(this);
	}
}

bool CCaptionPanel::SelectAnyControl(CControl *control)
{
	if(control->Visible() && control->Enabled())
	{
		if(control->TabStop())
		{
			control->SetFocus();
			return true;
		}
		else
			for(int i = 0; i < control->ControlCount(); i++)
				if(SelectAnyControl(control->GetControl(i))) return true;
	}
	return false;
}

void CCaptionPanel::SetActiveControl(CControl *value)
{
	if(active_control != value)
	{
		if(!value && active_control->Focused()) ParentForm()->SetFocus();
		else if(ContainsControl(value))
				if(HandleAllocated())
				{
					if(value->CanFocus()) value->SetFocus();
				}
				else active_control = value;
	}
}

///////////////
//           //
// CSplitter //
//           //
///////////////

CSplitter::CSplitter()
	: CControl(), splitting(false), ratio(0.5f), prev_control(NULL), next_control(NULL)
{
}

void CSplitter::SetRatio(float value)
{
	if(value < 0.0f) value = 0.0f;
	if(value > 1.0f) value = 1.0f;
	ratio = value;
	SendCommand(SN_SPLITTED);
}

void CSplitter::SetPrevControl(CControl *value)
{
	prev_control = value;
}

void CSplitter::SetNextControl(CControl *value)
{
	next_control = value;
}

LRESULT CSplitter::OnCancelMode(WPARAM wParam, LPARAM lParam)
{
	CControl::OnCancelMode(wParam, lParam);
	splitting = false;
	return 0;
}

LRESULT CSplitter::OnLButtonDown(DWORD /*keys*/, int x, int y)
{
	SetCapture(true);
	splitting = true;
	x0 = x;
	y0 = y;
	return 0;
}

LRESULT CSplitter::OnLButtonUp(DWORD /*keys*/, int /*x*/, int /*y*/)
{
	splitting = false;
	SetCapture(false);
	return 0;
}

LRESULT CSplitter::OnMouseMove(DWORD keys, int x, int y)
{
	if(splitting && Parent())
	{
		OnSplit(x + X() - x0, y + Y() - y0);
		return  0;
	}
	return CControl::OnMouseMove(keys, x, y);
}


////////////////
//            //
// CHSplitter //
//            //
////////////////

CHSplitter::CHSplitter()
	: CSplitter()
{
	SetHeight(4);
}

void CHSplitter::OnSplit(int /*x*/, int y)
{
	SetRatio((float)(y - AreaTop()) / (float)AreaHeight());
}

int CHSplitter::AreaTop()
{
	return PrevControl() ? PrevControl()->Bottom() : 0;
}

int CHSplitter::AreaBottom()
{
	return NextControl() ? NextControl()->Top() : Parent()->ClientHeight();
}

int CHSplitter::AreaHeight()
{
	return AreaBottom() - AreaTop() - Height();
}

int CHSplitter::TopHeight()
{
	return Round(Ratio() * AreaHeight());
}

int CHSplitter::BottomHeight()
{
	return Round((1.0f - Ratio()) * AreaHeight());
}

void CHSplitter::UpdateControls(int x, int width, CControl * top_control, CControl * bottom_control)
{
	if(top_control)
	{
		top_control->SetLocation(x, AreaTop());
		top_control->SetSize(width, TopHeight());
	}
	SetLocation(x, AreaTop() + TopHeight());
	SetWidth(width);
	if(bottom_control)
	{
		bottom_control->SetLocation(x, Bottom());
		bottom_control->SetSize(width, BottomHeight());
	}
}

////////////////
//            //
// CVSplitter //
//            //
////////////////

CVSplitter::CVSplitter()
	: CSplitter()
{
	SetWidth(4);
}

void CVSplitter::OnSplit(int x, int /*y*/)
{
	SetRatio((float)(x - AreaLeft()) / (float)AreaWidth());
}

int CVSplitter::AreaLeft()
{
	return PrevControl() ? PrevControl()->Right() : 0;
}

int CVSplitter::AreaRight()
{
	return NextControl() ? NextControl()->Left() : Parent()->ClientWidth();
}

int CVSplitter::AreaWidth()
{
	return AreaRight() - AreaLeft() - Width();
}

int CVSplitter::LeftWidth()
{
	return Round(Ratio() * AreaWidth());
}

int CVSplitter::RightWidth()
{
	return Round((1.0f - Ratio()) * AreaWidth());
}

void CVSplitter::UpdateControls(int y, int height, CControl * left_control, CControl * right_control)
{
	if(left_control)
	{
		left_control->SetLocation(AreaLeft(), y);
		left_control->SetSize(LeftWidth(), height);
	}
	SetLocation(AreaLeft() + LeftWidth(), y);
	SetHeight(height);
	if(right_control)
	{
		right_control->SetLocation(Right(), y);
		right_control->SetSize(RightWidth(), height);
	}
}

/////////////
//         //
// CStatic //
//         //
/////////////

CStatic::CStatic()
	: CControl(true)
{
}

DWORD CStatic::GetWndStyle()
{
	return CControl::GetWndStyle() | GetStaticStyle();
}

////////////
//        //
// CLabel //
//        //
////////////

CLabel::CLabel()
	: CStatic(), style(lsStandard), textAlignment(taLeft), ellipsis(teNone), wordWrap(false), noPrefix(false)
{
	SetSize(32, 14);
}

void CLabel::SetStyle(LabelStyle value)
{
	if(style != value)
	{
		style = value;
		RecreateWnd();
	}
}

void CLabel::SetTextAlignment(::TextAlignment value)
{
	if(value == taDefault) value = taLeft;
	if(textAlignment != value)
	{
		textAlignment = value;
		RecreateWnd();
	}
}

void CLabel::SetEllipsis(TextEllipsis value)
{
	if(ellipsis != value)
	{
		ellipsis = value;
		RecreateWnd();
	}
}

void CLabel::SetWordWrap(bool value)
{
	if(wordWrap != value)
	{
		wordWrap = value;
		RecreateWnd();
	}
}

void CLabel::SetNoPrefix(bool value)
{
	if(noPrefix != value)
	{
		noPrefix = value;
		RecreateWnd();
	}
}

DWORD CLabel::GetStaticStyle()
{
	return style
		| (textAlignment == taRight ? SS_RIGHT : textAlignment == taCenter ? SS_CENTER : wordWrap ? SS_LEFT : SS_LEFTNOWORDWRAP)
		| (ellipsis == teEnd ? SS_ENDELLIPSIS : ellipsis == teWord ? SS_WORDELLIPSIS : ellipsis == tePath ? SS_PATHELLIPSIS : 0)
		| (noPrefix ? SS_NOPREFIX : 0);
}

//////////////
//          //
// CPicture //
//          //
//////////////

CPicture::CPicture()
	: CStatic(), bitmap(NULL)
{
	SetSize(100, 100);
}

DWORD CPicture::GetStaticStyle()
{
	return CControl::GetWndStyle() | SS_BITMAP;
}

void CPicture::OnCreate()
{
	UpdateBitmap();
}

void CPicture::SetBitmap(HBITMAP value)
{
	bitmap = value;
	UpdateBitmap();
}

void CPicture::UpdateBitmap()
{
	if(bitmap)
	{
		BITMAP b;
		GetObject(bitmap, sizeof(b), &b);
		SetSize(b.bmWidth, b.bmHeight);
	}
	if(HandleAllocated())
	{
		SendMessage(STM_SETIMAGE, IMAGE_BITMAP, (LPARAM)bitmap);
	}
}

/////////////////////
//                 //
// CSimpleGraphics //
//                 //
/////////////////////

CSimpleGraphics::CSimpleGraphics()
	: CStatic(), style(sgsEtchedFrame)
{
	SetSize(4, 4);
}

DWORD CSimpleGraphics::GetStaticStyle()
{
	return style;
}

void CSimpleGraphics::SetStyle(SimpleGraphicsStyle value)
{
	if(style != value)
	{
		style = value;
		RecreateWnd();
	}
}

////////////////////
//                //
// CButtonControl //
//                //
////////////////////

CButtonControl::CButtonControl()
	: CControl(true)
{
}

DWORD CButtonControl::GetWndStyle()
{
	return CControl::GetWndStyle() | GetButtonStyle();
}

void CButtonControl::Click()
{
	if(CanFocus())
	{
		SendMessage(BM_CLICK, 0, 0);
	}
}

void CButtonControl::OnCommandSent(WORD event)
{
	if(event == BN_CLICKED) OnClick();
}

/////////////
//         //
// CButton //
//         //
/////////////

CButton::CButton()
	: CButtonControl(), def(false), cancel(false), modal_result(0), changing_button_style(false)
{
	SetSize(64, 24);
	SetTabStop(true);
}

DWORD CButton::GetButtonStyle()
{
	return def ? BS_DEFPUSHBUTTON : BS_PUSHBUTTON;
}

void CButton::OnClick()
{
	if(modal_result) ParentForm()->ExitModal(modal_result);
}

bool CButton::OnDefault()
{
	if(def && CanFocus())
	{
		Click();
		return true;
	}
	return false;
}

bool CButton::OnCancel()
{
	if(cancel && CanFocus())
	{
		Click();
		return true;
	}
	return false;
}

void CButton::UpdateButtonStyle()
{
	if(HandleAllocated())
	{
		CForm *form = ParentForm();
		bool active = form ? form->Active() : false;
		CControl *control = form ? form->ActiveControl() : NULL;
		changing_button_style = true;
		SendMessage(BM_SETSTYLE, active &&
			(control == this || (def && (!control || !(control->GetDlgCode() & (DLGC_DEFPUSHBUTTON | DLGC_UNDEFPUSHBUTTON)))))
			? BS_DEFPUSHBUTTON : BS_PUSHBUTTON, TRUE);
		changing_button_style = false;
	}
}

void CButton::OnActiveControlChanged()
{
	UpdateButtonStyle();
}

void CButton::OnActiveChanged()
{
	UpdateButtonStyle();
}

LRESULT CButton::OnMessage(UINT message, WPARAM wParam, LPARAM lParam)
{
	if(message == WM_KEYDOWN && wParam == VK_RETURN)
	{
		Click();
		return 0;
	}
	else if(message == BM_SETSTYLE && !changing_button_style) return 0;
	else return CButtonControl::OnMessage(message, wParam, lParam);
}

void CButton::SetDefault(bool value)
{
	if(def != value)
	{
		def = value;
		RecreateWnd();
	}
}

void CButton::SetCancel(bool value)
{
	cancel = value;
}

void CButton::SetModalResult(int value)
{
	modal_result = value;
}

///////////////
//           //
// CCheckBox //
//           //
///////////////

CCheckBox::CCheckBox()
	: CButtonControl(), checked(false)
{
	SetSize(60, 16);
	SetTabStop(true);
}

DWORD CCheckBox::GetButtonStyle()
{
	return BS_CHECKBOX;
}

void CCheckBox::UpdateChecked()
{
	if(HandleAllocated())
	{
		SendMessage(BM_SETCHECK, checked ? BST_CHECKED : BST_UNCHECKED, 0);
	}
}

void CCheckBox::OnCreate()
{
	UpdateChecked();
}

void CCheckBox::OnClick()
{
	Toggle();
}

void CCheckBox::SetChecked(bool value)
{
	if(checked != value)
	{
		checked = value;
		UpdateChecked();
	}
}

void CCheckBox::Toggle()
{
	SetChecked(!checked);
}

//////////////////
//              //
// CRadioButton //
//              //
//////////////////

CRadioButton::CRadioButton()
	: CButtonControl(), checked(false)
{
	SetSize(60, 16);
}

DWORD CRadioButton::GetButtonStyle()
{
	return BS_RADIOBUTTON;
}

void CRadioButton::UpdateChecked()
{
	if(HandleAllocated())
	{
		SendMessage(BM_SETCHECK, checked ? BST_CHECKED : BST_UNCHECKED, 0);
	}
	if(checked && Parent())
	{
		int j = Parent()->IndexOfControl(this), i;
		for(i = j - 1; i >= 0 && !Parent()->GetControl(i + 1)->Group(); i--)
		{
			CRadioButton *rb = dynamic_cast<CRadioButton*>(Parent()->GetControl(i));
			if(rb) rb->SetChecked(false);
		}
		for(i = j + 1; i < Parent()->ControlCount() && !Parent()->GetControl(i)->Group(); i++)
		{
			CRadioButton *rb = dynamic_cast<CRadioButton*>(Parent()->GetControl(i));
			if(rb) rb->SetChecked(false);
		}		
	}
}

void CRadioButton::OnCreate()
{
	UpdateChecked();
}

void CRadioButton::OnClick()
{
	Check();
}

void CRadioButton::SetParent(CControl *value)
{
	CButtonControl::SetParent(value);
	UpdateChecked();
}

void CRadioButton::SetChecked(bool value)
{
	if(checked != value)
	{
		checked = value;
		UpdateChecked();
	}
}

void CRadioButton::Check()
{
	SetChecked(true);
}

///////////////
//           //
// CGroupBox //
//           //
///////////////

CGroupBox::CGroupBox()
	: CButtonControl()
{
	SetSize(200, 100);
}

void CGroupBox::OnCreate()
{
	CControl::OnCreate();
	SetClassLong(Handle(), GCL_HBRBACKGROUND, (LONG)GetBackgroundBrush());
}

DWORD CGroupBox::GetButtonStyle()
{
	return BS_GROUPBOX;
}

LRESULT CGroupBox::OnEraseBackground(HDC dc, LPARAM /*lParam*/)
{
	RECT rect;
	GetClientRect(rect);
	FillRect(dc, &rect, BackgroundBrush());
	return TRUE;
}

///////////
//       //
// CEdit //
//       //
///////////

CEdit::CEdit()
	: CControl(true), multi_line(false), read_only(false),
	horz_scroll_bar(false), vert_scroll_bar(false), word_wrap(false),
	password_char(0), limit_text(0), changing_text(false)
{
	lines.Add(emptyString);
	SetSize(50, 20);
	SetClientEdge(true);
	SetTabStop(true);
}

DWORD CEdit::GetWndStyle()
{
	return CControl::GetWndStyle()
		| (multi_line ? ES_MULTILINE : 0)
		| (read_only ? ES_READONLY : 0)
		| (horz_scroll_bar ? WS_HSCROLL : 0)
		| (vert_scroll_bar ? WS_VSCROLL : 0)
		| (word_wrap ? 0 : ES_AUTOHSCROLL);
}

void CEdit::OnCreate()
{
	SetPasswordChar(password_char);
	SetLimitText(limit_text);
}

void CEdit::SetMultiLine(bool value)
{
	if(multi_line != value)
	{
		multi_line = value;
		RecreateWnd();
	}
}

void CEdit::SetReadOnly(bool value)
{
	if(read_only != value)
	{
		read_only = value;
		RecreateWnd();
	}
}

void CEdit::SetHorzScrollBar(bool value)
{
	if(horz_scroll_bar != value)
	{
		horz_scroll_bar = value;
		RecreateWnd();
	}
}

void CEdit::SetVertScrollBar(bool value)
{
	if(vert_scroll_bar != value)
	{
		vert_scroll_bar = value;
		RecreateWnd();
	}
}

void CEdit::SetWordWrap(bool value)
{
	if(word_wrap != value)
	{
		word_wrap = value;
		RecreateWnd();
	}
}

void CEdit::OnCommandSent(WORD event)
{
	if(event == EN_CHANGE)
	{
		if(!changing_text)
		{
			int l = GetWindowTextLength(Handle());
			text.SetLength(l);
			GetWindowText(Handle(), text, l + 1);
			DoTextChanged();
		}
	}
}

void CEdit::DoTextChanged()
{
	if(!changing_text)
	{
		UpdateLines();
	}
	CControl::DoTextChanged();
}

LRESULT CEdit::OnSetFocus(HWND hwnd, LPARAM lParam)
{
	LRESULT r = CControl::OnSetFocus(hwnd, lParam);
	if(!multi_line) SelectAll();
	return r;
}

LRESULT CEdit::OnGetDlgCode(WPARAM wParam, MSG *message)
{
	LRESULT r = CControl::OnGetDlgCode(wParam, message);
	r &= ~DLGC_HASSETSEL;
	if(message && (message->message == WM_KEYDOWN || message->message == WM_CHAR) && message->wParam == VK_ESCAPE)
		r &= ~(DLGC_WANTALLKEYS | DLGC_WANTMESSAGE | DLGC_WANTCHARS);
	return r;
}

int CEdit::LineCount()
{
	return lines.Length();
}

void CEdit::InsertLine(int index, const string &value)
{
	CheckIndex(index, true);
	if(index == 0 && text.IsEmpty())
	{
		SetLine(0, value);
		return;
	}
	string line = GetOneLine(value);
	lines.Insert(index, line);
	string line1;
	if(HandleAllocated())
	{
		changing_text = true;
		int s = (int)SendMessage(EM_LINEINDEX, index, 0);
		bool ok = true;
		if(s >= 0) line1 = line + TEXT("\r\n");
		else
		{
			s = (int)SendMessage(EM_LINEINDEX, index - 1, 0);
			if(s >= 0)
			{
				int l = (int)SendMessage(EM_LINELENGTH, s, 0);
				if(l >= 0)
				{
					s += l;
					line1 = TEXT("\r\n") + line;
				}
				else ok = false;
			}
			else ok = false;
		}
		if(ok)
		{
			SendMessage(EM_SETSEL, s, s);
			SendMessage(EM_REPLACESEL, 0, (LPARAM)(LPCTSTR)line1);
		}
		changing_text = false;
	}
	int i;
	for(i = 0; i < text.Length(); i++)
	{
		if(index == 0) break;
		if(text[i] == TEXT('\n')) index--;
	}
	line1 = text.Length() == 0 ? line : i == text.Length() ? STRING("\r\n") + line : line + STRING("\r\n");
	text.Insert(i, line1);
	OnTextChanged();
}

void CEdit::RemoveLine(int index)
{
	CheckIndex(index);
	lines.Remove(index);
	if(lines.IsEmpty()) lines.Add(emptyString);
	if(HandleAllocated())
	{
		changing_text = true;
		int s = (int)SendMessage(EM_LINEINDEX, index, 0);
		if(s >= 0)
		{
			int e = (int)SendMessage(EM_LINEINDEX, index + 1, 0);
			if(e < 0) e = s + (int)SendMessage(EM_LINELENGTH, s, 0);
			SendMessage(EM_SETSEL, s, e);
			SendMessage(EM_REPLACESEL, 0, (LPARAM)(LPCTSTR)emptyString);
		}
		changing_text = false;
	}
	for(int i = 0; i < text.Length(); i++)
	{
		if(index == 0)
		{
			int j = text.IndexOf(TEXT('\n'), i);
			if(j == -1)
			{
				if(i > 0) i -= 2;
				j = text.Length();
			}
			else j++;
			text.Remove(i, j - i);
			break;
		}
		if(text[i] == TEXT('\n')) index--;
	}
	OnTextChanged();
}

void CEdit::ClearLines()
{
	SetText(emptyString);
}

const string& CEdit::Line(int index)
{
	return lines[index];
}

void CEdit::SetLine(int index, const string &value)
{
	CheckIndex(index);
	string line = GetOneLine(value);
	lines[index] = line;
	if(HandleAllocated())
	{
		changing_text = true;
		int s = (int)SendMessage(EM_LINEINDEX, index, 0);
		if(s >= 0)
		{
			SendMessage(EM_SETSEL, s, s + SendMessage(EM_LINELENGTH, s, 0));
			SendMessage(EM_REPLACESEL, 0, (LPARAM)(LPCTSTR)line);
		}
		changing_text = false;
	}
	const TCHAR *t = text.Begin();
	for(int i = 0; t <= text.End(); i++, t++)
	{
		if(index == 0)
		{
			int j = text.IndexOf(TEXT('\n'), i);
			if(j == -1) j = text.Length();
			else j--;
			text.Remove(i, j - i);
			text.Insert(i, line);
			break;
		}
		if(t != text.End() && *t == TEXT('\n')) index--;
	}
	OnTextChanged();
}

void CEdit::SetText(const string &value)
{
	string text = AdjustLineBreaks(value);
	CControl::SetText(text);
}

void CEdit::UpdateLines()
{
	lines.Clear();
	for(int i = 0, j = 0; i <= text.Length(); i++)
		if(i == text.Length() || text[i] == TEXT('\r'))
		{
			lines.Add(text.Substring(j, i - j));
			if(i == text.Length()) break;
			if(text[i] == TEXT('\r')) i++;
			j = i + 1;
		}
}

bool CEdit::CanUndo()
{
	return HandleAllocated() ? SendMessage(EM_CANUNDO, 0, 0) != 0 : false;
}

void CEdit::ClearUndo()
{
	if(HandleAllocated()) SendMessage(EM_EMPTYUNDOBUFFER, 0, 0);
}

void CEdit::Undo()
{
	if(HandleAllocated()) SendMessage(EM_UNDO, 0, 0);
}

bool CEdit::Modified()
{
	return HandleAllocated() ? SendMessage(EM_GETMODIFY, 0, 0) != 0 : false;
}

void CEdit::SetModified(bool value)
{
	if(HandleAllocated()) SendMessage(EM_SETMODIFY, value, 0);
}

int CEdit::SelStart()
{
	if(HandleAllocated())
	{
		DWORD v;
		SendMessage(EM_GETSEL, (WPARAM)&v, NULL);
		return v;
	}
	else return -1;
}

int CEdit::SelEnd()
{
	if(HandleAllocated())
	{
		DWORD v;
		SendMessage(EM_GETSEL, NULL, (WPARAM)&v);
		return v;
	}
	else return -1;
}

int CEdit::SelLength()
{
	if(HandleAllocated())
	{
		DWORD s, e;
		SendMessage(EM_GETSEL, (WPARAM)&s, (LPARAM)&e);
		return e - s;
	}
	else return 0;
}

void CEdit::SetSelStart(int value)
{
	if(HandleAllocated())
	{
		SendMessage(EM_SETSEL, value, SelEnd());
	}
}

void CEdit::SetSelEnd(int value)
{
	if(HandleAllocated())
	{
		SendMessage(EM_SETSEL, SelEnd(), value);
	}
}

void CEdit::SetSelLength(int value)
{
	if(HandleAllocated())
	{
		int s = SelStart();
		SendMessage(EM_SETSEL, s, s + value);
	}
}

void CEdit::SelectAll()
{
	if(HandleAllocated())
	{
		SendMessage(EM_SETSEL, 0, -1);
	}
}

void CEdit::SelectNone()
{
	if(HandleAllocated())
	{
		SendMessage(EM_SETSEL, (WPARAM)-1, -1);
	}
}

string CEdit::SelText()
{
	if(HandleAllocated())
	{
		int s = SelStart();
		int l = SelLength();
		return l >= 0 ? text.Substring(s, l) : text.Substring(s - l, -l);
	}
	else return emptyString;
}

void CEdit::SetSelText(const string &value)
{
	if(HandleAllocated())
		SendMessage(EM_REPLACESEL, 0, (LPARAM)(LPCTSTR)value);
}

void CEdit::ClearSelection()
{
	if(HandleAllocated())
		SendMessage(WM_CLEAR, 0, 0);
}

void CEdit::CopyToClipboard()
{
	if(HandleAllocated())
		SendMessage(WM_COPY, 0, 0);
}

void CEdit::CutToClipboard()
{
	if(HandleAllocated())
		SendMessage(WM_CUT, 0, 0);
}

void CEdit::PasteFromClipboard()
{
	if(HandleAllocated())
		SendMessage(WM_PASTE, 0, 0);
}

TCHAR CEdit::PasswordChar()
{
	return password_char;
}

void CEdit::SetPasswordChar(TCHAR value)
{
	password_char = value;
	if(HandleAllocated())
	{
		SendMessage(EM_SETPASSWORDCHAR, value, 0);
		password_char = (TCHAR)SendMessage(EM_GETPASSWORDCHAR, 0, 0);
	}
}

int CEdit::LimitText()
{
	return limit_text;
}

void CEdit::SetLimitText(int value)
{
	limit_text = value;
	if(HandleAllocated())
	{
		SendMessage(EM_SETLIMITTEXT, value, 0);
		limit_text = SendMessage(EM_GETLIMITTEXT, 0, 0);
	}
}

//////////////
//          //
// CListBox //
//          //
//////////////

CListBox::CListBox()
	: CControl(true), sel_style(lbsSingle)
{
	SetSize(200, 100);
	SetClientEdge(true);
	SetTabStop(true);
}

DWORD CListBox::GetWndStyle()
{
	return CControl::GetWndStyle() | LBS_NOINTEGRALHEIGHT | WS_VSCROLL | LBS_NOTIFY
		| sel_style;
}

void CListBox::OnCreate()
{
	CControl::OnCreate();
	UpdateItems();
}

void CListBox::UpdateItems()
{
	if(HandleAllocated())
	{
		SetRedrawState(false);
		SendMessage(LB_RESETCONTENT, 0, 0);
		for(string* i = items.Begin(); i < items.End(); i++)
			SendMessage(LB_ADDSTRING, 0, (LPARAM)(LPCTSTR)(*i));
		SetRedrawState(true);
	}
}

void CListBox::SetSelStyle(ListBoxSelStyle value)
{
	if(sel_style != value)
	{
		sel_style = value;
		RecreateWnd();
	}
}

int CListBox::ItemCount()
{
	return items.Length();
}

void CListBox::InsertItem(int index, const string &value)
{
	CheckIndex(index, true);
	items.Insert(index, value);
	if(HandleAllocated())
	{
		SendMessage(LB_INSERTSTRING, index, (LPARAM)(LPCTSTR)value);
	}
}

void CListBox::RemoveItem(int index)
{
	CheckIndex(index);
	items.Remove(index);
	if(HandleAllocated())
	{
		SendMessage(LB_DELETESTRING, index, 0);
	}
}

void CListBox::ClearItems()
{
	items.Clear();
	if(HandleAllocated())
	{
		SendMessage(LB_RESETCONTENT, 0, 0);
	}
}

const string& CListBox::Item(int index)
{
	CheckIndex(index);
	return items[index];
}

void CListBox::SetItem(int index, const string &value)
{
	CheckIndex(index);
	items[index] = value;
	if(HandleAllocated())
	{
		bool restoreSel = sel_style == lbsNone ? false
			: sel_style == lbsSingle ? SelIndex() == index
			: ItemSelected(index);
		SendMessage(LB_DELETESTRING, index, 0);
		SendMessage(LB_INSERTSTRING, index, (LPARAM)(LPCTSTR)value);
		if(restoreSel)
			if(sel_style == lbsSingle) SendMessage(LB_SETCURSEL, index, 0);
			else SendMessage(LB_SETSEL, (WPARAM)true, index);
	}
}

int CListBox::SelIndex()
{
	if(HandleAllocated() && sel_style == lbsSingle)
	{
		int i = (int)SendMessage(LB_GETCURSEL, 0, 0);
		return i == LB_ERR ? -1 : i;
	}
	else return -1;
}

void CListBox::SetSelIndex(int index)
{
	if(index >= 0) CheckIndex(index);
	if(HandleAllocated() && sel_style == lbsSingle)
	{
		SendMessage(LB_SETCURSEL, index, 0);
	}
}

int CListBox::SelCount()
{
	if(HandleAllocated() && (sel_style == lbsMultiple || sel_style == lbsExtended))
	{
		int i = (int)SendMessage(LB_GETSELCOUNT, 0, 0);
		return i == LB_ERR ? 0 : i;
	}
	else return 0;
}

bool CListBox::ItemSelected(int index)
{
	CheckIndex(index);
	if(HandleAllocated() && (sel_style == lbsMultiple || sel_style == lbsExtended))
	{
		int i = (int)SendMessage(LB_GETSEL, index, 0);
		return i > 0 ? true : false;
	}
	else return false;
}

void CListBox::SetItemSelected(int index, bool value)
{
	CheckIndex(index);
	if(HandleAllocated() && (sel_style == lbsMultiple || sel_style == lbsExtended))
	{
		SendMessage(LB_SETSEL, (WPARAM)value, index);
	}
}

void CListBox::SelectAll()
{
	if(HandleAllocated() && (sel_style == lbsMultiple || sel_style == lbsExtended))
	{
		SendMessage(LB_SETSEL, TRUE, -1);
		SendCommand(LBN_SELCHANGE);
	}
}

void CListBox::SelectNone()
{
	if(HandleAllocated() && (sel_style == lbsMultiple || sel_style == lbsExtended))
	{
		SendMessage(LB_SETSEL, FALSE, -1);
		SendCommand(LBN_SELCHANGE);
	}
}

///////////////
//           //
// CComboBox //
//           //
///////////////

CComboBox::CComboBox()
	: CControl(true), style(cbsDropDown), item_index(-1)
{
	SetSize(60, GetDefaultControlHeight(GetWndClassName(), CBS_DROPDOWN));
	SetTabStop(true);
}

DWORD CComboBox::GetWndStyle()
{
	return CControl::GetWndStyle() | WS_VSCROLL | style;
}

void CComboBox::SetStyle(ComboBoxStyle value)
{
	if(style != value)
	{
		style = value;
		if(style == cbsDropDownList) text.Clear();
		item_index = -1;
		RecreateWnd();
	}
}

void CComboBox::UpdateItems()
{
	if(HandleAllocated())
	{
		SetRedrawState(false);
		SendMessage(CB_RESETCONTENT, 0, 0);
		for(string * i = items.Begin(); i < items.End(); i++)
		{
			SendMessage(CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)(*i));
		}
		SetText(text);
		SetRedrawState(true);
		UpdateList();
	}
}

void CComboBox::UpdateList()
{
	if(HandleAllocated())
	{
		// Store selection;
		DWORD s, e;
		SendMessage(CB_GETEDITSEL, (WPARAM)&s, (LPARAM)&e);
		// Set list height
		int height = Height() + SendMessage(CB_GETITEMHEIGHT, 0, 0) * min(ItemCount(), 16) + 2;
		SetWindowPos(Handle(), NULL, 0, 0, Width(), height, SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOREPOSITION | SWP_NOZORDER);
		// Restore selection
		SendMessage(CB_SETEDITSEL, 0, MAKELPARAM((WORD)s, (WORD)e));
	}
}

void CComboBox::UpdateItemIndex()
{
	if(HandleAllocated())
	{
		SendMessage(CB_SETCURSEL, item_index, 0);
	}
}

void CComboBox::OnCreate()
{
	CControl::OnCreate();
	UpdateItems();
	if(item_index != -1) UpdateItemIndex();
}

void CComboBox::SetText(const string &value)
{
	if(style != cbsDropDownList) CControl::SetText(value);
}

void CComboBox::OnCommandSent(WORD event)
{
	if(event == CBN_SELCHANGE)
	{
		item_index = SendMessage(CB_GETCURSEL, 0, 0);
		if(item_index == CB_ERR)
		{
			item_index = -1;
			text.Clear();
		}
		else text = items[item_index];
	}
	else if(event == CBN_EDITCHANGE)
	{
		int l = GetWindowTextLength(Handle());
		text.SetLength(l);
		GetWindowText(Handle(), text, l + 1);
		DoTextChanged();
	}
}

int CComboBox::ItemCount()
{
	return items.Length();
}

void CComboBox::ClearItems()
{
	items.Clear();
	values.Clear();
	if(HandleAllocated())
	{
		SendMessage(LB_RESETCONTENT, 0, 0);
		UpdateList();
	}
}

void CComboBox::InsertItem(int index, const string &text, void * value)
{
	CheckIndex(index, true);
	items.Insert(index, text);
	values.Insert(index, value);
	if(HandleAllocated())
	{
		SendMessage(CB_INSERTSTRING, index, (LPARAM)(LPCTSTR)text);
		UpdateList();
	}
}

void CComboBox::RemoveItem(int index)
{
	CheckIndex(index);
	items.Remove(index);
	values.Remove(index);
	if(HandleAllocated())
	{
		SendMessage(CB_DELETESTRING, index, 0);
		UpdateList();
	}
}

const string & CComboBox::Item(int index)
{
	CheckIndex(index);
	return items[index];
}

void CComboBox::SetItem(int index, const string &value)
{
	CheckIndex(index);
	items[index] = value;
	if(HandleAllocated())
	{
		SendMessage(CB_SETITEMDATA, index, (LPARAM)(LPCTSTR)value);
	}
}

void * CComboBox::Value(int index)
{
	CheckIndex(index);
	return values[index];
}

void CComboBox::SetValue(int index, void * value)
{
	CheckIndex(index);
	values[index] = value;
}

int CComboBox::ItemIndex()
{
	return item_index;
}

void CComboBox::SetItemIndex(int value)
{
	if(value >= 0) CheckIndex(value);
	if(value < 0) value = -1;
	if(item_index != value)
	{
		item_index = value;
		if(value == -1) text.Clear();
		else text = items[value];
	}
	UpdateItemIndex();
}

int CComboBox::IndexOfItem(const string &value)
{
	int ii = 0;
	for(string * i = items.Begin(); i < items.End(); i++, ii++)
		if(*i == value) return ii;
	return -1;
}

int CComboBox::IndexOfValue(void * value)
{
	int ii = 0;
	for(void ** i = values.Begin(); i < values.End(); i++, ii++)
		if(*i == value) return ii;
	return -1;
}

///////////////
//           //
// CRichEdit //
//           //
///////////////

CRichEdit::CRichEdit()
	: CEdit()
{
	SetSize(200, 100);
	SetMultiLine(true);
	SetVertScrollBar(true);
	SetHorzScrollBar(true);
}

void CRichEdit::OnCreate()
{
	SendMessage(EM_SETEVENTMASK, 0, ENM_CHANGE);
}

///////////////
//           //
// CListView //
//           //
///////////////

CListView::CListView()
	: CControl(true), owner_data(false), no_column_header(false), no_sort_header(false), show_sel_always(false), single_sel(true),
	full_row_select(false), itemCount(0)
{
	SetSize(200, 100);
	SetClientEdge(true);
	SetTabStop(true);
}

DWORD CListView::GetWndStyle()
{
	return CControl::GetWndStyle() | LVS_REPORT
		| (owner_data ? LVS_OWNERDATA : 0)
		| (no_column_header ? LVS_NOCOLUMNHEADER : 0)
		| (no_sort_header ? LVS_NOSORTHEADER : 0)
		| (show_sel_always ? LVS_SHOWSELALWAYS : 0)
		| (single_sel ? LVS_SINGLESEL : 0);
}

void CListView::OnCreate()
{
	CControl::OnCreate();
	ListView_SetExtendedListViewStyleEx(Handle(), LVS_EX_FULLROWSELECT, full_row_select ? LVS_EX_FULLROWSELECT : 0);
	UpdateColumns();
	UpdateItems();
}

const int max_lvcolumn_header_text = 1000;

void CListView::UpdateColumns()
{
	if(HandleAllocated())
	{
		for(int i = 0; i < columns.Length(); i++)
		{
			LVCOLUMN col;
			TCHAR buf[max_lvcolumn_header_text];
			col.pszText = buf;
			col.cchTextMax = max_lvcolumn_header_text;
			col.mask = LVCF_TEXT;
			columns[i].Text.Copy(buf, 0, max_lvcolumn_header_text - 1);
			buf[columns[i].Text.Length()] = TEXT('\0');
			if(columns[i].Width != -1)
			{
				col.mask |= LVCF_WIDTH;
				col.cx = columns[i].Width;
			}
			if(columns[i].Align != taDefault)
			{
				col.mask |= LVCF_FMT;
				switch(columns[i].Align)
				{
					case taLeft: col.fmt = LVCFMT_LEFT; break;
					case taCenter: col.fmt = LVCFMT_CENTER; break;
					case taRight: col.fmt = LVCFMT_RIGHT; break;
					default: col.fmt = 0;
				}
			}
			ListView_InsertColumn(Handle(), i, &col);
			col.mask = LVCF_TEXT | LVCF_WIDTH | LVCF_FMT;
			ListView_GetColumn(Handle(), i, &col);
			columns[i].Text.Assign(col.pszText);
			columns[i].Width = col.cx;
			switch(col.mask & LVCFMT_JUSTIFYMASK)
			{
				case LVCFMT_LEFT: columns[i].Align = taLeft; break;
				case LVCFMT_CENTER: columns[i].Align = taCenter; break;
				case LVCFMT_RIGHT: columns[i].Align = taRight; break;
				default: columns[i].Align = taDefault;
			}
		}
	}
}

void CListView::SetOwnerData(bool value)
{
	if(owner_data != value)
	{
		owner_data = value;
		if(owner_data) items.Clear();
		else itemCount = 0;
		RecreateWnd();
	}
}

int CListView::ColumnCount()
{
	return columns.Length();
}

void CListView::ClearColumns()
{
	columns.Clear();
	for(ListViewItem * i = items.Begin(); i < items.End(); i++)
		i->Subitems.Clear();
	RecreateWnd();
}

void CListView::InsertColumn(int index, const string & text, int width, TextAlignment align)
{
	CheckColumnIndex(index, true);
	columns.Insert(index, ListViewColumn(text, width, align));
	for(ListViewItem * i = items.Begin(); i < items.End(); i++)
		i->Subitems.Insert(index, emptyString);
	RecreateWnd();
}

void CListView::RemoveColumn(int index)
{
	CheckColumnIndex(index);
	columns.Remove(index);
	for(ListViewItem * i = items.Begin(); i < items.End(); i++)
		i->Subitems.Remove(index);
	RecreateWnd();
}

void CListView::SetNoColumnHeader(bool value)
{
	if(no_column_header != value)
	{
		no_column_header = value;
		RecreateWnd();
	}
}

void CListView::SetNoSortHeader(bool value)
{
	if(no_sort_header != value)
	{
		no_sort_header = value;
		RecreateWnd();
	}
}

void CListView::SetShowSelAlways(bool value)
{
	if(show_sel_always != value)
	{
		show_sel_always = value;
		RecreateWnd();
	}
}

void CListView::SetSingleSel(bool value)
{
	if(single_sel != value)
	{
		single_sel = value;
		RecreateWnd();
	}
}

void CListView::SetFullRowSelect(bool value)
{
	if(full_row_select != value)
	{
		full_row_select = value;
		ListView_SetExtendedListViewStyleEx(Handle(), LVS_EX_FULLROWSELECT, full_row_select ? LVS_EX_FULLROWSELECT : 0);
	}
}

void CListView::UpdateItems()
{
	if(HandleAllocated())
	{
		if(owner_data)
		{
			ListView_SetItemCountEx(Handle(), itemCount, 0);
		}
		else
		{
			SetRedrawState(false);
			ListView_SetItemCount(Handle(), items.Length());
			int c = columns.Length();
			for(int i = 0; i < items.Length(); i++)
			{
					LVITEM lvi;
					lvi.mask = LVIF_TEXT;
					lvi.iItem = (INT)i;
					lvi.iSubItem = 0;
					lvi.pszText = items[i].Text;
					ListView_InsertItem(Handle(), &lvi);
					for(int j = 1; j < c; j++)
					{
						lvi.iSubItem = j;
						lvi.pszText = items[i].Subitems[j];
						ListView_SetItem(Handle(), &lvi);
					}
			}
			SetRedrawState(true);
		}
	}
}

int CListView::ItemCount()
{
	return owner_data ? itemCount : items.Length();
}

void CListView::SetItemCount(int value)
{
	if(owner_data)
	{
		itemCount = value;
		if(HandleAllocated())
			ListView_SetItemCountEx(Handle(), itemCount, 0);
	}
	else
	{
		items.SetLength(value);
		if(HandleAllocated())
			ListView_SetItemCount(Handle(), value);
	}
}

void CListView::ClearItems()
{
	items.Clear();
	itemCount = 0;
	if(HandleAllocated())
	{
		ListView_DeleteAllItems(Handle());
	}
}

void CListView::InsertItem(int index, const string & text)
{
	if(!owner_data)
	{
		CheckIndex(index, true);
		items.Insert(index, ListViewItem(text));
		for(int j = 0; j < columns.Length(); j++)
			items[index].Subitems.Insert(j, emptyString);
	}
	else itemCount++;
	if(HandleAllocated())
	{
		LVITEM lvi;
		lvi.mask = owner_data ? 0 : LVIF_TEXT;
		lvi.iItem = index;
		lvi.iSubItem = 0;
		if(!owner_data) lvi.pszText = (LPTSTR)(LPCTSTR)text;
		ListView_InsertItem(Handle(), &lvi);
	}
}

void CListView::RemoveItem(int index)
{
	if(!owner_data)
	{
		CheckIndex(index);
		items.Remove(index);
	}
	else itemCount--;
	if(HandleAllocated())
	{
		ListView_DeleteItem(Handle(), index);
	}
}

const string & CListView::ItemText(int index)
{
	if(!owner_data)
	{
		CheckIndex(index);
		return items[index].Text;
	}
	else return emptyString;
}

const string & CListView::SubItemText(int index, int subindex)
{
	if(!owner_data)
	{
		CheckSubIndex(index, subindex);
		return subindex == 0 ? ItemText(index) : items[index].Subitems[subindex];
	}
	else return emptyString;
}

void CListView::SetItemText(int index, const string & text)
{
	if(!owner_data)
	{
		CheckIndex(index);
		items[index].Text = text;
	}
	if(HandleAllocated())
	{
		ListView_SetItemText(Handle(), index, 0, (LPTSTR)(LPCTSTR)text);
	}
}

void CListView::SetSubItemText(int index, int subindex, const string & text)
{
	CheckSubIndex(index, subindex);
	if(subindex == 0) SetItemText(index, text);
	else
	{
		if(!owner_data)
		{
			items[index].Subitems[subindex] = text;
		}
		if(HandleAllocated())
		{
			ListView_SetItemText(Handle(), index, subindex, (LPTSTR)(LPCTSTR)text);
		}
	}
}

int CListView::FocusIndex()
{
	return HandleAllocated() ? ListView_GetNextItem(Handle(), -1, LVNI_ALL | LVNI_FOCUSED) : -1;
}

void CListView::SetFocusIndex(int value)
{
	if(value >= 0) CheckIndex(value);
	if(HandleAllocated())
	{
		ListView_SetItemState(Handle(), value, 0, LVIS_FOCUSED);
	}
}

int CListView::SelIndex()
{
	return HandleAllocated() ? ListView_GetNextItem(Handle(), -1, LVNI_ALL | LVNI_SELECTED) : -1;
}

void CListView::SetSelIndex(int value)
{
	if(value >= 0) CheckIndex(value);
	if(HandleAllocated())
	{
		ListView_SetItemState(Handle(), -1, 0, LVIS_SELECTED);
		if(value != -1)
			ListView_SetItemState(Handle(), value, LVIS_SELECTED, LVIS_SELECTED);
	}
}

int CListView::SelCount()
{
	return HandleAllocated() ? ListView_GetSelectedCount(Handle()) : 0;
}

bool CListView::ItemSelected(int index)
{
	CheckIndex(index);
	return HandleAllocated() ? (ListView_GetItemState(Handle(), index, LVIS_SELECTED) & LVIS_SELECTED) != 0 : false;
}

void CListView::SetItemSelected(int index, bool value)
{
	CheckIndex(index);
	if(HandleAllocated())
	{
		ListView_SetItemState(Handle(), index, value ? LVIS_SELECTED : 0, LVIS_SELECTED);
	}
}

void CListView::SelectAll()
{
	if(HandleAllocated())
		ListView_SetItemState(Handle(), -1, LVIS_SELECTED, LVIS_SELECTED);
}

void CListView::SelectNone()
{
	if(HandleAllocated())
		ListView_SetItemState(Handle(), -1, 0, LVIS_SELECTED);
}

void CListView::EnsureVisible(int index, bool partialOK)
{
	CheckIndex(index);
	if(HandleAllocated())
	{
		ListView_EnsureVisible(Handle(), index, partialOK);
	}
}

RECT CListView::ItemRect(int index, int code)
{
	CheckIndex(index);
	RECT r;
	ListView_GetItemRect(Handle(), index, &r, code);
	return r;
}

RECT CListView::SubItemRect(int index, int subIndex, int code)
{
	CheckSubIndex(index, subIndex);
	RECT r;
	if(subIndex) ListView_GetSubItemRect(Handle(), index, subIndex, code, &r);
	else
	{
		ListView_GetItemRect(Handle(), index, &r, code);
		if(columns.Length() > 1)
		{
			RECT r1;
			ListView_GetSubItemRect(Handle(), index, 1, code, &r1);
			r.right = r1.left;
		}
	}
	return r;
}

void CListView::RedrawItems(int index, int count)
{
	CheckIndex(index);
	CheckIndex(index + count, true);
	ListView_RedrawItems(Handle(), index, index + count - 1);
}

void CListView::RedrawItems()
{
	Invalidate();
}


//////////////////
//              //
// CProgressBar //
//              //
//////////////////

CProgressBar::CProgressBar()
	: CControl(true)
{
	min = 0;
	max = 100;
	position = 0;
	SetSize(100, 16);
}

void CProgressBar::SetMin(int value)
{
	if(min != value)
	{
		min = value;
		if(max < min) max = min;
		if(position < min) position = min;
		UpdateRange();
	}
}

void CProgressBar::SetMax(int value)
{
	if(max != value)
	{
		max = value;
		if(min > max) min = max;
		if(position > max) position = max;
		UpdateRange();
	}
}

void CProgressBar::SetPosition(int value)
{
	value = max(min, min(max, value));
	if(position != value)
	{
		position = value;
		UpdatePosition();
	}
}

void CProgressBar::UpdateRange()
{
	if(HandleAllocated())
	{
		SendMessage(PBM_SETRANGE32, (WPARAM)min, (LPARAM)max);
		UpdatePosition();
	}
}

void CProgressBar::UpdatePosition()
{
	if(HandleAllocated())
	{
		SendMessage(PBM_SETPOS, (WPARAM)position, 0);
	}
}

///////////////
//           //
// CTrackBar //
//           //
///////////////

CTrackBar::CTrackBar()
	: CControl(true)
{
	min = 0;
	max = 10;
	position = 0;
	line_size = 1;
	page_size = 2;
	SetSize(100, 30);
	SetTabStop(true);
}

void CTrackBar::OnCreate()
{
	CControl::OnCreate();
	UpdateRange();
	UpdateLineSize();
	UpdatePageSize();
	UpdateFirstBuddy();
	UpdateLastBuddy();
}

void CTrackBar::OnHScrollSent(WORD /*code*/, WORD /*pos*/)
{
	position = SendMessage(TBM_GETPOS, 0, 0);
}

void CTrackBar::OnVScrollSent(WORD /*code*/, WORD /*pos*/)
{
	position = SendMessage(TBM_GETPOS, 0, 0);
}

void CTrackBar::UpdateRange()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETRANGEMIN, FALSE, min);
		SendMessage(TBM_SETRANGEMAX, FALSE, max);
	}
	UpdatePosition();
}

void CTrackBar::UpdatePosition()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETPOS, TRUE, position);
	}
}

void CTrackBar::UpdateLineSize()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETLINESIZE, 0, line_size);
	}
}

void CTrackBar::UpdatePageSize()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETPAGESIZE, 0, line_size);
		line_size = SendMessage(TBM_GETPAGESIZE, 0, 0);
	}
}

void CTrackBar::UpdateFirstBuddy()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETBUDDY, TRUE, first_buddy ? (LPARAM)first_buddy->Handle() : NULL);
	}
}

void CTrackBar::UpdateLastBuddy()
{
	if(HandleAllocated())
	{
		SendMessage(TBM_SETBUDDY, FALSE, last_buddy ? (LPARAM)last_buddy->Handle() : NULL);
	}
}

void CTrackBar::SetMin(int value)
{
	if(min != value)
	{
		min = value;
		if(max < min) max = min;
		if(position < min) position = min;
		UpdateRange();
	}
}

void CTrackBar::SetMax(int value)
{
	if(max != value)
	{
		max = value;
		if(min > max) min = max;
		if(position > max) position = max;
		UpdateRange();
	}
}

void CTrackBar::SetPosition(int value)
{
	value = max(min, min(max, value));
	if(position != value)
	{
		position = value;
		UpdatePosition();
	}
}

void CTrackBar::SetLineSize(int value)
{
	value = max(1, value);
	if(line_size != value)
	{
		line_size = value;
		UpdateLineSize();
	}
}

void CTrackBar::SetPageSize(int value)
{
	value = max(1, value);
	if(page_size != value)
	{
		page_size = value;
		UpdatePageSize();
	}
}

void CTrackBar::SetFirstBuddy(CControl * value)
{
	if(first_buddy != value)
	{
		first_buddy = value;
		UpdateFirstBuddy();
	}
}

void CTrackBar::SetLastBuddy(CControl * value)
{
	if(last_buddy != value)
	{
		last_buddy = value;
		UpdateLastBuddy();
	}
}

////////////////
//            //
// CStatusBar //
//            //
////////////////

#ifndef SB_SIMPLEID
#define SB_SIMPLEID 0x00FF
#endif

const int max_statusbar_parts = 256;

CStatusBar::CStatusBar()
	: CControl(true), size_grip(true), simple(false), text_align(taDefault), text_border(bsDefault)
{
	SetHeight(GetDefaultControlHeight(GetWndClassName()));
}

DWORD CStatusBar::GetWndStyle()
{
	return CControl::GetWndStyle() | (size_grip ? SBARS_SIZEGRIP : 0);
}

void CStatusBar::OnCreate()
{
	UpdateParts();
	UpdateSimple();
	UpdatePart(SB_SIMPLEID);
}

void CStatusBar::SetSizeGrip(bool value)
{
	if(size_grip != value)
	{
		size_grip = value;
		RecreateWnd();
	}
}

void CStatusBar::SetSimple(bool value)
{
	if(simple != value)
	{
		simple = value;
		UpdateSimple();
	}
}

void CStatusBar::UpdateSimple()
{
	if(HandleAllocated())
	{
		SendMessage(SB_SIMPLE, simple, 0);
	}
}

void CStatusBar::UpdatePart(int index)
{
	if(HandleAllocated())
	{
		string text;
		TextAlignment align;
		BorderStyle border;
		if(index == SB_SIMPLEID)
		{
			text = this->text;
			align = text_align;
			border = text_border;
		}
		else
		{
			text = parts[index].Text;
			align = parts[index].Align;
			border = parts[index].Border;
		}
		switch(align)
		{
			case taCenter: text = STRING("\t") + text; break;
			case taRight: text = STRING("\t\t") + text; break;
		}
		DWORD type = 0;
		switch(border)
		{
			case bsNone: type = SBT_NOBORDERS; break;
			case bsRaised: type = SBT_POPOUT; break;
		}
		SendMessage(SB_SETTEXT, (DWORD)index | type, (LPARAM)(LPCTSTR)text);
	}
}

void CStatusBar::UpdateParts()
{
	if(HandleAllocated())
	{
		INT widths[max_statusbar_parts];
		if(parts.Length()) widths[0] = parts[0].Width;
		int i;
		for(i = 1; i < parts.Length(); i++)
			widths[i] = parts[i].Width < 0 ? -1 : widths[i - 1] + parts[i].Width;
		SendMessage(SB_SETPARTS, parts.Length(), (LPARAM)widths);
		for(i = 0; i < parts.Length(); i++)
			UpdatePart(i);
	}
}

int CStatusBar::PartCount()
{
	return parts.Length();
}

void CStatusBar::InsertPart(int index, int width, const string & text, TextAlignment align, BorderStyle border)
{
	CheckIndex(index, true);
	if(parts.Length() == max_statusbar_parts) throw exception("Max status bar part count achieved");
	parts.Insert(index, StatusBarPart(width, text, align, border));
	UpdateParts();
}

void CStatusBar::RemovePart(int index)
{
	CheckIndex(index);
	parts.Remove(index);
	UpdateParts();
}

void CStatusBar::ClearParts()
{
	parts.Clear();
	UpdateParts();
}

void CStatusBar::SetText(const string & value)
{
	CControl::SetText(value);
	UpdatePart(SB_SIMPLEID);
}

void CStatusBar::SetTextAlign(TextAlignment value)
{
	if(text_align != value)
	{
		text_align = value;
		UpdatePart(SB_SIMPLEID);
	}
}

void CStatusBar::SetTextBorder(BorderStyle value)
{
	if(text_border != value)
	{
		text_border = value;
		UpdatePart(SB_SIMPLEID);
	}
}

const string & CStatusBar::PartText(int index)
{
	CheckIndex(index);
	return parts[index].Text;
}

TextAlignment CStatusBar::PartAlign(int index)
{
	CheckIndex(index);
	return parts[index].Align;
}

BorderStyle CStatusBar::PartBorder(int index)
{
	CheckIndex(index);
	return parts[index].Border;
}

int CStatusBar::PartWidth(int index)
{
	CheckIndex(index);
	return parts[index].Width;
}

void CStatusBar::SetPartText(int index, const string & value)
{
	CheckIndex(index);
	parts[index].Text = value;
	UpdatePart(index);
}

void CStatusBar::SetPartAlign(int index, TextAlignment value)
{
	CheckIndex(index);
	if(parts[index].Align != value)
	{
		parts[index].Align = value;
		UpdatePart(index);
	}
}

void CStatusBar::SetPartBorder(int index, BorderStyle value)
{
	CheckIndex(index);
	if(parts[index].Border != value)
	{
		parts[index].Border = value;
		UpdatePart(index);
	}
}

void CStatusBar::SetPartWidth(int index, int value)
{
	CheckIndex(index);
	if(parts[index].Width != value)
	{
		parts[index].Width = value;
		UpdateParts();
	}
}

////////////////
//            //
// CImageList //
//            //
////////////////

CImageList::CImageList()
{
	Init(16, 16);
}

CImageList::CImageList(int width, int height)
{
	if(width < 1 || height < 1) throw exception("Size of image list images too small");
	Init(width, height);
}

CImageList::~CImageList()
{
	if(handle) ImageList_Destroy(handle);
}

void CImageList::Init(int width, int height)
{
	handle = ImageList_Create(width, height, ILC_COLOR24 | ILC_MASK, 0, 4);
	if(!handle) throw exception("Can not create image list");
	this->width = width;
	this->height = height;
}

int CImageList::Count()
{
	return ImageList_GetImageCount(handle);
}

int CImageList::Add(HICON icon)
{
	int i = ImageList_AddIcon(handle, icon);
	if(i < 0) throw exception("Can not add icon to image list");
	return i;
}

int CImageList::Add(HBITMAP bitmap)
{
	BITMAP bmp;
	if(!GetObject(bitmap, sizeof(BITMAP), &bmp)) throw exception("Can not access bitmap information");
	HDC dc = CreateCompatibleDC(NULL);
	if(dc)
	{
		HBITMAP old_bitmap = SelectBitmap(dc, bitmap);
		COLORREF color = GetPixel(dc, 0, bmp.bmHeight - 1);
		SelectBitmap(dc, old_bitmap);
		DeleteDC(dc);
		return Add(bitmap, color);
	}
	else throw exception("Error accessing bitmap information");
}

int CImageList::Add(HBITMAP bitmap, COLORREF maskColor)
{
	int i = ImageList_AddMasked(handle, bitmap, maskColor);
	if(i < 0) throw exception("Can not add bitmap to image list");
	return i;
}

//////////////
//          //
// CToolBar //
//          //
//////////////

CToolBar::CToolBar()
	: CControl(true)
{
}
