#include "System.h"
#include "Threads.h"

CThread::CThread(bool createSuspended)
	: running(false)
{
	handle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)ThreadProc, this, createSuspended ? CREATE_SUSPENDED : 0, &id);
	if(!handle) throw exception("Can not create thread");
}

CThread::~CThread()
{
	WaitFor();
	CloseHandle(handle);
}

DWORD WINAPI CThread::ThreadProc(LPVOID parameter)
{
	return ((CThread *)parameter)->Execute();
}

DWORD CThread::Execute()
{
	OnStartExecute();
	DWORD exitCode = OnExecute();
	OnEndExecute(exitCode);
	return exitCode;
}

void CThread::OnStartExecute()
{
	running = true;
}

void CThread::OnEndExecute(DWORD /*exitCode*/)
{
	running = false;
}

void CThread::Resume()
{
	if(ResumeThread(handle) == -1) throw exception("Can not resume thread");
}

void CThread::Suspend()
{
	if(SuspendThread(handle) == -1) throw exception("Can not suspend thread");
}

void CThread::Terminate(DWORD exit_code)
{
	if(!TerminateThread(handle, exit_code)) throw exception("Can not terminate thread");
}

void CThread::WaitFor(DWORD ms)
{
	WaitForSingleObject(handle, ms);
}

DWORD CThread::get_ExitCode()
{
	DWORD exitCode;
	if(!GetExitCodeThread(handle, &exitCode)) throw exception("Can not retrieve thread exit code");
	return exitCode;
}
