import java.awt.Container;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.DefaultListModel;
import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import neurotechnologija.utils.BMPLoader;
import neurotechnologija.utils.ImagePanel;
import neurotechnologija.utils.Utils;

import ScanMan.ScanMan;
import ScanMan.ScanManException;
import ScanMan.ScannerEventListener;
import VeriFinger.FingerDatabase;
import VeriFinger.FingerprintData;
import VeriFinger.VeriFingerException;
import VeriFinger.VeriFingerExtractionResult;
import VeriFinger.VeriFingerFeatures;
import VeriFinger.VeriFingerGeneralizationResult;
import VeriFinger.VeriFingerMatchDetails;
import VeriFinger.VeriFingerWrapper;


/**
 *
 * To change this generated comment edit the template variable "typecomment":
 * Window>Preferences>Java>Templates.
 * To enable and disable the creation of type comments go to
 * Window>Preferences>Java>Code Generation.
 */
/**
 * VeriFinger 4.2 Demo<br/>
 * using VeriFinger and UareU wrappers<br/>
 *
 * &copy;2002 by <a href="http://www.neurotechnologija.com/">Neurotechnologija</a>
 * @version 1.0 02/07/2002
 */
public class VFingerJava extends JFrame implements ActionListener {

        // fields
        private FingerDatabase fdb = new FingerDatabase();
        private DefaultListModel listModel = new DefaultListModel();

        public static final String CONFIG_FILE = "vfdemo.cfg";

        // zoom
        private int zoom = 100;
        private static final int ZOOM_MIN = 50;
        private static final int ZOOM_MAX = 1000;
        private static final int ZOOM_STEP = 50;

        // number of images for enrollment with generalization
        private static final int GEN_IMAGE_COUNT = 3;

        /**
         * Main frame constructor. User interface creation.
         * */
        public VFingerJava() {
                loadSettings();
                initVFinger();

                // destroy on exit
                addWindowListener(new WindowAdapter() {
                        public void windowClosing(WindowEvent e) {
                                onClose();
                        }
                });

                // initialize
                initGui();
                initMenu();

                // disable menu to prevent use of uninitialized features
                file.setEnabled(false);
                service.setEnabled(false);
                source.setEnabled(false);

                // init frame
                setSize(640, 480);
                Utils.centerOnScreen(this);
                setTitle("VFinger Demo");
                setVisible(true);


                int cnt = 0;
                try {
                        cnt = ScanMan.SMGetScannerCount();
                } catch (Exception e) {
                        System.out.println(
                                "Error getting scanner count - " + e.getMessage());
                                e.printStackTrace();
                }

                if (cnt == 0) {
                        sourceScanner.setEnabled(false);
                        sourceFile.setSelected(true);
                        file.setText("File");
                        Settings.imageSource = Settings.IMAGE_SOURCE_FILE;
                }
                else {
                        sourceScanner.setSelected(true);
                        file.setText("Scanner");
                        Settings.imageSource = Settings.IMAGE_SOURCE_SCANNER;
                }

                logEvent("Loading fingerprint database: " + Settings.dbFile + "...");
                fdb.load(Settings.dbFile);
                logEvent(
                        "Done (" + Integer.toString(fdb.getCount()) + " fingerprints)");

                // enable menu
                file.setEnabled(true);
                service.setEnabled(true);
                source.setEnabled(true);
        }

        /**
         * User interface action handler
         * */
        public void actionPerformed(ActionEvent e) {
                Object source = e.getSource();

                if (source == fileExit) {
                        onClose();
                } else if (source == fileEnrollment) {
                        onEnrollment();
                } else if (source == enrollWithGen) {
                        onEnrollWithGen();
                } else if (source == fileRecognitionFast) {
                        onRecognition(false);
                } else if (source == fileRecognitionFull) {
                        onRecognition(true);
                } else if (source == serviceProperties) {
                        onProperties();
                } else if (source == fileAbout) {
                        AboutDlg dlg = new AboutDlg(this);
                        dlg.setVisible(true);
                } else if (source == sourceScanner) {
                        Settings.imageSource = Settings.IMAGE_SOURCE_SCANNER;
                        file.setText("Scanner");
                        Settings.imageDpi = 500;
                } else if (source == sourceFile) {
                        file.setText("File");
                        Settings.imageSource = Settings.IMAGE_SOURCE_FILE;
                } else if (source == clearLog) {
                        logPanel.setText("");
                } else if (source == emptyDatabase) {
                        onEmptyDatabase();
                } else if (source == zoomIn) {
                        setZoom(ZOOM_STEP);
                } else if (source == zoomOut) {
                        setZoom(-ZOOM_STEP);
                } else if (source == viewOriginalSize) {
                        zoom = 100;
                        fingerPanel.setZoom(zoom);
                        imagePanel.setZoom(zoom);
                }
        }

        /**
         * Set image panel zoom values
         * */
        private void setZoom(int value) {
                zoom += value;
                if (zoom > ZOOM_MAX) {
                        JOptionPane.showMessageDialog(this, "Maximum zoom level reached!");
                        zoom = ZOOM_MAX;
                }
                if (zoom < ZOOM_MIN) {
                        JOptionPane.showMessageDialog(this, "Minimum zoom level reached!");
                        zoom = ZOOM_MIN;
                }
                fingerPanel.setZoom(zoom);
                imagePanel.setZoom(zoom);
        }

        /**
         * Display properties dialog.
         * */
        private void onProperties() {
                PropertiesDlg dlg = new PropertiesDlg(this);
                dlg.setVisible(true);
                updateSettings();
                imagePanel.invalidate();
                imagePanel.repaint();
        }

        /**
         * Empty fingerprint database
         * */
        private void onEmptyDatabase() {
                int res = JOptionPane.showConfirmDialog(
                                        this,
                                        "Do you want to empty the fingerprint database?",
                                        "VeriFinger Demo",
                                        JOptionPane.YES_NO_OPTION);

                if (res == JOptionPane.OK_OPTION) {
                        fdb.removeAll();
                        listModel.removeAllElements();
                        logEvent("\nFingerprint database is now empty.");
                }
        }

        /**
         * Get images from image source
         * @param count - number of images to retrieve
         * @param id - on return contains image id's, pass <code>null<code> if you don't need image id's
         * */
        private boolean getImages(String title, int count, String[] id, Image[] images) {

                if (Settings.imageSource == Settings.IMAGE_SOURCE_FILE) {
                        StringBuffer msg = new StringBuffer("Select ");
                        msg.append(count);
                        msg.append(count == 1 ? " file" : " files");

                        String[] files = new String[count];
                        if (!Utils.selectImages(this, msg.toString(), files)) return false;
                        if (files == null || files[count-1] == null) {
                                msg = new StringBuffer("You have to select ");
                                msg.append(count);
                                msg.append(count == 1 ? " image!" : " images!");
                                JOptionPane.showMessageDialog(
                                        this,
                                        msg.toString(),
                                        "Error",
                                        JOptionPane.ERROR_MESSAGE);
                                return false;
                        }

                        if (id != null)
                                System.arraycopy(files, 0, id, 0, count);
                        ImageIcon icon = null;
                        msg = new StringBuffer("Loaded: ");

                        for (int i = 0; i < count; ++i) {
                                msg.append(files[i]);
                                msg.append(" ");
                        }
                        logEvent(msg.toString());

                        for (int i = 0; i < count; ++i) {
                                if (files[i].toLowerCase().endsWith(".bmp")) {
                                        BMPLoader bmp = new BMPLoader();
                                        icon = new ImageIcon(bmp.load(files[i]));
                                } else {
                                        icon = new ImageIcon(files[i]);
                                }
                                images[i] = icon.getImage();
                        }
                        Settings.imageDpi = 500;
                } else {
                        // get images from fingerprint scanner
                        FingerprintDlg dlg = null;
                        dlg = new FingerprintDlg(this, title, count, id);
                        try {
                                startCapturing(dlg);
                        }
                        catch (Exception e) {
                                JOptionPane.showMessageDialog(this, "Image capturing failed, check fingerprint scanner!");
                                return false;
                        }
                        dlg.getImages(images);
                        dlg.dispose();

                        try {
                                stopCapturing();
                        } catch (Exception e) {
                                System.out.println(
                                "Error stopping scanner listeners - " + e.getMessage());
                                e.printStackTrace();
                        }
                }

                zoom = 100;
                return true;
        }


        /**
         * Start capturing events from fingerprint scanner.
         * */
        private void startCapturing(ScannerEventListener listener) throws Exception {
                int cnt = ScanMan.SMGetScannerCount();

                for (int i = 0; i < cnt; ++i) {
                        String id = ScanMan.SMGetScannerId(i);
                        ScanMan.SMStartCapturing(id, listener);
                }
        }

        /**
         * Stop capturing events from fingerprint scanner.
         * */
        private void stopCapturing() throws Exception {
                int cnt = ScanMan.SMGetScannerCount();

                for (int i = 0; i < cnt; ++i) {
                        String id = ScanMan.SMGetScannerId(i);
                        ScanMan.SMStopCapturing(id);
                }
        }

        /**
         * Select image for enrollment and add its features to fingerprint database
         * */
        private void onEnrollment() {
                String id[] = new String[1];
                Image[] images = new Image[1];

                if (!getImages("Enrollment", 1, id, images)) {
                        return;
                }

                if (images == null || images.length != 1 || images[0] == null) {
                        logEvent("Failed to get images for enrollment!");
                        return;
                }

                logEvent("\nFingerprint enrollment...");
                Image fingerImage = images[0];
                fingerPanel.setImage(fingerImage, 500.0 / Settings.imageDpi);

                int h = fingerImage.getHeight(null);
                int w = fingerImage.getWidth(null);

                // get byte dump from image
                byte[] imageData = Utils.getImageData(fingerImage, w, h);

                listModel.removeAllElements();

                // extract features
                long start = System.currentTimeMillis();
                byte[] features = new byte[VeriFingerWrapper.VF_MAX_FEATURES_SIZE];
                try {
                        VeriFingerExtractionResult res = VeriFingerWrapper.VFExtract(
                                w, h,
                                imageData,
                                Settings.imageDpi,
                                features, 0);
                } catch (Exception e) {
                        System.out.println(
                                "Extracting features - " + e.getMessage());
                        e.printStackTrace();
                }

                long duration = System.currentTimeMillis() - start;
                logEvent("Features extracted in " + duration + "ms");

                // create image from generalized dump
                Image featuresImage = Utils.createImage(imageData, w, h);

                // display image
                VeriFingerFeatures vf = null;
                try {
                        vf = VeriFingerWrapper.VFDecompressFeatures(features);
                } catch (Exception e) {
                        System.out.println("Error decompressing features - " + e.getMessage());
                        e.printStackTrace();
                }
                imagePanel.setData(vf);
                imagePanel.setImage(featuresImage, 500.0 / Settings.imageDpi);

                view.setEnabled(true);

                fdb.add(new FingerprintData(id[0], features));
        }

        /**
         * Select three fingerprint images for enrollment with generalization,
         * extract their features and generalize them
         * */
        private void onEnrollWithGen() {
                String[] id = new String[GEN_IMAGE_COUNT];
                Image[] images = new Image[GEN_IMAGE_COUNT];

                if (!getImages("Enrollment with generalization", GEN_IMAGE_COUNT, id, images)) {
                        return;
                }

                if (Utils.isNull(images))       {
                        logEvent("Failed to get images for enrollment with generalization!");
                        return;
                }

                logEvent("\nFingerprint enrollment with generalization...");

                if (Settings.imageSource == Settings.IMAGE_SOURCE_FILE) {
                        StringBuffer msg = new StringBuffer("Loaded: ");

                        for (int i = 0; i < GEN_IMAGE_COUNT; ++i) {
                                msg.append(id[i]);
                                msg.append(" ");
                        }

                        logEvent(msg.toString());
                } else logEvent("Loaded: " + id[0]);

                byte[][] imageData = new byte[GEN_IMAGE_COUNT][];
                long start, duration = 0;
                Image fingerImage = null;
                listModel.removeAllElements();

                // extract features from all images
                byte[][] gen_features = new byte[GEN_IMAGE_COUNT][VeriFingerWrapper.VF_MAX_FEATURES_SIZE];
                for (int i = 0; i < GEN_IMAGE_COUNT; ++i) {
                        fingerImage = images[i];
                        int height = fingerImage.getHeight(null);
                        int width = fingerImage.getWidth(null);

                        // get byte dump from image
                        imageData[i] = Utils.getImageData(fingerImage, width, height);

                        // extract features
                        start = System.currentTimeMillis();
                        try {
                                VeriFingerExtractionResult res = VeriFingerWrapper.VFExtract(
                                        width, height,
                                        imageData[i],
                                        Settings.imageDpi,
                                        gen_features[i], 0);
                        } catch (Exception e) {
                                System.out.println(
                                "Error extracting features - " + e.getMessage());
                                e.printStackTrace();
                        }
                        duration = duration + (System.currentTimeMillis() - start);
                }
                logEvent("Features extracted in " + Long.toString(duration) + "ms");

                // get generalized features
                start = System.currentTimeMillis();

                byte[] features = new byte[VeriFingerWrapper.VF_MAX_FEATURES_SIZE];
                VeriFingerGeneralizationResult res = null;
                try {
                        res = VeriFingerWrapper.VFGeneralize(GEN_IMAGE_COUNT, gen_features, features, 0);
                } catch (Exception e) {
                        JOptionPane.showMessageDialog(this, "Generalization failed!");
                        System.out.println(
                                "Error generalizing features - " + e.getMessage());
                        e.printStackTrace();
                        return;
                }
                duration = System.currentTimeMillis() - start;
                logEvent("Features generalized in " + Long.toString(duration) + "ms");

                fingerImage = images[res.indexOfBase];
                int height = fingerImage.getHeight(null);
                int width = fingerImage.getWidth(null);

                fingerPanel.setImage(fingerImage, 500.0 / Settings.imageDpi);

                // create image from generalized dump
                Image featuresImage = Utils.createImage(imageData[res.indexOfBase], width, height);

                VeriFingerFeatures vf = null;
                try {
                        vf = VeriFingerWrapper.VFDecompressFeatures(features);
                } catch (Exception e) {
                        System.out.println(
                                "Error decompressing features - " + e.getMessage());
                                e.printStackTrace();
                }
                // display last image with generalized features
                imagePanel.setData(vf);
                imagePanel.setImage(featuresImage, 500.0 / Settings.imageDpi);
                view.setEnabled(true);

                fdb.add(new FingerprintData(id[0], features));
        }

        /**
         * Select image for recognition and search the fingerprint
         * database for the first/all matching items
         * */
        private void onRecognition(boolean full) {
                // fingerprint db is empty?
                if (fdb.getCount() == 0) {
                        JOptionPane.showMessageDialog(
                                this,
                                "You have to enroll fingerprints first!",
                                "Error",
                                JOptionPane.ERROR_MESSAGE);
                        return;
                }

                Image[] images = new Image[1];

                if (!getImages("Recognition", 1, null, images)) {
                        return;
                }

                if (images == null || images.length != 1 || images[0] == null) {
                        logEvent("Failed to get images for recognition!");
                        return;
                }
                logEvent("\nFingerprint recognition...");
                Image fingerImage = images[0];
                fingerPanel.setImage(fingerImage, 500.0 / Settings.imageDpi);

                int width = fingerImage.getWidth(null);
                int height = fingerImage.getHeight(null);

                // get byte dump from image
                byte[] imageData = Utils.getImageData(fingerImage, width, height);

                // extract features
                byte[] features = new byte[VeriFingerWrapper.VF_MAX_FEATURES_SIZE];
                long start = System.currentTimeMillis();
                try {
                        VeriFingerExtractionResult res = VeriFingerWrapper.VFExtract(
                                width,
                                height,
                                imageData,
                                Settings.imageDpi,
                                features,
                                0);
                } catch (Exception e) {
                        System.out.println(
                                "Error extracting features - " + e.getMessage());
                                e.printStackTrace();
                }
                long duration = System.currentTimeMillis() - start;
                logEvent("Features extracted in " + duration + "ms");

                // create image from generalized dump
                Image featuresImage = Utils.createImage(imageData, width, height);
                // display image
                VeriFingerFeatures vf = null;
                try {
                        vf = VeriFingerWrapper.VFDecompressFeatures(features);
                } catch (Exception e) {
                        System.out.println(
                                "Error decompressing features - " + e.getMessage());
                                e.printStackTrace();
                }
                imagePanel.setData(vf);
                imagePanel.setImage(featuresImage, 500.0 / Settings.imageDpi);

                view.setEnabled(true);

                // search database for matching fingerprint
                String speed = "low";
                if (Settings.matchingSpeed == 1) speed = "high";
                logEvent("Matching speed: " + speed);

                String[] FAR = {"0.1%", "0.01%", "0.001%"};
                logEvent(
                        "False acceptance rate(FAR): " + FAR[Settings.falseAcceptanceRate]);
                logEvent("Maximal rotation: " + Integer.toString(Settings.maximalRotation));
                int matches = 0;
                listModel.removeAllElements();


                listModel.addElement("Title : Similarity");
                duration = 0;
                int i = 0;
                int startG, g, gMin, gMax;
                int range = 1;
                int fCount = 0;
                boolean directionUp = true;
                Vector v = null;

                startG = (int) (features[1] & 0xFF);
                g = startG;
                gMin = startG;
                gMax = startG;

                try {
                        VeriFingerWrapper.VFIdentifyStart(features, 0);
                } catch (Exception e) {
                        System.out.println("Error starting identification - " + e.getMessage());
                        e.printStackTrace();
                        return;
                }

                do {
                        v = fdb.getByG(g);
                        i = 0;

                        while (v != null && i < v.size()) {
                                FingerprintData data = (FingerprintData) v.get(i++);
                                VeriFingerMatchDetails md = null;
                                try {
                                        start = System.currentTimeMillis();
                                        md = VeriFingerWrapper.VFIdentifyNext(data.getFeatures(), false, 0);
                                        duration = duration + (System.currentTimeMillis() - start);
                                        ++fCount;
                                } catch (Exception e) {
                                        System.out.println("Error while matching features - " + e.getMessage());
                                        e.printStackTrace();
                                        return;
                                }

                                int threshold = VeriFingerWrapper.VF_FARMT_MT[Settings.falseAcceptanceRate+1];
                                if (md.similarity > threshold) {
                                        ++matches;
                                        listModel.addElement(data.getTitle() + " : " + md.similarity);
                                        if (!full) break;
                                }
                        }
                        if (!full && matches != 0) break;

                        if (gMin == 0 && gMax == 255) break;
                        else {
                                if (directionUp && gMax == 255) directionUp = false;
                                if (!directionUp && gMin == 0) directionUp = true;
                        }

                        if (directionUp) {
                                ++gMax;
                                g = gMax;
                        } else {
                                --gMin;
                                g = gMin;
                        }
                        directionUp = !directionUp;
                } while (g != -1);

                try {
                        VeriFingerWrapper.VFIdentifyEnd(0);
                } catch (Exception e) {
                        System.out.println("Error ending identification - " + e.getMessage());
                        e.printStackTrace();
                        return;
                }

                logEvent(fCount + " fingerprint(s) processed");
                if (matches != 0)
                        logEvent(Integer.toString(matches) + " match found");
                else
                        logEvent("No matching fingerprint found in database!");
                int mean = (int) (duration / fCount);
                if (mean == 0) mean = 1;
                logEvent("Processing time: " + duration + "ms");
                logEvent("Mean: " + mean + "ms");
                logEvent("Fingers per second: " + 1000 / mean);
        }

        /**
         * Load settings from config file
         * */
        private void loadSettings() {
                try {
                        Settings.load(VFingerJava.CONFIG_FILE);
                } catch (FileNotFoundException e) {
                        Settings.reset();
                        System.out.println("Couldn't load settings - " + e.getMessage());
                        e.printStackTrace();
                }
        }

        /**
         * Save fingerprint database and close main frame
         * */
        private void onClose() {
                try {
                        logEvent("Saving fingerprint database...");
                        fdb.save(Settings.dbFile);
                        Settings.save(CONFIG_FILE);
                } catch (Exception e) {
                        System.out.println(
                                "Error saving fingerprint database - " + e.getMessage());
                        e.printStackTrace();
                }
                dispose();

                // Uninitialize
                try {
                        VeriFingerWrapper.VFFinalize();
                        ScanMan.SMFinalize();
                } catch (VeriFingerException vfe) {
                        System.out.println(
                                "Error terminating VeriFinger - " + vfe.getMessage());
                        vfe.printStackTrace();
                } catch (ScanManException sme) {
                        System.out.println(
                                "Error terminating ScanMan - " + sme.getMessage());
                        sme.printStackTrace();
                } catch (Exception e) {
                        System.out.println(
                                "Error - " + e.getMessage());
                        e.printStackTrace();
                }

                System.out.println("Terminated.");
                System.exit(0);
        }


        /**
         * Initialize VeriFinger
         * */
        private void initVFinger() {
                try {
                        VeriFingerWrapper.VFInitialize();
                        ScanMan.SMInitialize();
                        int protection = VeriFingerWrapper.VFRegistrationType();
                        if (protection == VeriFingerWrapper.VF_RT_UNREGISTERED) {
                                        // get registration key
                                        RegistrationDlg dlg = new RegistrationDlg(this);
                                        VeriFingerWrapper.VFRegister(Settings.regParam, Settings.registrationKey);

                                        protection = VeriFingerWrapper.VFRegistrationType();
                                        if (protection == VeriFingerWrapper.VF_RT_UNREGISTERED) {
                                                JOptionPane.showMessageDialog(
                                                        this,
                                                        "Registration failed!",
                                                        "Error",
                                                        JOptionPane.ERROR_MESSAGE);
                                                onClose();
                                        }
                                        if (protection == VeriFingerWrapper.VF_RT_PC ||
                                                protection == VeriFingerWrapper.VF_RT_UAREU) {
                                                VeriFingerWrapper.VFRegister(Settings.regParam, Settings.registrationKey);
                                        }
                        }

                        updateSettings();

                } catch (Exception e) {
                        System.out.println(e.getMessage());
                        e.printStackTrace();
                }
        }

        private void updateSettings() {
                int speed = VeriFingerWrapper.VF_MATCHING_SPEED_HIGH;
                if (Settings.matchingSpeed == 0) speed = VeriFingerWrapper.VF_MATCHING_SPEED_LOW;

                try {
                        VeriFingerWrapper.setParameterValue(
                                VeriFingerWrapper.VFP_MATCHING_SPEED,
                                Integer.toString(speed), 0);
                        /*
                        VeriFingerWrapper.setParameterValue(
                                VeriFingerWrapper.VF_IMAGE_RESOLUTION,
                                Integer.toString(Settings.imageDpi), 0);
                        */
                        VeriFingerWrapper.setParameterValue(
                                VeriFingerWrapper.VFP_MODE,
                                Integer.toString(Settings.vfMode * 100), 0);

                        VeriFingerWrapper.setParameterValue(
                                VeriFingerWrapper.VFP_MAXIMAL_ROTATION,
                                Integer.toString(Settings.maximalRotation), 0);


                } catch (Exception e) {
                        System.out.println("Error updating settings - " + e.getMessage());
                        e.printStackTrace();
                }
        }

        /**
         * Log event to textarea
         * */
        private void logEvent(String event) {
                logPanel.append(event + "\n");
                logPanel.setCaretPosition(logPanel.getDocument().getLength());
        }

        /**
         * Application entry point
         * */
        public static void main(String args[]) {
                System.out.println("Starting VFingerJava...");
                VFingerJava frame = new VFingerJava();
        }

        /**
         * Just to save few code lines with menu construction
         * */
        private JMenuItem addMenuItem(String title, JMenu menu) {
                JMenuItem item = new JMenuItem(title);
                item.addActionListener(this);
                menu.add(item);

                return item;
        }

        /**
         * Helper function to create the menu and attach it to main frame
         * */
        private void initMenu() {
                JMenuBar menu = new JMenuBar();

                file = new JMenu("File");
                JMenuItem item;

                fileEnrollment = addMenuItem("Enrollment...", file);
                enrollWithGen = addMenuItem("Enrollment with feature generalization...", file);
                fileRecognitionFast = addMenuItem("Recognition (fast) ...", file);
                fileRecognitionFull = addMenuItem("Recognition (full) ...", file);
                file.addSeparator();
                fileAbout = addMenuItem("About", file);
                fileExit = addMenuItem("Exit", file);
                menu.add(file);

                source = new JMenu("Source");

                ButtonGroup group = new ButtonGroup();

                sourceScanner = new JRadioButtonMenuItem("Scanner");
                sourceScanner.addActionListener(this);
                group.add(sourceScanner);
                source.add(sourceScanner);

                sourceFile = new JRadioButtonMenuItem("Image file");
                sourceFile.addActionListener(this);
                group.add(sourceFile);
                source.add(sourceFile);

                menu.add(source);

                service = new JMenu("Service");

                emptyDatabase = addMenuItem("Empty database...", service);
                clearLog = addMenuItem("Clear log window", service);
                serviceProperties = addMenuItem("Properties...", service);

                menu.add(service);

                view = new JMenu("View");
                zoomIn = addMenuItem("Zoom In", view);
                zoomOut = addMenuItem("Zoom out", view);
                viewOriginalSize = addMenuItem("Original size", view);
                menu.add(view);

                view.setEnabled(false);
                setJMenuBar(menu);
        }

        private void initGui() {
                Container content = this.getContentPane();
                content.setLayout(new GridLayout(2, 2));

                final int b = 5;

                // fingerprint display panel
                fingerPanel = new ImagePanel();
                fingerPanel.setBorder(
                        BorderFactory.createCompoundBorder(
                                BorderFactory.createEmptyBorder(b, b, b, b),
                                BorderFactory.createBevelBorder(1)));
                content.add(fingerPanel);

                // fingerprint features display panel
                imagePanel = new FeaturesImagePanel();
                imagePanel.setBorder(
                        BorderFactory.createCompoundBorder(
                                BorderFactory.createEmptyBorder(b, b, b, b),
                                BorderFactory.createBevelBorder(1)));
                content.add(imagePanel);

                // application log area
                logPanel = new JTextArea();
                logPanel.setEditable(false);
                logPanel.setBorder(BorderFactory.createEmptyBorder(b, b, b, b));
                JScrollPane jsp = new JScrollPane(logPanel);
                jsp.setBorder(
                        BorderFactory.createCompoundBorder(
                                BorderFactory.createEmptyBorder(b, b, b, b),
                                BorderFactory.createBevelBorder(1)));

                // matches list
                logPanel2 = new JList(listModel);
                JScrollPane jsp2 = new JScrollPane(logPanel2);
                jsp2.setBorder(
                        BorderFactory.createCompoundBorder(
                                BorderFactory.createEmptyBorder(b, b, b, b),
                                BorderFactory.createBevelBorder(1)));
                content.add(jsp);
                content.add(jsp2);
        }

        // components
        private ImagePanel fingerPanel;
        private FeaturesImagePanel imagePanel;
        private JTextArea logPanel;
        private JList logPanel2;
        private JMenu view;
        private JMenu file;
        private JMenu source;
        private JMenu service;
        private JMenuItem fileAbout;
        private JMenuItem emptyDatabase;
        private JMenuItem fileExit;
        private JMenuItem fileRecognitionFast;
        private JMenuItem fileRecognitionFull;
        private JMenuItem fileEnrollment;
        private JMenuItem serviceProperties;
        private JMenuItem enrollWithGen;
        private JMenuItem clearLog;
        private JMenuItem zoomIn;
        private JMenuItem zoomOut;
        private JMenuItem viewOriginalSize;
        private JRadioButtonMenuItem sourceScanner;
        private JRadioButtonMenuItem sourceFile;
}
