/**
 * Database used to store fingerprint features
 *
 */

package VeriFinger;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.List;
import java.util.Vector;

public class FingerDatabase {
        private static final int NUM_G = 256;
        private Object[] data = new Object[NUM_G];

        /**
         * Constructor
         * */
        public FingerDatabase() {

        }

        /**
         * Add fingerpint features to database
         * */
        public void add(FingerprintData fd) {
                int g = fd.getG();
                if      (data[g] == null) {
                        Vector v = new Vector();
                        v.add(fd);
                        data[g] = v;
                }
                else {
                        Vector v = (Vector) data[g];
                        if (v.indexOf(fd) != -1) v.remove(fd);
                        v.add(fd);
                }
        }

        /**
         * Remove all entries from database
         * */
        public void removeAll() {
                for (int i = 0; i < NUM_G; ++i) {
                        if (data[i] != null) {
                                Vector v = (Vector) data[i];
                                v.clear();
                        }
                }
        }

        /**
         * Load the database from disk
         * @param fileName database file
         * */
        public void load(String fileName) {
                ObjectInputStream in = null;
                Object[] tmp = new Object[NUM_G];
                try {
                        FileInputStream fin = new FileInputStream(fileName);
                        in = new ObjectInputStream(fin);
                        for (int i = 0; i < NUM_G; ++i) {
                                Object o = in.readObject();
                                if (o != null) {
                                        Vector v = (Vector) o;
                                        tmp[i] = v;
                                }
                        }
                        in.close();
                        removeAll();
                        data = tmp;
                }
                catch (Exception e) {
                        System.out.println("Error loading database - " + e.getMessage());
                        e.printStackTrace();
                }
        }

        /**
         * Save the database to disk
         * @param fileName file to use for database serialization
         * */
        public void save(String fileName) throws IOException {
                ObjectOutputStream out = null;
                try {
                        FileOutputStream fout = new FileOutputStream(fileName, false);
                        out = new ObjectOutputStream(fout);
                        for (int i = 0; i < NUM_G; ++i) {
                                out.writeObject(data[i]);
                        }
                        out.close();
                }
                catch (FileNotFoundException e) {
                        System.out.println(e.getMessage());
                        e.printStackTrace();
                }
        }

        /**
         * Gets number of fingerprints in database
         * @return number of fingerprint features in database
         * */
        public int getCount() {
                int result = 0;
                for (int i = 0; i < NUM_G; ++i) {
                        if (data[i] != null) {
                                Vector v = (Vector) data[i];
                                result += v.size();
                        }
                }

                return result;
        }


        public Vector getByG(int g) {
                Vector result = null;
                if (data[g] != null) {
                        result = (Vector) data[g];
                }
                return result;
        }
}

