package VeriFinger;


public class VeriFingerException extends Exception {

    protected int errorCode;

    public VeriFingerException(int errorcode, String message)
    {
        super(message);
        errorCode = errorcode;
    }

    public int getErrorCode()
    {
        return errorCode;
    }

    public static boolean failed(int errorCode)
    {
        return ((errorCode) < 0);
    }

    public static boolean succeeded(int errorCode)
    {
        return ((errorCode) >= 0);
    }

    // VeriFinger Error Codes:

    // General
    public static final int VFE_OK                              =  0;
    public static final int VFE_FAILED                          = -1;
    public static final int VFE_OUT_OF_MEMORY                   = -2;
    public static final int VFE_NOT_INITIALIZED                 = -3;
    public static final int VFE_ARGUMENT_NULL                   = -4;
    public static final int VFE_INVALID_ARGUMENT                = -5;
    // Registration
    public static final int VFE_NOT_REGISTERED                  = -2000;
    public static final int VFE_INVALID_SERIAL_NUMBER           = -2001;
    public static final int VFE_INVALID_REGISTRATION_KEY        = -2002;
    public static final int VFE_SCANNER_DRIVER_ERROR            = -2003;
    public static final int VFE_REGISTRATION_NOT_NEEDED         = -2004;
    public static final int VFE_NO_SCANNER                      = -2005;
    public static final int VFE_MORE_THAN_ONE_SCANNER           = -2006;
    public static final int VFE_LM_CONNECTION_ERROR             = -2007;
    public static final int VFE_LM_NO_MORE_LICENSES             = -2008;
    // Parameters
    public static final int VFE_INVALID_PARAMETER               = -10;
    public static final int VFE_PARAMETER_READ_ONLY             = -11;
    // Features extraction
    public static final int VFE_ILLEGAL_IMAGE_RESOLUTION        = -101;
    public static final int VFE_ILLEGAL_IMAGE_SIZE              = -102;
    public static final int VFE_LOW_QUALITY_IMAGE               = -103;
    // VeriFinger specific
    public static final int VFE_INVALID_MODE                    = -1000;

    public static String VFErrorToString(int error)
    {
	    if (error >= 0)
	        return "";
	    switch(error)
	    {
		    case VFE_FAILED: return "Failed";
		    case VFE_OUT_OF_MEMORY: return "Out of memory";
		    case VFE_NOT_INITIALIZED: return "Not initialized";
		    case VFE_ARGUMENT_NULL: return "Argument null";
		    case VFE_INVALID_ARGUMENT: return "Invalid argument";
		    case VFE_NOT_REGISTERED: return "Not registered";
		    case VFE_INVALID_SERIAL_NUMBER: return "Invalid serial number";
		    case VFE_INVALID_REGISTRATION_KEY: return "Invalid registration key";
		    case VFE_SCANNER_DRIVER_ERROR: return "Scanner driver error";
		    case VFE_REGISTRATION_NOT_NEEDED: return "Registration not needed";
		    case VFE_NO_SCANNER: return "No scanner connected";
		    case VFE_MORE_THAN_ONE_SCANNER: return "More than one scanner connected";
		    case VFE_LM_CONNECTION_ERROR: return "Communication with License Manager error";
		    case VFE_LM_NO_MORE_LICENSES: return "License Manager error: no more licenses";
		    case VFE_INVALID_PARAMETER: return "Invalid parameter";
		    case VFE_PARAMETER_READ_ONLY: return "Parameter is read only";
		    case VFE_ILLEGAL_IMAGE_RESOLUTION: return "Illegal image resolution";
		    case VFE_ILLEGAL_IMAGE_SIZE: return "Illegal image size";
		    case VFE_LOW_QUALITY_IMAGE: return "Low quality image";
		    case VFE_INVALID_MODE: return "Invalid mode";
		    default: return Integer.toString(error);
	    }
    }

    public String VFResultToString(int result)
    {
	    if(failed(result))
	        return VFErrorToString(result);
	    else
	        return result == VFE_OK ? "OK" : "OK (" + Integer.toString(result) + ")";
    }
}
