unit FormOptions;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, VFOptions, FormFileListEntry, Utils, VFinger;

type
  TOptionsForm = class(TForm)
    LabelSM: TLabel;
    ComboBoxStartupMode: TComboBox;
    LabelVFM: TLabel;
    ComboBoxVeriFingerMode: TComboBox;
    CheckBoxUseLowQualityImages: TCheckBox;
    GroupBoxFile: TGroupBox;
    GroupBoxScanner: TGroupBox;
    GroupBoxEnrollment: TGroupBox;
    GroupBoxIdentification: TGroupBox;
    GroupBoxExtraction: TGroupBox;
    GroupBoxGeneralization: TGroupBox;
    GroupBoxMatching: TGroupBox;
    ButtonOK: TButton;
    ButtonCancel: TButton;
    ButtonDefault: TButton;
    LabelFR: TLabel;
    LabelFileResolution: TLabel;
    LabelFDPI: TLabel;
    LabelFFID: TLabel;
    LabelFileFingerID: TLabel;
    CheckBoxFilePrompt: TCheckBox;
    ButtonFileEdit: TButton;
    LabelSFID: TLabel;
    LabelScannerFingerID: TLabel;
    CheckBoxScannerPrompt: TCheckBox;
    ButtonScannerEdit: TButton;
    EditMinMinutiaCount: TEdit;
    CheckBoxSearchForDuplicates: TCheckBox;
    CheckBoxWarnDifferentFingerID: TCheckBox;
    CheckBoxUseG: TCheckBox;
    LabelDatabasePercent1: TLabel;
    EditDatabasePercent: TEdit;
    LabelDatabasePercent2: TLabel;
    CheckBoxMatchAll: TCheckBox;
    CheckBoxShowAllResults: TCheckBox;
    LabelGenFAR: TLabel;
    ComboBoxGenFARate: TComboBox;
    LabelFAR: TLabel;
    ComboBoxFARate: TComboBox;
    LabelMS: TLabel;
    TrackBarMatchingSpeed: TTrackBar;
    LabelTBMSFB: TLabel;
    LabelTBMSLB: TLabel;
    LabelMR: TLabel;
    EditMaximalRotation: TEdit;
    LabelDeg: TLabel;
    CheckBoxMinMinutiaCount: TCheckBox;
    LabelReturnedImage: TLabel;
    ComboBoxReturnedImage: TComboBox;
    LabelGenMaximalRotation: TLabel;
    EditGenMaximalRotation: TEdit;
    LabelGenDeg: TLabel;
    procedure ButtonDefaultClick(Sender: TObject);
    procedure ButtonFileEditClick(Sender: TObject);
    procedure ButtonScannerEditClick(Sender: TObject);
    procedure CheckBoxSearchForDuplicatesClick(Sender: TObject);
    procedure CheckBoxUseGClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure CheckBoxMinMinutiaCountClick(Sender: TObject);
  private
    procedure SetDefaultOptions;
    procedure EditInfo(LabelFingerID, LabelResolution: TLabel; FromScanner: Boolean);
    procedure UpdateControls;
  public
    procedure SetOptions(var Options: TVFOptions);
    procedure GetOptions(var Options: TVFOptions);
  end;

function OptionsBox(Owner: TForm; var Options: TVFOptions): Boolean;

implementation

{$R *.dfm}

function OptionsBox(Owner: TForm; var Options: TVFOptions): Boolean;
var
  Form: TOptionsForm;
begin
  Form := TOptionsForm.Create(Owner);
  Form.SetOptions(Options);
  Result := Form.ShowModal = mrOk;
  if Result then
  begin
    Form.GetOptions(Options);
  end;
  Form.Free;
end;

{ TOptionsForm }

procedure TOptionsForm.GetOptions(var Options: TVFOptions);
begin
  Options.StartupMode := TVFDemoMode(ComboBoxStartupMode.ItemIndex);

  Options.VeriFingerMode := Integer(ComboBoxVeriFingerMode.Items.Objects[ComboBoxVeriFingerMode.ItemIndex]);

  Options.UseLowQualityImages := CheckBoxUseLowQualityImages.Checked;

  Options.FileResolution := StrToInt(LabelFileResolution.Caption);
  Options.FileFingerID := LabelFileFingerID.Caption;
  Options.FilePrompt := CheckBoxFilePrompt.Checked;

  Options.ScannerFingerID := LabelScannerFingerID.Caption;
  Options.ScannerPrompt := CheckBoxScannerPrompt.Checked;

  Options.CheckMinMinutiaCount := CheckBoxMinMinutiaCount.Checked;
  if Options.CheckMinMinutiaCount then Options.MinMinutiaCount := StrToInt(EditMinMinutiaCount.Text);
  Options.SearchForDuplicates := CheckBoxSearchForDuplicates.Checked;
  if Options.SearchForDuplicates then Options.WarnDifferentFingerID := CheckBoxWarnDifferentFingerID.Checked;

  Options.UseG := CheckBoxUseG.Checked;
  if Options.UseG then Options.DatabasePercent := StrToInt(EditDatabasePercent.Text);
  Options.MatchAll := CheckBoxMatchAll.Checked;
  Options.ShowAllResults := CheckBoxShowAllResults.Checked;

  Options.ReturnedImage := ComboBoxReturnedImage.ItemIndex;

  Options.GenFARate := StringToFloat(ComboBoxGenFARate.Text, '%');
  Options.GenMaximalRotation := VFDegToDir(StrToInt(EditGenMaximalRotation.Text));

  Options.FARate := StringToFloat(ComboBoxFARate.Text, '%');
  if TrackBarMatchingSpeed.Position = 1 then Options.MatchingSpeed := VF_MATCHING_SPEED_HIGH
  else Options.MatchingSpeed := VF_MATCHING_SPEED_LOW;
  Options.MaximalRotation := VFDegToDir(StrToInt(EditMaximalRotation.Text));
end;

procedure TOptionsForm.SetOptions(var Options: TVFOptions);
begin
  ComboBoxStartupMode.ItemIndex := Ord(Options.StartupMode);

  ComboBoxVeriFingerMode.ItemIndex := ComboBoxVeriFingerMode.Items.IndexOfObject(TObject(Options.VeriFingerMode));

  CheckBoxUseLowQualityImages.Checked := Options.UseLowQualityImages;

  LabelFileResolution.Caption := IntToStr(Options.FileResolution);
  LabelFileFingerID.Caption := Options.FileFingerID;
  CheckBoxFilePrompt.Checked := Options.FilePrompt;

  LabelScannerFingerID.Caption := Options.ScannerFingerID;
  CheckBoxScannerPrompt.Checked := Options.ScannerPrompt;

  CheckBoxMinMinutiaCount.Checked := Options.CheckMinMinutiaCount;
  EditMinMinutiaCount.Text := IntToStr(Options.MinMinutiaCount);
  CheckBoxSearchForDuplicates.Checked := Options.SearchForDuplicates;
  CheckBoxWarnDifferentFingerID.Checked := Options.WarnDifferentFingerID;

  CheckBoxUseG.Checked := Options.UseG;
  EditDatabasePercent.Text := IntToStr(Options.DatabasePercent);
  CheckBoxMatchAll.Checked := Options.MatchAll;
  CheckBoxShowAllResults.Checked := Options.ShowAllResults;

  ComboBoxReturnedImage.ItemIndex := Options.ReturnedImage;

  ComboBoxGenFARate.Text := FloatToStr(Options.GenFARate) + '%';
  EditGenMaximalRotation.Text := IntToStr(VFDirToDeg(Options.GenMaximalRotation));

  ComboBoxFARate.Text := FloatToStr(Options.FARate) + '%';
  if Options.MatchingSpeed = VF_MATCHING_SPEED_HIGH then
    TrackBarMatchingSpeed.Position := 1
  else TrackBarMatchingSpeed.Position := 0;
  EditMaximalRotation.Text := IntToStr(VFDirToDeg(Options.MaximalRotation));

  UpdateControls;
end;

procedure TOptionsForm.ButtonDefaultClick(Sender: TObject);
begin
  SetDefaultOptions;
end;

procedure TOptionsForm.ButtonFileEditClick(Sender: TObject);
begin
  EditInfo(LabelFileFingerID, LabelFileResolution, False);
end;

procedure TOptionsForm.ButtonScannerEditClick(Sender: TObject);
begin
  EditInfo(LabelScannerFingerID, nil, True);
end;

procedure TOptionsForm.CheckBoxSearchForDuplicatesClick(Sender: TObject);
begin
  UpdateControls;
end;

procedure TOptionsForm.CheckBoxUseGClick(Sender: TObject);
begin
  UpdateControls;
end;

procedure TOptionsForm.EditInfo(LabelFingerID, LabelResolution: TLabel;
  FromScanner: Boolean);
var
  Flags: LongWord;
  Source: string;
  Resolution: Integer;
  FingerID: string;
begin
  Flags := FLEB_NEW_ENTRY or FLEB_MANY_ENTRIES;
  FingerID := LabelFingerID.Caption;
  if FromScanner then
  begin
    Source := 'Name\ID';
    SetFlag(Flags, FLEB_FROM_SCANNER);
    Resolution := -1;
  end
  else begin
    Source := 'C:\Dir1\Dir2\filename.ext';
    Resolution := StrToInt(LabelResolution.Caption);
  end;
  if FileListEntryBox(Self, Flags, Source, Resolution, FingerID) then
  begin
    LabelFingerID.Caption := FingerID;
    if not FromScanner then LabelResolution.Caption := IntToStr(Resolution);
  end;
end;

procedure TOptionsForm.SetDefaultOptions;
begin
  SetOptions(DefaultOptions);
end;

procedure TOptionsForm.UpdateControls;
var
  UseDP: Boolean;
begin
  EditMinMinutiaCount.Enabled := CheckBoxMinMinutiaCount.Checked;
  CheckBoxWarnDifferentFingerID.Enabled := CheckBoxSearchForDuplicates.Checked;
  UseDP := CheckBoxUseG.Checked;
  LabelDatabasePercent1.Enabled := UseDP;
  EditDatabasePercent.Enabled := UseDP;
  LabelDatabasePercent2.Enabled := UseDP;
end;

procedure TOptionsForm.FormCreate(Sender: TObject);
begin
  with ComboBoxStartupMode.Items do
  begin
    Add(VFDemoModeToString(vfdmEnrollment));
    Add(VFDemoModeToString(vfdmEnrollmentWithGen));
    Add(VFDemoModeToString(vfdmVerification));
    Add(VFDemoModeToString(vfdmIdentification));
  end;
  with ComboBoxVeriFingerMode.Items do
  begin
    AddObject('General', TObject(VF_MODE_GENERAL));
    AddObject('DigitalPersona U.are.U', TObject(VF_MODE_DIGITALPERSONA_UAREU));
    AddObject('BiometriKa FX2000', TObject(VF_MODE_BIOMETRIKA_FX2000));
    AddObject('Keytronic SecureDesktop', TObject(VF_MODE_KEYTRONIC_SECUREDESKTOP));
    AddObject('Identix TouchView', TObject(VF_MODE_IDENTIX_TOUCHVIEW));
    AddObject('PreciseBiometrics 100CS', TObject(VF_MODE_PRECISEBIOMETRICS_100CS));
    AddObject('STMicroelectronics TouchChip', TObject(VF_MODE_STMICROELECTRONICS_TOUCHCHIP));
    AddObject('IdenticatorTechnology DF90', TObject(VF_MODE_IDENTICATORTECHNOLOGY_DF90));
    AddObject('Authentec AFS2', TObject(VF_MODE_AUTHENTEC_AFS2));
    AddObject('Authentec AES4000', TObject(VF_MODE_AUTHENTEC_AES4000));
    AddObject('Atmel FingerChip', TObject(VF_MODE_ATMEL_FINGERCHIP));
    AddObject('BMF BLP100', TObject(VF_MODE_BMF_BLP100));
    AddObject('Secugen Hamster', TObject(VF_MODE_SECUGEN_HAMSTER));
    AddObject('Ethentica', TObject(VF_MODE_ETHENTICA));
    AddObject('CrossMatch Verifier 300', TObject(VF_MODE_CROSSMATCH_VERIFIER300));
  end;
end;

procedure TOptionsForm.FormCloseQuery(Sender: TObject;
  var CanClose: Boolean);
var
  I: Integer;
begin
  if ModalResult = mrOk then
  begin
    try
      if CheckBoxMinMinutiaCount.Checked then
      begin
        try
          I := StrToInt(EditMinMinutiaCount.Text);
        except
          EditMinMinutiaCount.SetFocus;
          raise Exception.Create('Minimal minutia count have to be an integer value');
        end;
        if I < 0 then
        begin
          EditMinMinutiaCount.SetFocus;
          raise Exception.Create('Minimal minutia count have to be not less than zero');
        end;
      end;
      if CheckBoxUseG.Checked then
      begin
        try
          I := StrToInt(EditDatabasePercent.Text);
        except
          EditDatabasePercent.SetFocus;
          raise Exception.Create('Database percent have to be an integer value');
        end;
        if (I < 0) or (I > 100) then
        begin
          EditDatabasePercent.SetFocus;
          raise Exception.Create('Database percent have to be in range from 0 to 100');
        end;
      end;
      try
        StringToFloat(ComboBoxGenFARate.Text, '%');
      except
        ComboBoxGenFARate.SetFocus;
        raise Exception.Create('Generalization FAR have to be a float value');
      end;
      try
        StringToFloat(ComboBoxFARate.Text, '%');
      except
        ComboBoxFARate.SetFocus;
        raise Exception.Create('FAR have to be a float value');
      end;
      try
        I := StrToInt(EditGenMaximalRotation.Text);
      except
        EditGenMaximalRotation.SetFocus;
        raise Exception.Create('Generalization maximal rotation have to be an integer value');
      end;
      if (I < 0) or (I > 180) then
      begin
        EditGenMaximalRotation.SetFocus;
        raise Exception.Create('Generalization maximal rotation have to be in range from 0 to 180');
      end;
      try
        I := StrToInt(EditMaximalRotation.Text);
      except
        EditMaximalRotation.SetFocus;
        raise Exception.Create('Maximal rotation have to be an integer value');
      end;
      if (I < 0) or (I > 180) then
      begin
        EditMaximalRotation.SetFocus;
        raise Exception.Create('Maximal rotation have to be in range from 0 to 180');
      end;
      CanClose := True;
    except
      on E: Exception do
      begin
        MessageDlg(E.Message, mtError, [mbOk], 0);
        CanClose := False;
      end;
    end;
  end
  else CanClose := True;
end;

procedure TOptionsForm.CheckBoxMinMinutiaCountClick(Sender: TObject);
begin
  UpdateControls;
end;

end.
