unit VFDatabase;

interface

uses SysUtils, Classes, Windows, DB, ADODB, VFinger;

const
  SOHEAP_MAX_BLOCK_COUNT = 1024;

  FINGER_ID_LENGTH = 50;

type

  TSOHeap = class
  private
    BlockSize: LongWord;
    BlockCount: Integer;
    Blocks: array[0..SOHEAP_MAX_BLOCK_COUNT - 1] of PByte;
    Sizes: array[0..SOHEAP_MAX_BLOCK_COUNT - 1] of LongWord;
  public
    constructor Create(BlockSize: LongWord = 1048576);
    destructor Destroy; override;
    procedure Clear;
    function Alloc(Size: Word): PByte;
    procedure Free(Obj: PByte);
  end;

  PVFDBItem = ^TVFDBItem;
  TVFDBItem = record
    Index: Integer;
    GIndex: Integer;
    ID: Integer;
    Features: PByte;
    FeaturesSize: LongWord;
    FingerIDLength: LongWord;
    FingerID: array[0..FINGER_ID_LENGTH] of Char;
  end;

  IVFDBOpenCallback = interface
    procedure OnOpenState(State: string);
    function OnOpenProgress(Started: Boolean; Value: Integer): Boolean;
  end;

  TVFDBFindData = record
    TestG: Integer;
    G: Integer;
    Index: Integer;
  end;

  TVFDatabase = class
  private
    Heap: TSOHeap;
    FItems: array[0..255] of array of PVFDBItem;
    FAllItems: array of PVFDBItem;
    FCapacity: array[0..255] of Integer;
    FAllCapacity: Integer;
    FCount: array[0..255] of Integer;
    FAllCount: Integer;
    Connection: TADOConnection;
    FConnected: Boolean;
    Recordset: TADOTable;
    FOpened: Boolean;
    RecordsetName: string;

    function GetFeatures(Index: Integer): PByte;
    function GetFeaturesSize(Index: Integer): LongWord;
    function GetFingerID(Index: Integer): string;
    function GetID(Index: Integer): Integer;
    procedure SetFingerID(Index: Integer; const Value: string);

    procedure OpenRecordset(CursorType: TCursorType = ctKeyset;
      LockType: TADOLockType = ltOptimistic; CursorLocation: TCursorLocation = clUseServer);
    procedure CloseRecordset;

    class procedure CopyFingerID(Dst: PChar; var DstLength: LongWord; Src: PChar; SrcLength: LongWord);

    procedure ClearItems;
    function AddItem(ID: Integer; Features: PByte; Size: LongWord; FingerID: string): Integer;

    procedure CheckIndex(Index: Integer);

    function FindRecord(ID: Integer): Boolean;
    procedure UpdateRecord(Index: Integer; IsNew: Boolean = False);

  public
    constructor Create;
    destructor Destroy; override;

    procedure Open(Provider, DataSource, UserID, Password: string;
      Callback: IVFDBOpenCallback = nil);
    procedure Close;
    property Opened: Boolean read FOpened;

    property Count: Integer read FAllCount;
    property ID[Index: Integer]: Integer read GetID;
    property Features[Index: Integer]: PByte read GetFeatures;
    property FeaturesSize[Index: Integer]: LongWord read GetFeaturesSize;
    property FingerID[Index: Integer]: string read GetFingerID write SetFingerID;
    procedure SetFeatures(Index: Integer; Features: PByte; Size: LongWord);
    function Add(Features: PByte; Size: LongWord; FingerID: string): Integer;
    procedure Remove(Index: Integer);
    procedure Clear;

    procedure FindStart(Features: PByte; var FD: TVFDBFindData);
    function FindNext(var FD: TVFDBFindData): Integer;
    procedure FindEnd(var FD: TVFDBFindData);
  end;

implementation

{ TVFDatabase }

function VFDBCapacity(C: Integer): Integer;
begin
  if C <= 2 then Result := C
  else if C <=   4 then Result :=   4
  else if C <=   8 then Result :=   8
  else if C <=  16 then Result :=  16
  else if C <=  32 then Result :=  32
  else if C <=  64 then Result :=  64
  else if C <= 128 then Result := 128
  else if C <= 256 then Result := 256
  else if C <= 512 then Result := 512
  else begin
    Result := C div 1024;
    if C mod 1024 <> 0 then Inc(Result);
    Result := Result * 1024;
  end;
end;

function VFDBAllCapacity(C: Integer): Integer;
begin
  if C <= 2 then Result := C
  else if C <=     4 then Result :=     4
  else if C <=     8 then Result :=     8
  else if C <=    16 then Result :=    16
  else if C <=    32 then Result :=    32
  else if C <=    64 then Result :=    64
  else if C <=   128 then Result :=   128
  else if C <=   256 then Result :=   256
  else if C <=   512 then Result :=   512
  else if C <=  1024 then Result :=  1024
  else if C <=  2048 then Result :=  2048
  else if C <=  4096 then Result :=  4096
  else if C <=  8192 then Result :=  8192
  else if C <= 16384 then Result := 16384
  else if C <= 32768 then Result := 32768
  else if C <= 65536 then Result := 65536
  else begin
    Result := C div 131072;
    if C mod 131072 <> 0 then Inc(Result);
    Result := Result * 131072;
  end;
end;

function VFDBIncCapacityDelta(CP: Integer): Integer;
begin
  if CP < 2 then Result := 1
  else if CP =   2 then Result :=   2
  else if CP =   4 then Result :=   4
  else if CP =   8 then Result :=   8
  else if CP =  16 then Result :=  16
  else if CP =  32 then Result :=  32
  else if CP =  64 then Result :=  64
  else if CP = 128 then Result := 128
  else if CP = 256 then Result := 256
  else if CP = 512 then Result := 512
  else Result := 1024;
end;

function VFDBIncAllCapacityDelta(CP: Integer): Integer;
begin
  if CP < 2 then Result := 1
  else if CP =     2 then Result :=     2
  else if CP =     4 then Result :=     4
  else if CP =     8 then Result :=     8
  else if CP =    16 then Result :=    16
  else if CP =    32 then Result :=    32
  else if CP =    64 then Result :=    64
  else if CP =   128 then Result :=   128
  else if CP =   256 then Result :=   256
  else if CP =   512 then Result :=   512
  else if CP =  1024 then Result :=  1024
  else if CP =  2048 then Result :=  2048
  else if CP =  4098 then Result :=  4096
  else if CP =  8192 then Result :=  8192
  else if CP = 16384 then Result := 16384
  else if CP = 32768 then Result := 32768
  else if CP = 65536 then Result := 65536
  else Result := 131072;
end;

function TVFDatabase.Add(Features: PByte; Size: LongWord;
  FingerID: string): Integer;
begin
  Result := AddItem(0, Features, Size, FingerID);
  if FOpened then
  begin
    try
      UpdateRecord(Result, True);
      FAllItems[Result].ID := Recordset.Fields[0].AsInteger;
    except
      on E: Exception do
        raise Exception.Create('Can not add record to the database'#13 + E.Message);
    end;
  end;
end;

function TVFDatabase.AddItem(ID: Integer; Features: PByte; Size: LongWord;
  FingerID: string): Integer;
var
  G: Integer;
  C: Integer;
  CP: Integer;
  Item: PVFDBItem;
begin
  G := VFFeatGetG(Features); VFCheckResult(G);
  C := FCount[G];
  CP := FCapacity[G];
  if C = CP then
  begin
    Inc(CP, VFDBIncCapacityDelta(CP));
    SetLength(FItems[G], CP);
    FCapacity[G] := CP;
  end;
  if FAllCount = FAllCapacity then
  begin
    Inc(FAllCapacity, VFDBIncAllCapacityDelta(FAllCapacity));
    SetLength(FAllItems, FAllCapacity);
  end;
  Item := PVFDBItem(Heap.Alloc(SizeOf(TVFDBItem)));
  Item.Index := FAllCount;
  Item.GIndex := C;
  Item.ID := ID;
  Item.Features := Heap.Alloc(Size);
  Item.FeaturesSize := Size;
  Move(Features^, Item.Features^, Size);
  CopyFingerID(Item.FingerID, Item.FingerIDLength, PChar(FingerID), Length(FingerID));
  FItems[G][C] := Item;
  FAllItems[FAllCount] := Item;
  Inc(C);
  FCount[G] := C;
  Result := FAllCount;
  Inc(FAllCount);
end;

procedure TVFDatabase.CheckIndex(Index: Integer);
begin
  if (Index < 0) or (Index >= FAllCount) then
    raise Exception.Create('Database record index out of bounds');
end;

procedure TVFDatabase.Clear;
begin
  ClearItems;
  if FOpened then
  begin
    try
      CloseRecordset;
      Connection.Execute(WideString(Format('DELETE FROM %s', [RecordsetName])),
        cmdText, [eoExecuteNoRecords]);
      OpenRecordset;
    except
      on E: Exception do
        raise Exception.Create('Can not clear database'#13 + E.Message);
    end;
  end;
end;

procedure TVFDatabase.ClearItems;
var
  G: Integer;
begin
  Heap.Clear;
  for G := 0 to 255 do
  begin
    FItems[G] := nil;
    FCapacity[G] := 0;
    FCount[G] := 0;
  end;
  FAllItems := nil;
  FAllCount := 0;
  FAllCapacity := 0;
end;

procedure TVFDatabase.Close;
begin
  ClearItems;
  try
    CloseRecordset;
    Recordset.Free;
    Recordset := nil;
    if FConnected then
    begin
      Connection.Close;
      FConnected := False;
    end;
    Connection.Free;
    Connection := nil;
  except
    on E: Exception do
      raise Exception.Create('Can not close database'#13 + E.Message);
  end;
end;

procedure TVFDatabase.CloseRecordset;
begin
  if FOpened then
  begin
    Recordset.Close;
    FOpened := False;
  end;
end;

class procedure TVFDatabase.CopyFingerID(Dst: PChar; var DstLength: LongWord;
  Src: PChar; SrcLength: LongWord);
begin
  if SrcLength > FINGER_ID_LENGTH then SrcLength := FINGER_ID_LENGTH;
  DstLength := SrcLength;
  StrLCopy(Dst, Src, SrcLength);
end;

constructor TVFDatabase.Create;
var
  G: Integer;
begin
  Heap := TSOHeap.Create;
  for G := 0 to 255 do
  begin
    FItems[G] := nil;
    FCapacity[G] := 0;
    FCount[G] := 0;
  end;
  FAllItems := nil;
  FAllCount := 0;
  FAllCapacity := 0;
  FConnected := False;
  FOpened := False;
  RecordsetName := 'Fingerprints';
end;

destructor TVFDatabase.Destroy;
begin
  Close;
  TObject(Heap).Free;
  inherited;
end;

procedure TVFDatabase.FindEnd(var FD: TVFDBFindData);
begin

end;

function TVFDatabase.FindNext(var FD: TVFDBFindData): Integer;
var
  GDif: Integer;
begin
  Inc(FD.Index);
  while FD.Index >= FCount[FD.G] do
  begin
    FD.Index := 0;
    GDif := FD.G - FD.TestG;
    if GDif = 0 then FD.G := 1 else if GDif > 0 then FD.G := -GDif else FD.G := -GDif + 1;
    Inc(FD.G, FD.TestG);
    GDif := FD.G - FD.TestG;
    if FD.G >= 255 then
    begin
      FD.G := FD.TestG - GDif;
      if FD.G < 0 then
      begin
        Result := -1;
        Exit;
      end;
    end
    else if FD.G < 0 then
    begin
      FD.G := FD.TestG - GDif + 1;
      if FD.G >= 255 then
      begin
        Result := -1;
        Exit;
      end;
    end;
  end;
  Result := FItems[FD.G][FD.Index].Index;
end;

function TVFDatabase.FindRecord(ID: Integer): Boolean;
begin
  Result := Recordset.Locate('ID', ID, []);
end;

procedure TVFDatabase.FindStart(Features: PByte; var FD: TVFDBFindData);
begin
  FD.TestG := VFFeatGetG(Features); VFCheckResult(FD.TestG);
  FD.G := FD.TestG;
  FD.Index := -1;
end;

function TVFDatabase.GetFeatures(Index: Integer): PByte;
begin
  CheckIndex(Index);
  Result := FAllItems[Index].Features;
end;

function TVFDatabase.GetFeaturesSize(Index: Integer): LongWord;
begin
  CheckIndex(Index);
  Result := FAllItems[Index].FeaturesSize;
end;

function TVFDatabase.GetFingerID(Index: Integer): string;
begin
  CheckIndex(Index);
  Result := FAllItems[Index].FingerID;
end;

function TVFDatabase.GetID(Index: Integer): Integer;
begin
  CheckIndex(Index);
  Result := FAllItems[Index].ID;
end;

procedure TVFDatabase.Open(Provider, DataSource, UserID, Password: string;
  Callback: IVFDBOpenCallback);
var
  R: _RecordSet;
  RC: Integer;
  DT: LongWord;
  Features: array[0..VF_MAX_FEATURES_SIZE - 1] of Byte;
  Stream: TStream;
  Size: LongWord;
begin
  Close;
  try
    Connection := TADOConnection.Create(nil);
    Recordset := TADOTable.Create(nil);
    Recordset.Connection := Connection;
    Recordset.TableName := RecordsetName;
    Recordset.TableDirect := True;

    if Assigned(Callback) then Callback.OnOpenState('Opening connection');
    Connection.ConnectionString := Format('Provider=%s;Data Source=%s;', [Provider, DataSource]);
    Connection.Open(UserID, Password);
    FConnected := True;

    if Assigned(Callback) then Callback.OnOpenState('Retrieving record count');
    R := Connection.Execute(WideString('SELECT COUNT(*) AS RC FROM ' + RecordsetName));
    RC := R.Fields.Item['RC'].Value;

    if Assigned(Callback) then Callback.OnOpenState('Allocating memory');
    FAllCapacity := VFDBAllCapacity(RC);
    SetLength(FAllItems, FAllCapacity);

    if Assigned(Callback) then Callback.OnOpenState('Opening recordset in forward only mode');
    OpenRecordset(ctOpenForwardOnly);
    if Assigned(Callback) then
    begin
      Callback.OnOpenState('Loading data');
      Callback.OnOpenProgress(False, RC);
    end;
    DT := GetTickCount;
    while not Recordset.Eof do
    begin
      if Assigned(Callback) and (GetTickCount - DT > 300) then
      begin
        if not Callback.OnOpenProgress(True, FAllCount) then Break;
        DT := GetTickCount;
      end;
      Stream := Recordset.CreateBlobStream(Recordset.Fields[1], bmRead);
      Size := Stream.Read(Features[0], VF_MAX_FEATURES_SIZE);
      Stream.Free;
      AddItem(Recordset.Fields[0].AsInteger, @Features[0], Size, Recordset.Fields[2].AsString);
      Recordset.Next;
    end;

    if Assigned(Callback) then Callback.OnOpenState('Opening recordset');
    OpenRecordset;

    if Assigned(Callback) then Callback.OnOpenState('Done');
  except
    on E: Exception do
    begin
      if Assigned(Callback) then Callback.OnOpenState('Error');
      try
        Close;
      except
        // do nothing
      end;
      raise Exception.Create('Can not open database'#13 + E.Message);
    end;
  end;
end;

procedure TVFDatabase.OpenRecordset(CursorType: TCursorType;
  LockType: TADOLockType; CursorLocation: TCursorLocation);
begin
  CloseRecordset;
  Recordset.CursorType := CursorType;
  Recordset.LockType := LockType;
  Recordset.CursorLocation := CursorLocation;
  Recordset.Open;
  FOpened := True;
end;

procedure TVFDatabase.Remove(Index: Integer);
var
  Item: PVFDBItem;
  ID: Integer;
  G: Integer;
  I: Integer;
  C: Integer;
  CP: Integer;
  K: Integer;
begin
  CheckIndex(Index);
  Item := FAllItems[Index];
  ID := Item.ID;
  G := VFFeatGetG(Item.Features); VFCheckResult(G);
  I := Item.GIndex;
  Dec(FCount[G]);
  C := FCount[G];
  Move(Pointer(LongWord(FItems[G]) + LongWord(I + 1) * SizeOf(PVFDBItem))^,
    Pointer(LongWord(FItems[G]) + LongWord(I) * SizeOf(PVFDBItem))^,
    SizeOf(PVFDBItem) * (C - I));
  for K := I to C - 1 do
    Dec(FItems[G][K].GIndex);
  CP := VFDBCapacity(C);
  if CP <> FCapacity[G] then
  begin
    SetLength(FItems[G], CP);
    FCapacity[G] := CP;
  end;
  Dec(FAllCount);
  Move(Pointer(LongWord(FAllItems) + LongWord(Index + 1) * SizeOf(PVFDBItem))^,
    Pointer(LongWord(FAllItems) + LongWord(Index) * SizeOf(PVFDBItem))^,
    SizeOf(PVFDBItem) * (FAllCount - Index));
  for K := Index to FAllCount - 1 do
    Dec(FAllItems[K].Index);
  CP := VFDBAllCapacity(FAllCount);
  if FAllCapacity <> CP then
  begin
    SetLength(FAllItems, CP);
    FAllCapacity := CP;
  end;
  Heap.Free(Item.Features);
  Heap.Free(PByte(Item));
  if FOpened then
  begin
    try
      if FindRecord(ID) then Recordset.Delete;
    except
      on E: Exception do
        raise Exception.Create('Can not delete record from the database'#13 + E.Message);
    end;
  end;
end;

procedure TVFDatabase.SetFeatures(Index: Integer; Features: PByte;
  Size: LongWord);
begin
  CheckIndex(Index);
  Heap.Free(FAllItems[Index].Features);
  FAllItems[Index].Features := Heap.Alloc(Size);
  Move(Features^, FAllItems[Index].Features^, Size);
  FAllItems[Index].FeaturesSize := Size;
  if FOpened then
  begin
    UpdateRecord(Index);
  end;
end;

procedure TVFDatabase.SetFingerID(Index: Integer; const Value: string);
begin
  CheckIndex(Index);
  CopyFingerID(FAllItems[Index].FingerID, FAllItems[Index].FingerIDLength, PChar(Value), Length(Value));
  if FOpened then
  begin
    UpdateRecord(Index);
  end;
end;

procedure TVFDatabase.UpdateRecord(Index: Integer; IsNew: Boolean = False);
var
  Stream: TStream;
begin
  try
    if IsNew then Recordset.Insert
    else begin
      FindRecord(FAllItems[Index].ID);
      Recordset.Edit;
    end;
    Stream := Recordset.CreateBlobStream(Recordset.Fields[1], bmWrite);
    try
      Stream.WriteBuffer(FAllItems[Index].Features^, FAllItems[Index].FeaturesSize);
    finally
      Stream.Free;
    end;
    Recordset.Fields[2].AsString := FAllItems[Index].FingerID;
    Recordset.Post;
  except
    on E: Exception do
      raise Exception.Create('Can not update record in the database'#13 + E.Message);
  end;
end;

{ TSOHeap }

function TSOHeap.Alloc(Size: Word): PByte;
var
  MemSize: LongWord;
  I, J: Integer;
  Block: PByte;
begin
  if Size = 0 then
  begin
    Result := nil;
    Exit;
  end;
  MemSize := Size + SizeOf(Word);
  Block := nil;
  J := -1;
  for I := 0 to BlockCount - 1 do
  begin
    if BlockSize - Sizes[I] >= MemSize then
    begin
      Block := Blocks[I];
      J := I;
      Break;
    end;
  end;
  if Block = nil then
  begin
    if BlockCount = SOHEAP_MAX_BLOCK_COUNT then raise Exception.Create('SOHeap: max block count riched');
    GetMem(Block, BlockSize);
    Blocks[BlockCount] := Block;
    Sizes[BlockCount] := 0;
    J := BlockCount;
    Inc(BlockCount);
  end;
  Inc(Block, Sizes[J]);
  PWord(Block)^ := Size; Inc(Block, SizeOf(Word));
  Inc(Sizes[J], MemSize);
  Result := Block;
end;

procedure TSOHeap.Clear;
var
  I: Integer;
begin
  for I := 0 to BlockCount - 1 do
    FreeMem(Blocks[I]);
  BlockCount := 0;
end;

constructor TSOHeap.Create(BlockSize: LongWord);
begin
  BlockCount := 0;
  Self.BlockSize := BlockSize;
end;

destructor TSOHeap.Destroy;
begin
  Clear;
  inherited;
end;

procedure TSOHeap.Free(Obj: PByte);
begin
  // Do nothing - memory will be allocated until heap is cleared
end;

end.
