//#if defined(linux) || defined (__LINUX__)
//	#include "linux_include.h"
//#else
	#include "stdafx.h"
//#endif


 /*
 * Class:     VeriFinger_VeriFingerWrapper
 * Method:    VFExtract
 * Signature: (II[BI[BI)LVeriFinger/VeriFingerExtractionResult;

   public static native int VFExtract(int width, int height, byte[] image,
                                      int resolution, byte[] features, int context) throws VeriFingerException;
 */

JNIEXPORT jobject JNICALL Java_VeriFinger_VeriFingerWrapper_VFExtract
  (JNIEnv * env, jclass jcls, jint width, jint height, jbyteArray jimage,
   jint resolution, jbyteArray jfeatures, jint context)
{
    int res = VFE_OK;

    try
    {
        unsigned long size = 0;

        // get image
        jbyte* image = env->GetByteArrayElements(jimage, NULL);
        if (image == NULL)
            ThrowJavaIllegalArgumentException(env);
        else
        {
            // get features
            jbyte* features = env->GetByteArrayElements(jfeatures, NULL);
            if (features == NULL)
            {
                // release image
                env->ReleaseByteArrayElements(jimage, image, 0);

                ThrowJavaIllegalArgumentException(env);
            }
            else
            {
                res = VFExtract(width, height, (BYTE*) image, resolution, (BYTE*) features, &size,
                                (HVFCONTEXT) context);

                // release image
                env->ReleaseByteArrayElements(jimage, image, 0);

                // release features
                env->ReleaseByteArrayElements(jfeatures, features, 0);

                if ((res != VFE_LOW_QUALITY_IMAGE) && VFFailed(res))
                    ThrowVeriFingerException(env, res);
                else
                {
                    // preparing result
                    jclass newExcCls = env->FindClass("VeriFinger/VeriFingerExtractionResult");
                    if (newExcCls == NULL)
                    {
                        ThrowJavaClassNotFoundException(env);
                    }
                    else
                    {
                        jmethodID constr = env->GetMethodID(newExcCls, "<init>", "(IZ)V");
                        jobject result = (jobject)env->NewObject(newExcCls, constr, (jint) size, (jboolean) (res == VFE_LOW_QUALITY_IMAGE));

                        if (result == NULL)
                            ThrowJavaException(env);

                        env->DeleteLocalRef(newExcCls);

                        return result;
                    }
                }
            }
        }
    }
    catch(...)
    {
        ThrowJavaException(env);
    }

    return NULL;
}
