/********************************************************\
|*                                                      *|
|*                  VeriFinger 5.0 SDK                  *|
|*                                                      *|
|*  VF42Parser.cpp                                      *|
|*    Visual Basic 6 parser implementation              *|
|*                                                      *|
|*  Copyright (C) 1998-2006 Neurotechnologija           *|
|*                                                      *|
\********************************************************/

#include <windows.h>
#include <oleauto.h>
#include <comdef.h>
#include <math.h>
#include <windowsx.h>
#include "Image.h"
#include "vfinger.h"

// -------------------------------------------------------------
// Useful defines
// -------------------------------------------------------------

#define CheckArray(var) if (!(var->vt & VT_ARRAY) || !(var->vt & VT_UI1)) return VFE_INVALID_ARGUMENT
#define CheckArrayVariant(var) if (!(var->vt & VT_ARRAY) || !(var->vt & VT_VARIANT)) return VFE_INVALID_ARGUMENT
#define GetSafeArray(pSA, var) if (var->vt & VT_BYREF) pSA = *(var->pparray); else pSA = var->parray

typedef BYTE* BYTEPTR;

#define VF_MAX_MINUTIA_COUNT 1024

#define VF_MAX_SINGULAR_POINT_COUNT 64

#define VF_BLOCK_SIZE 16
#define VF_MAX_IMAGE_DIMENSION  2048

#define VF_MAX_BLOCKED_ORIENTS_DIMENSION (VF_MAX_IMAGE_DIMENSION / VF_BLOCK_SIZE)

#define VF_MAX_FEATURES_SIZE 10000

struct VFMINUTIAE
{
        int count;
        int x[VF_MAX_MINUTIA_COUNT];
        int y[VF_MAX_MINUTIA_COUNT];
        BYTE d[VF_MAX_MINUTIA_COUNT];
        BYTE c[VF_MAX_MINUTIA_COUNT];
};

struct VFSINGULARPOINTS
{
        int count;
        int x[VF_MAX_SINGULAR_POINT_COUNT];
        int y[VF_MAX_SINGULAR_POINT_COUNT];
        BYTE t[VF_MAX_SINGULAR_POINT_COUNT];
        BYTE d[VF_MAX_SINGULAR_POINT_COUNT];
};

struct VFBLOCKEDORIENTS
{
        int width;
        int height;
        BYTE bits[VF_MAX_BLOCKED_ORIENTS_DIMENSION][VF_MAX_BLOCKED_ORIENTS_DIMENSION];
};

struct VFFEATURES
{
        BYTE g;
        VFMINUTIAE m;
        VFSINGULARPOINTS sp;
        VFBLOCKEDORIENTS bo;
};

INT WINAPI VBVFDecompressFeatures(VARIANT* vfeatures, VFFEATURES* f)
{
    BYTE*       bfeatures   = NULL;
    SAFEARRAY*  features    = NULL;

    if (vfeatures == NULL || f == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vfeatures);

    GetSafeArray(features, vfeatures);

    if (features == NULL)
        return VFE_INVALID_ARGUMENT;

    SafeArrayAccessData(features, (void**)&(bfeatures));
    // decompression

	f->g = VFFeatGetG(bfeatures);

	VFFeatGetBOSize(bfeatures, &f->bo.width, &f->bo.height);
	VFFeatGetBO(bfeatures, (BYTE*) f->bo.bits);

	f->m.count = VFFeatGetMinutiaCount(bfeatures);
	if (f->m.count > 0)
	{
		VFMinutia* m = new VFMinutia[f->m.count];
		VFFeatGetMinutiae(bfeatures, m);

		for (int i = 0; i < f->m.count; i++)
		{
			f->m.x[i] = m[i].X;
			f->m.y[i] = m[i].Y;
			f->m.c[i] = m[i].C;
			f->m.d[i] = m[i].D;
		}
		delete[] m;
	}

	f->sp.count = VFFeatGetSPCount(bfeatures);
	if (f->sp.count > 0)
	{
		VFSingularPoint* sp = new VFSingularPoint[f->sp.count];
		VFFeatGetSP(bfeatures, sp);

		for (int i = 0; i < f->sp.count; i++)
		{
			f->sp.x[i] = sp[i].X;
			f->sp.y[i] = sp[i].Y;
			f->sp.t[i] = sp[i].T;
			f->sp.d[i] = sp[i].D;
		}
		delete[] sp;
	}

	SafeArrayUnaccessData(features);

	return VFE_OK;
}

INT WINAPI VBVFCompressFeatures(VFFEATURES* f, VARIANT* vfeatures, LONG* size)
{
    BYTE bfeatures[VF_MAX_FEATURES_SIZE];

    if (vfeatures == NULL || size == NULL)
        return VFE_ARGUMENT_NULL;

	// compress
	VFMinutia* m = NULL;
	VFSingularPoint* sp = NULL;

	if (f->m.count > 0)
	{
		m = new VFMinutia[f->m.count];
		for (int i = 0; i < f->m.count; i++)
		{
			m[i].X = f->m.x[i];
			m[i].Y = f->m.y[i];
			m[i].C = f->m.c[i];
			m[i].D = f->m.d[i];
		}
	}

	if (f->sp.count > 0)
	{
		sp = new VFSingularPoint[f->sp.count];
		for (int i = 0; i < f->sp.count; i++)
		{
			sp[i].X = f->sp.x[i];
			sp[i].Y = f->sp.y[i];
			sp[i].T = (VFSingularPointType) f->sp.t[i];
			sp[i].D = f->sp.d[i];
		}
	}

	*size = VFFeatSet(f->g, f->m.count, m, f->sp.count, sp, f->bo.width, f->bo.height, (BYTE*) f->bo.bits, bfeatures);

	if (m != NULL)
		delete[] m;
	
	if (sp != NULL)
		delete[] sp;

    if (*size <= 0)
        return VFE_FAILED;

    if (FAILED(VariantClear(vfeatures)))
        return VFE_FAILED;

    SAFEARRAYBOUND sabound;

    sabound.lLbound   = 0;
        sabound.cElements = *size;

    SAFEARRAY* pSA = SafeArrayCreate(VT_UI1, 1, &sabound);

    if (pSA == NULL)
        return VFE_FAILED;

    BYTE *psaData = NULL;

    // Get a pointer to the elements of the array.
    if (FAILED(SafeArrayAccessData(pSA, (void**)&psaData)))
        return VFE_FAILED;

    for (int i = 0; i < *size; ++i)
        psaData[i] = bfeatures[i];

    vfeatures->vt = VT_UI1 | VT_ARRAY;
    vfeatures->parray = pSA;

    SafeArrayUnaccessData(pSA);

    return VFE_OK;
}

// -------------------------------------------------------------
// Initialization
// -------------------------------------------------------------

INT WINAPI VBVFInitialize()
{
	return VFInitialize();
}

INT WINAPI VBVFFinalize()
{
	return VFFinalize();
}

// -------------------------------------------------------------
// Registration
// -------------------------------------------------------------

INT WINAPI VBVFIsRegistered()
{
	return (VFRegistrationType() == VF_RT_UNREGISTERED)?0:1;
}

// -------------------------------------------------------------
// Contexts
// -------------------------------------------------------------

INT WINAPI VBVFCreateContext()
{
	return (INT) VFCreateContext();
}

INT WINAPI VBVFFreeContext(INT context)
{
	return VFFreeContext((HVFCONTEXT)context);
}

// -------------------------------------------------------------
// Parameters
// -------------------------------------------------------------

INT WINAPI VBVFGetParameterRaw(INT parameter, INT value, INT context)
{
	return VFGetParameter(parameter, (void*)value, (HVFCONTEXT) context);
}

INT WINAPI VBVFSetParameterRaw(INT parameter, INT value, INT context)
{
	return VFSetParameter(parameter, value, (HVFCONTEXT) context);
}

INT WINAPI VBVFGetParameterType(LONG parameter)
{
    LONG param = parameter;
    LONG result = VFGetParameter(VFP_TYPE, (void*)param, NULL);
    return result;
}
// -------------------------------------------------------------
// Parameters with Variants
// -------------------------------------------------------------

INT WINAPI VBVFGetParameter(INT parameter, VARIANT* value, INT context)
{
    INT     result = VFE_OK;
    LONG    pvalue = 0;
    LONG    ptype;
    VARTYPE type;
    BOOL    boolValue;
    char*   strValue = NULL;
    _bstr_t bstrValue;
    INT     l;

    if (value == NULL)
        return VFE_ARGUMENT_NULL;

    ptype = VBVFGetParameterType(parameter);
    if (VFFailed(ptype))
        return VFE_INVALID_PARAMETER;

    switch (ptype)
    {
    case VF_TYPE_BYTE:  type = VT_I1; break;
    case VF_TYPE_SBYTE: type = VT_I1; break;
    case VF_TYPE_WORD:  type = VT_I2; break;
    case VF_TYPE_SHORT: type = VT_I2; break;
    case VF_TYPE_DWORD: type = VT_I4; break;
    case VF_TYPE_INT:   type = VT_I4; break;
    case VF_TYPE_BOOL:  type = VT_BOOL; break;
    case VF_TYPE_CHAR:  type = VT_I1; break;
    case VF_TYPE_STRING: type = VT_BSTR; break;
    default:
        result = VFE_INVALID_PARAMETER;
    }

    if (result == VFE_OK)
    {
        VariantClear(value);
        value->vt = type;
        switch (type)
        {
        case VT_I1: pvalue = (int)&(value->bVal); break;
        case VT_I2: pvalue = (int)&(value->iVal); break;
        case VT_I4: pvalue = (int)&(value->lVal); break;
        case VT_BOOL: pvalue = (int)&boolValue; break;
        case VT_BSTR:
            l = VFGetParameter(parameter, NULL, NULL);
            strValue = new char[l+1];
            pvalue = (int)strValue;
            break;
        default:
            result = VFE_INVALID_ARGUMENT;
        }

        if (result == VFE_OK)
            result = VFGetParameter(parameter, (void*)pvalue, (HVFCONTEXT) context);

        if (type == VT_BOOL)
            value->boolVal =  boolValue;

        if (type == VT_BSTR)
        {
            bstrValue = strValue;
            value->bstrVal = bstrValue.copy();
            delete[] strValue;
        }
    }
    return result;
}

INT WINAPI VBVFSetParameter(INT parameter, VARIANT value, INT context)
{
    LONG    pvalue = 0;
    INT     result = VFE_OK;
    VARIANT vvalue;
    VARTYPE type;
    LONG    ptype;

    _bstr_t bstrValue;

    ptype = VBVFGetParameterType(parameter);
    if (VFFailed(ptype))
        return VFE_INVALID_PARAMETER;

    switch (ptype)
    {
    case VF_TYPE_BYTE:  type = VT_I1; break;
    case VF_TYPE_SBYTE: type = VT_I1; break;
    case VF_TYPE_WORD:  type = VT_I2; break;
    case VF_TYPE_SHORT: type = VT_I2; break;
    case VF_TYPE_DWORD: type = VT_I4; break;
    case VF_TYPE_INT:   type = VT_I4; break;
    case VF_TYPE_BOOL:  type = VT_BOOL; break;
    case VF_TYPE_CHAR:  type = VT_I1; break;
    case VF_TYPE_STRING: type = VT_BSTR; break;
    default:
        result = VFE_INVALID_ARGUMENT;
    }

    if (result == VFE_OK)
    {
        VariantInit(&vvalue);
        if (FAILED(VariantChangeType(&vvalue, &value, VARIANT_ALPHABOOL, type)))
            result = VFE_INVALID_ARGUMENT;

        switch (type)
        {
        case VT_I1: pvalue = vvalue.bVal; break;
        case VT_I2: pvalue = vvalue.iVal; break;
        case VT_I4: pvalue = vvalue.lVal; break;
        case VT_BOOL: pvalue = vvalue.boolVal; break;
        case VT_BSTR:
            bstrValue = vvalue.bstrVal; // bstrValue.Attach(vvalue.bstrVal);
            pvalue = (int)((char*)bstrValue);
            break;
        default:
            result = VFE_INVALID_ARGUMENT;
        }

        if (result == VFE_OK)
            result = VFSetParameter(parameter, pvalue, (HVFCONTEXT) context);

        VariantClear(&vvalue);
    }
    return result;
}

// -------------------------------------------------------------
// Features extraction
// -------------------------------------------------------------

INT WINAPI VBVFExtract(INT width, INT height, VARIANT* vimage, LONG resolution, VARIANT* vfeatures, INT* size, INT context)
{
    if (vimage == NULL || vfeatures  == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vimage);
    CheckArray(vfeatures);

    SAFEARRAY* image = NULL;
    SAFEARRAY* features = NULL;

    GetSafeArray(image, vimage);
    GetSafeArray(features, vfeatures);

    if (image == NULL || features == NULL)
        return VFE_INVALID_ARGUMENT;

        BYTE* bimage    = NULL;
        BYTE* bfeatures = NULL;
        INT   result    = VFE_FAILED;

        if (size == NULL)
                return VFE_ARGUMENT_NULL;

        if (FAILED(SafeArrayAccessData(image, (void**)&bimage)))
                return VFE_INVALID_ARGUMENT;

        if (FAILED(SafeArrayAccessData(features, (void**)&bfeatures)))
        {
                SafeArrayUnaccessData(image);
                return VFE_INVALID_ARGUMENT;
        }
        __try
        {
                result = VFExtract(width, height, bimage, resolution, bfeatures, (DWORD*) size, (HVFCONTEXT) context);
        }
        __finally
        {
        }

        SafeArrayUnaccessData(image);
        SafeArrayUnaccessData(features);

        return result;
}

// -------------------------------------------------------------
// Features generalization
// -------------------------------------------------------------

INT WINAPI VBVFGeneralize(INT count, VARIANT* vgen_features, VARIANT* vfeatures, INT* size, INT context)
{
    if (vgen_features == NULL || vfeatures == NULL)
        return VFE_ARGUMENT_NULL;

    SAFEARRAY* gen_features = NULL;
    SAFEARRAY* features = NULL;

    CheckArrayVariant(vgen_features);
    CheckArray(vfeatures);

    GetSafeArray(gen_features, vgen_features);
    GetSafeArray(features, vfeatures);

    if (features == NULL || gen_features == NULL)
        return VFE_INVALID_ARGUMENT;

        VARIANT* bgenfeatures   = NULL;
        BYTE*    bfeatures              = NULL;
        INT              result                 = VFE_FAILED;
    BYTE**   gfeatures      = NULL;

        if (size == NULL)
                return VFE_ARGUMENT_NULL;

    if (count <= 0)
        return VFE_INVALID_ARGUMENT;

        if (FAILED(SafeArrayAccessData(gen_features, (void**)&bgenfeatures)))
                return VFE_INVALID_ARGUMENT;

        if (FAILED(SafeArrayAccessData(features, (void**)&bfeatures)))
        {
                SafeArrayUnaccessData(gen_features);
                return VFE_INVALID_ARGUMENT;
        }

    __try
        {
        // accessing features
        gfeatures = new BYTEPTR[count];

        __try
        {
            int i;

            for (i = 0; i < count; ++i)
            {
                SAFEARRAY *pSA = NULL;
                if (bgenfeatures[i].vt & VT_BYREF)
                    pSA = *(bgenfeatures[i].pparray);
                else
                    pSA = bgenfeatures[i].parray;

                SafeArrayAccessData(pSA, (void**)&(gfeatures[i]));
            }

                __try
                {
                        result = VFGeneralize(count, gfeatures, bfeatures, (DWORD*)size, (HVFCONTEXT) context);
                }
                __finally
                {
                }

            for (i = 0; i < count; ++i)
            {
                SAFEARRAY *pSA = NULL;
                if (bgenfeatures[i].vt & VT_BYREF)
                    pSA = *(bgenfeatures[i].pparray);
                else
                    pSA = bgenfeatures[i].parray;

                SafeArrayUnaccessData(pSA);
            }
        }
            __finally
            {
            }

        delete[] gfeatures;
    }
    __finally
        {
        }

        SafeArrayUnaccessData(gen_features);
        SafeArrayUnaccessData(features);

        return result;
}

// -------------------------------------------------------------
// Verification
// -------------------------------------------------------------

INT WINAPI VBVFVerify(VARIANT* vfeatures1, VARIANT* vfeatures2, void* md, INT context)
{
    if (vfeatures1 == NULL || vfeatures2 == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vfeatures1);
    CheckArray(vfeatures2);

    SAFEARRAY* features1 = NULL;
    SAFEARRAY* features2 = NULL;

    GetSafeArray(features1, vfeatures1);
    GetSafeArray(features2, vfeatures2);

    if (features1 == NULL || features2 == NULL)
        return VFE_INVALID_ARGUMENT;

        INT		result      = VFE_FAILED;
        BYTE*	bfeatures1      = NULL;
        BYTE*	bfeatures2      = NULL;

        //if (md == NULL)
                //return VFE_ARGUMENT_NULL;

        if (FAILED(SafeArrayAccessData(features1, (void**)&bfeatures1)))
                return VFE_INVALID_ARGUMENT;

        if (FAILED(SafeArrayAccessData(features2, (void**)&bfeatures2)))
        {
                SafeArrayUnaccessData(features1);
                return VFE_INVALID_ARGUMENT;
        }

        __try
        {
                result = VFVerify(bfeatures1, bfeatures2, (VFMatchDetails*) md, (HVFCONTEXT) context);
        }
        __finally
        {
        }

        SafeArrayUnaccessData(features1);
        SafeArrayUnaccessData(features2);

        return result;
}

// -------------------------------------------------------------
// Identification
// -------------------------------------------------------------

INT WINAPI VBVFIdentifyStart(VARIANT* vtest_features, INT context)
{
    SAFEARRAY* test_features = NULL;

    if (vtest_features == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vtest_features);

    GetSafeArray(test_features, vtest_features);

    if (test_features == NULL)
        return VFE_INVALID_ARGUMENT;

        INT   result                    = VFE_FAILED;
        BYTE* btest_features    = NULL;

        if (FAILED(SafeArrayAccessData(test_features, (void**)&btest_features)))
                return VFE_INVALID_ARGUMENT;

        __try
        {
                result = VFIdentifyStart(btest_features, (HVFCONTEXT) context);
        }
        __finally
        {
        }

        SafeArrayUnaccessData(test_features);

        return result;
}

INT WINAPI VBVFIdentifyNext(VARIANT* vsample_features, void* md, INT context)
{
    SAFEARRAY* sample_features = NULL;

    if (vsample_features == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vsample_features);

    GetSafeArray(sample_features, vsample_features);

    if (sample_features == NULL)
        return VFE_INVALID_ARGUMENT;

        INT       result                        = VFE_FAILED;
        BYTE* bsample_features  = NULL;

        //if (md == NULL)
                //return VFE_ARGUMENT_NULL;

        if (FAILED(SafeArrayAccessData(sample_features, (void**)&bsample_features)))
                return VFE_INVALID_ARGUMENT;

        __try
        {
                result = VFIdentifyNext(bsample_features, (VFMatchDetails*) md, (HVFCONTEXT) context);
        }
        __finally
        {
        }

        SafeArrayUnaccessData(sample_features);

        return result;
}

INT WINAPI VBVFIdentifyEnd(INT context)
{
        return VFIdentifyEnd((HVFCONTEXT) context);
}


INT WINAPI VBDeleteHandle(INT handle)
{
	DeleteObject((HGDIOBJ)handle);

	return 0;
}

INT WINAPI VBImageToHandle(INT width, INT height, VARIANT* vimage, INT* handle, INT Pallete)
{
    SAFEARRAY* image = NULL;

    if (vimage == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vimage);

    GetSafeArray(image, vimage);

    if (image == NULL)
        return VFE_INVALID_ARGUMENT;

    INT   result = VFE_FAILED;
        BYTE* bimage = NULL;

    CImage img;

        if (handle == NULL)
                return VFE_ARGUMENT_NULL;

        if (FAILED(SafeArrayAccessData(image, (void**)&bimage)))
                return VFE_INVALID_ARGUMENT;

    img.SetSize(width, height);
    try
    {
    for (int x = 0; x < width; ++x)
        for (int y = 0; y < height; ++y)
            img.m_img.m_Buffer[y][x] = bimage[width*y+x];
    }
    catch(...)
    {}

    *handle = (INT) img.ToHandle(Pallete == 0);

        SafeArrayUnaccessData(image);

        return result;
}

INT WINAPI VBHandleToImage(INT handle, INT* width, INT* height, VARIANT* vimage)
{
    if (width == NULL || height == NULL || vimage == NULL)
        return VFE_ARGUMENT_NULL;

    CImage img;

    if (FAILED(img.FromHandle((HBITMAP) handle)))
        return VFE_FAILED;

    *width = img.w;
    *height = img.l;

    if (FAILED(VariantClear(vimage)))
        return VFE_FAILED;

    SAFEARRAYBOUND sabound;

    sabound.lLbound   = 0;
        sabound.cElements = img.w*img.l;

    SAFEARRAY* pSA = SafeArrayCreate(VT_UI1, 1, &sabound);

    if (pSA == NULL)
        return VFE_FAILED;

    BYTE *psaData = NULL;

    // Get a pointer to the elements of the array.
    if (FAILED(SafeArrayAccessData(pSA, (void**)&psaData)))
        return VFE_FAILED;

    for (int x = 0; x < *width; ++x)
        for (int y = 0; y < *height; ++y)
            //psaData[(*width)*((*height)-y-1)+x] = img.m_img.m_Buffer[y][x];
            psaData[(*width)*y+x] = img.m_img.m_Buffer[y][x];

    vimage->vt = VT_UI1 | VT_ARRAY;
    vimage->parray = pSA;

    SafeArrayUnaccessData(pSA);

    return VFE_OK;
}

INT WINAPI VBLoadImageFromFile(CHAR* filename, INT* width, INT* height, VARIANT* vimage)
{
    if (width == NULL || height == NULL || vimage == NULL)
        return VFE_ARGUMENT_NULL;

    CImage img;

    if (FAILED(img.Load(filename, 1)))
        return VFE_FAILED;

    *width = img.w;
    *height = img.l;

    if (*width <= 0 || *height <= 0)
        return VFE_FAILED;

    if (FAILED(VariantClear(vimage)))
        return VFE_FAILED;

    SAFEARRAYBOUND sabound;

    sabound.lLbound   = 0;
        sabound.cElements = img.w*img.l;

    SAFEARRAY* pSA = SafeArrayCreate(VT_UI1, 1, &sabound);

    if (pSA == NULL)
        return VFE_FAILED;

    BYTE *psaData = NULL;

    // Get a pointer to the elements of the array.
    if (FAILED(SafeArrayAccessData(pSA, (void**)&psaData)))
        return VFE_FAILED;

    for (int x = 0; x < *width; ++x)
        for (int y = 0; y < *height; ++y)
            //psaData[(*width)*((*height)-y-1)+x] = img.m_img.m_Buffer[y][x];
            psaData[(*width)*(y)+x] = img.m_img.m_Buffer[y][x];

    vimage->vt = VT_UI1 | VT_ARRAY;
    vimage->parray = pSA;

    SafeArrayUnaccessData(pSA);

    return VFE_OK;
}

inline int Round(double value) { return (int)(value < 0 ? value - 0.5 : value + 0.5); }

// Directions
#define VFDIR_0                         0
#define VFDIR_45                        30
#define VFDIR_90                        (VFDIR_45 * 2)
#define VFDIR_135                       (VFDIR_45 * 3)
#define VFDIR_180                       (VFDIR_45 * 4)
#define VFDIR_225                       (VFDIR_45 * 5)
#define VFDIR_270                       (VFDIR_45 * 6)
#define VFDIR_315                       (VFDIR_45 * 7)
#define VFDIR_360                       (VFDIR_45 * 8)
#define VFDIR_UNKNOWN           127
#define VFDIR_BACKGROUND        255
// Conversion from VFDIR_XXX to degrees and vice a versa
#define VFDirToDeg(dir) ((dir) * 180 / VFDIR_180)
#define VFDegToDir(deg) ((deg) * VFDIR_180 / 180)

// Singular point types
#define VFFEAT_SP_DELTA                 0x03
#define VFFEAT_SP_CORE                  0x01
#define VFFEAT_SP_DOUBLECORE    0x02

void PaintBO(HDC hdc, int x, int y, BYTE dir)
{
    double hbs = VF_BLOCK_SIZE / 2;
    double a = VFDirToRad(dir);
    int dx = Round(cos(a) * hbs);
    int dy = Round(sin(a) * hbs);
    MoveToEx(hdc, x, y, NULL);
    LineTo(hdc, x + dx, y + dy);
    MoveToEx(hdc, x, y, NULL);
    LineTo(hdc, x - dx, y - dy);
}

void PaintDir(HDC hdc, int x, int y, BYTE dir, int length)
{
    double a = VFDirToRad(dir);
    int dx = Round(cos(a) * length);
    int dy = Round(sin(a) * length);
    MoveToEx(hdc, x, y, NULL);
    LineTo(hdc, x + dx, y + dy);
}

INT WINAPI VBDrawFeatures(VFFEATURES* f, INT width, INT height, VARIANT* vimage, INT pallete, INT parameters, INT* handle)
{
    SAFEARRAY* image = NULL;

    if (vimage == NULL)
        return VFE_ARGUMENT_NULL;

    CheckArray(vimage);

    GetSafeArray(image, vimage);

    if (image == NULL)
        return VFE_INVALID_ARGUMENT;

    INT   result = VFE_FAILED;
        BYTE* bimage = NULL;

    CImage img;

        if (FAILED(SafeArrayAccessData(image, (void**)&bimage)))
                return VFE_INVALID_ARGUMENT;

    img.SetSize(width, height);
    try
    {
        for (int x = 0; x < width; ++x)
            for (int y = 0; y < height; ++y)
                img.m_img.m_Buffer[y][x] = bimage[width*y+x];
    }
    catch(...)
    {}

    BITMAPINFO bmi;

    ::ZeroMemory (&bmi, sizeof (bmi));
    bmi.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
    bmi.bmiHeader.biWidth = (width%4)?width+(4-(width%4)):width;
    bmi.bmiHeader.biHeight = height;
    bmi.bmiHeader.biPlanes = 1;
    bmi.bmiHeader.biBitCount = 24;
    bmi.bmiHeader.biCompression = BI_RGB;

    BYTE * buffi = NULL;
        long ihandle = (long) CreateDIBSection(
                NULL,
        &(bmi),
                DIB_RGB_COLORS,
                (VOID**)&buffi,
                NULL,0);

    try
    {
        for (int x = 0; x < width; ++x)
            for (int y = 0; y < height; ++y)
            {
                if (pallete == 0) // green
                {
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3] = bimage[width*y+x];
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3+1] = 255;
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3+2] = bimage[width*y+x];
                }
                else
                {
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3] = bimage[width*y+x];
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3+1] = bimage[width*y+x];
                    buffi[bmi.bmiHeader.biWidth*(height-y-1)*3+x*3+2] = bimage[width*y+x];
                }
            }
    }
    catch(...)
    {}

    HDC hdc = CreateCompatibleDC(NULL);

    HGDIOBJ hold = SelectObject(hdc, (HGDIOBJ) ihandle);

    long bo_color = 0xFF0000;
    long sp_color = 0x0000FF;
    long minutia_color = 0x0000FF;
    int x = 0;
    int y = 0;

    HBRUSH old_brush = SelectBrush(hdc, GetStockBrush(NULL_BRUSH));

    if(parameters & 0x2)  // draw blocked orientations
    {
        HPEN pen = CreatePen(PS_SOLID, max(1, Round(1.0)), bo_color);
        HPEN old_pen = SelectPen(hdc, pen);
        int bs = VF_BLOCK_SIZE;
        int hbs = bs / 2;
        for(int i = 0, yy = y + hbs; i < f->bo.height; i++, yy += bs)
            for(int j = 0, xx = x + hbs; j < f->bo.width; j++, xx += bs)
            {
                BYTE dir = f->bo.bits[i][j];
                if(dir != VFDIR_BACKGROUND)
                {
                    if(dir != VFDIR_UNKNOWN)
                        PaintBO(hdc, xx, yy, dir);
                    else
                    {
                        PaintBO(hdc, xx, yy, VFDIR_0);
                        PaintBO(hdc, xx, yy, VFDIR_45);
                        PaintBO(hdc, xx, yy, VFDIR_90);
                        PaintBO(hdc, xx, yy, VFDIR_135);
                    }
                }
            }
        SelectPen(hdc, old_pen);
        DeletePen(pen);
    }

    if(parameters & 0x4) // draw singular_points
    {
        HPEN pen = CreatePen(PS_SOLID, max(2, Round(1.0)), sp_color);
        HPEN old_pen = SelectPen(hdc, pen);
        int xx, yy, s = max(1, Round(6 * 1.0));
        for(int i = 0; i < f->sp.count; i++)
        {
            xx = x + f->sp.x[i] * 1;
            yy = y + f->sp.y[i] * 1;
            switch(f->sp.t[i])
            {
                case VFFEAT_SP_CORE:
                    Rectangle(hdc, xx - s, yy - s, xx + s + 1, yy + s + 1);
                    if(f->sp.d[i] != VFDIR_UNKNOWN)
                        PaintDir(hdc, xx, yy, f->sp.d[i], 20);
                    break;
                case VFFEAT_SP_DOUBLECORE:
                    Ellipse(hdc, xx - s, yy - s, xx + s + 1, yy + s + 1);
                    break;
                case VFFEAT_SP_DELTA:
                    MoveToEx(hdc, xx - s, yy + s, NULL);
                    LineTo(hdc, xx, yy - s);
                    LineTo(hdc, xx + s, yy + s);
                    LineTo(hdc, xx - s, yy + s);
                    break;
            }
        }
        SelectPen(hdc, old_pen);
        DeletePen(pen);
    }

    if(parameters & 0x1) // draw minutiea
    {
        HPEN pen = CreatePen(PS_SOLID, max(2, Round(1.0)), minutia_color);
        HPEN old_pen = SelectPen(hdc, pen);
        int xx, yy, s = max(1, Round(4 * 1.0));
        for(int i = 0; i < f->m.count; i++)
        {
            xx = x + f->m.x[i];
            yy = y + f->m.y[i];
            Ellipse(hdc, xx - s, yy - s, xx + s + 1, yy + s + 1);
            PaintDir(hdc, xx, yy, f->m.d[i], 15);
        }
        SelectPen(hdc, old_pen);
        DeletePen(pen);
    }

    SelectBrush(hdc, old_brush);

    byte * buff = new byte[height*bmi.bmiHeader.biWidth*3];

    GetDIBits (hdc, (HBITMAP) ihandle, 0, height, buff, &bmi, DIB_RGB_COLORS);

    BYTE * ibuff = NULL;
    *handle = (int) CreateDIBSection(
            NULL,
            &(bmi),
            DIB_RGB_COLORS,
            (VOID**)&ibuff,
            NULL,0);

    memcpy(ibuff,buff,height*bmi.bmiHeader.biWidth*3);

    delete[] buff;
    SelectObject(hdc, (HGDIOBJ) hold);
    DeleteDC(hdc);
    DeleteObject((HGDIOBJ) ihandle);

    SafeArrayUnaccessData(image);

    return result;
}

INT WINAPI VBSaveImageToFile(CHAR* filename, INT width, INT height, VARIANT* vimage)
{
    if (vimage == NULL)
        return VFE_ARGUMENT_NULL;

        SAFEARRAY* image = NULL;

    CheckArray(vimage);

    GetSafeArray(image, vimage);

    if (image == NULL)
        return VFE_INVALID_ARGUMENT;

    BYTE* bimage = NULL;

    CImage img;

    if (FAILED(SafeArrayAccessData(image, (void**)&bimage)))
                return VFE_INVALID_ARGUMENT;

    img.SetSize(width, height);
    try
    {
        for (int x = 0; x < width; ++x)
            for (int y = 0; y < height; ++y)
                img.m_img.m_Buffer[y][x] = bimage[width*y+x];
    }
    catch(...)
    {}

        SafeArrayUnaccessData(image);

    if (FAILED(img.SaveBMP(filename)))
        return VFE_FAILED;

    return VFE_OK;
}

// -------------------------------------------------------------
// Registration
// -------------------------------------------------------------

INT WINAPI VBVFRegistrationType()
{
	return VFRegistrationType();
}

INT WINAPI VBVFGenerateId(CHAR * serial, CHAR * id)
{
	return VFGenerateId(serial, id);
}

INT WINAPI VBVFRegister(CHAR * serial, CHAR * key)
{
	return VFRegister(serial, key);
}



BOOL APIENTRY DllMain(HANDLE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved)
{
	switch (ul_reason_for_call)
	{
		case DLL_PROCESS_ATTACH:
			VFInitialize();
			break;
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH:
			while (VFFinalize() > 0);
			break;
	}
	return TRUE;
}