﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public class Distribution
	{
		#region Private fields

		private int[] items;
		private int[] cumItems;
		private long sum;
		private int valueCount;

		#endregion

		public Distribution(IList<int> items, int min)
		{
			if(items == null) throw new ArgumentNullException("items");
			int count = items.Count;
			this.items = new int[count];
			items.CopyTo(this.items, 0);
			this.min = min;
			Init();
		}

		//public Distribution(string fileName)
		//{
		//    BinaryReader reader = new BinaryReader(new FileStream(fileName, FileMode.Open));
		//    min = reader.ReadInt32();
		//    int count = reader.ReadInt32();
		//    items = new int[count];
		//    for(int i = 0; i < count; i++)
		//        items[i] = reader.ReadInt32();
		//    reader.Close();
		//    Init();
		//}

		private void Init()
		{
			int c = items.Length;
			cumItems = new int[c];
			if(c != 0)
			{
				valueCount = cumItems[0] = items[0];
				sum = valueCount * min;
				for(int i = 1; i < c; i++)
				{
					int count = items[i];
					cumItems[i] = cumItems[i - 1] + count;
					valueCount += count;
					sum += count * (min + i);
				}
			}
			else
			{
				valueCount = 0;
				sum = 0;
			}
		}

		public int ValueCount
		{
			get
			{
				return valueCount;
			}
		}

		private int min;

		public int Min
		{
			get
			{
				return min;
			}
		}

		public int Max
		{
			get
			{
				return min + items.Length - 1;
			}
		}

		public double Ave
		{
			get
			{
				return (double)sum / (double)valueCount;
			}
		}

		public double Var
		{
			get
			{
				double ave = Ave;
				int c = items.Length;
				double sum = 0;
				for(int i = 0; i < c; i++)
				{
					double v = min + i - ave;
					sum += v * v * items[i];
				}
				return sum / valueCount;
			}
		}

		public bool IsEmpty
		{
			get
			{
				return items.Length == 0;
			}
		}

		public double CountToProbability(int count)
		{
			return count / (double)valueCount;
		}

		public int GetCount(int value)
		{
			return value < min || value > Max ? 0 : items[value - min];
		}

		public double GetProbability(int value)
		{
			return CountToProbability(GetCount(value));
		}

		public int GetLTCount(int value)
		{
			return value <= min ? 0 : value > Max ? valueCount : cumItems[value - min - 1];
		}

		public int GetLTECount(int value)
		{
			return value < min ? 0 : value >= Max ? valueCount : cumItems[value - min];
		}

		public int GetGTCount(int value)
		{
			return value < min ? valueCount : value >= Max ? 0 : valueCount - cumItems[value - min];
		}

		public int GetGTECount(int value)
		{
			return value <= min ? valueCount : value > Max ? 0 : valueCount - cumItems[value - min - 1];
		}

		public double GetLTProbability(int value)
		{
			return CountToProbability(GetLTCount(value));
		}

		public double GetLTEProbability(int value)
		{
			return CountToProbability(GetLTECount(value));
		}

		public double GetGTProbability(int value)
		{
			return CountToProbability(GetGTCount(value));
		}

		public double GetGTEProbability(int value)
		{
			return CountToProbability(GetGTECount(value));
		}

		public IEnumerator<CountValuePair> GetCountPoints()
		{
			int c = items.Length;
			for(int i = 0; i < c; i++)
			{
				if(items[i] != 0)
					yield return new CountValuePair(items[i], i + min);
			}
		}

		public IEnumerator<ProbabilityValuePair> GetProbabilityPoints()
		{
			int c = items.Length;
			for(int i = 0; i < c; i++)
			{
				if(items[i] != 0)
					yield return new ProbabilityValuePair(CountToProbability(items[i]), i + min);
			}
		}

		//public void Save(string fileName)
		//{
		//    BinaryWriter writer = new BinaryWriter(new FileStream(fileName, FileMode.Create));
		//    writer.Write(min);
		//    int count = items.Length;
		//    writer.Write(count);
		//    for(int i = 0; i < count; i++)
		//        writer.Write(items[i]);
		//    writer.Close();
		//}
	}

	public struct CountValuePair
	{
		public CountValuePair(int count, int value)
		{
			Count = count;
			Value = value;
		}

		public int Count;
		public int Value;
	}

	public struct ProbabilityValuePair
	{
		public ProbabilityValuePair(double probability, int value)
		{
			Probability = probability;
			Value = value;
		}

		public double Probability;
		public int Value;
	}
}
