/* ONLY FOR KERNEL 2.4 or more */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

//#define FXDRIVER_DEBUG

//#define CONFIG_DEVFS_FS


#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/errno.h>    /* error codes */
#include <linux/version.h>
#include <linux/config.h>
#include <linux/init.h>

#include <linux/fs.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include <linux/usb.h>

#include "fxdriver.h"

/*
 * Version Information
 */
#define DRIVER_VERSION "v1.01"
#define DRIVER_AUTHOR "Giorgio Alboni, alboni@biometrika.it; Mauro Bartolomeoli (mbarto@infosia.it)"
#define DRIVER_DESC "Fx2000/3000 USB interface driver for Linux (2002-2006)"
#define BUFFER_IN_SIZE 120000
#define BUFFER_OUT_SIZE 1200

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,13)
#define KERNEL2613
#endif

#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,6,0)
#define KERNEL26
#else
#define KERNEL24
#endif

typedef int BOOL;

#define FALSE	0
#define TRUE	1

/* read_scanner timeouts -- RD_NAK_TIMEOUT * RD_EXPIRE = Number of seconds */
#define WR_NAK_TIMEOUT (1*HZ)	/* Default number of X seconds to wait */
#define RD_NAK_TIMEOUT (1*HZ)	/* Default number of X seconds to wait */
#define RD_EXPIRE 3		/* Number of attempts to wait X seconds */

/*
 * We need a local data structure, as it must be allocated for each new
 * scanner device plugged in the USB bus
 */
typedef struct fxdriver_device {
	struct usb_device *udev;			/* save off the usb device pointer */
	struct usb_interface *interface;		/* the interface for this device */
	char bufferIn[BUFFER_IN_SIZE];
	char bufferOut[BUFFER_OUT_SIZE];
	BOOL deviceOpened;
	int type;
} fxdriver_device_t;

typedef fxdriver_device_t *fxdriver_device_p;

int f_fxdriver_open(struct inode *inode, struct file *filp);
ssize_t f_fxdriver_read(struct file *filp, char *buf, size_t count, loff_t *f_pos);
ssize_t f_fxdriver_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos);
int f_fxdriver_release(struct inode *inode, struct file *filp);

static struct file_operations fxdriver_fops = {
	.read=       f_fxdriver_read,
	.write=      f_fxdriver_write,
	.open=       f_fxdriver_open,
	.release=    f_fxdriver_release,
};

#ifdef CONFIG_SYSFS
#ifdef KERNEL2613
static struct class_device *fxdriverdev;
static struct class *sysfs_class;
#else
static struct class_simple *sysfs_class;
#endif
#endif

fxdriver_device_p dev_fxdriver[FXDRIVER_NR_DEVS];
BOOL fxdriver_presence[FXDRIVER_NR_DEVS];

int fxdriver_major =   FXDRIVER_MAJOR;
int fxdriver_nr_devs = FXDRIVER_NR_DEVS;    /* number of bare scull devices */

#ifdef FX3000
const char* UCASENAME="FX3000";
const char* LCASENAME="fx3000";
const int basePort=200;
#else
const char* UCASENAME="FX2000";
const char* LCASENAME="fx2000";
const int basePort=100;
#endif

MODULE_PARM(fxdriver_major,"i");
MODULE_PARM(fxdriver_nr_devs,"i");

#ifdef FXDRIVER_DEBUG
static void print_dev_fxdriver (void)
{
	int i;

	/*
	 * Check for fx2000/3000 devices
	 */


	for (i=0; i<FXDRIVER_NR_DEVS ;i++) {

		if (dev_fxdriver[i]!=NULL) {

		    	FXDRIVER_DEBUG("%s n.%i(%4x)=%x\n",UCASENAME,i ,dev_fxdriver[i]->udev->descriptor.idProduct,dev_fxdriver[i]);
		} else {

			FXDRIVER_DEBUG("%s n.%i(---)= %x\n",UCASENAME,i,dev_fxdriver[i]);
		}
	}
}
#endif

/* -------------------------------------------------------------------------------*/
/* Handle proc file																																*/
/* -------------------------------------------------------------------------------*/

/* Proc read function: print some driver information */
int fxdriver_read_procmem(char *buf, char **start, off_t offset,
                   int count, int *eof, void *data)
{
	int i, len = 0;
	//int limit = count - 80; /* Don't print more than this */

	sprintf (buf,"Fingerprint scanner driver running\n\nDevice status:\n");
	len=strlen (buf);


	/*
	 * print fx2000/3000 devices
	 */

	for (i=0; i<FXDRIVER_NR_DEVS ;i++)
	{

		if (fxdriver_presence[i]) {
#ifdef FX3000
		    sprintf(buf+len,"%s n.%3i (%04x)= Connected (%s)\n",UCASENAME,i+basePort,dev_fxdriver[i]->udev->descriptor.idProduct,dev_fxdriver[i]->deviceOpened? "open":"close");
#else
		    sprintf(buf+len,"%s n.%3i (%04x)= Connected (%s)\n",UCASENAME,i?i+basePort:0,dev_fxdriver[i]->udev->descriptor.idProduct,dev_fxdriver[i]->deviceOpened? "open":"close");
#endif
		}
		else
		    sprintf(buf+len,"%s n.%3i (----)= Not Connected\n",UCASENAME,i?i+basePort:0);
		len=strlen (buf);
	}

  *eof = 1;
  return len;
}



/* Remove proc file */
static void fxdriver_create_proc()
{
    	create_proc_read_entry(LCASENAME, 0 /* default mode */,
                           NULL /* parent dir */, fxdriver_read_procmem,
                           NULL /* client data */);

}

/* Register proc file */
static void fxdriver_remove_proc()
{
    	remove_proc_entry(LCASENAME, NULL /* parent dir */);
}

static int get_scanner_index(int idProduct) {
    // Fx3000 scanner
    if(idProduct>=0x0300)
	return idProduct-0x0300;
    // Fx2000 scanner
    else
	return idProduct-0x0200;
}

static int get_scanner_minor(int idProduct) {
    // Fx3000 scanner
    if(idProduct>=0x0300)
	return idProduct-0x0300;
    // Fx2000 scanner
    else
	return idProduct-0x0200;
}

/* -------------------------------------------------------------------------------*/
/* Handle USB events																															*/
/* -------------------------------------------------------------------------------*/

/*
 * These two callbacks are invoked when an USB device is detached or attached
 * to the bus
 */
#ifdef KERNEL26
static void fxdriver_disconnect(struct usb_interface *intf)
#else
static void fxdriver_disconnect(struct usb_device *udev,void* clientdata)
#endif
{

#ifdef KERNEL26
	struct usb_device *udev=interface_to_usbdev(intf);
	struct fxdriver_device *scanner = usb_get_intfdata(intf);
	usb_set_intfdata(intf,NULL);
#else
	struct fxdriver_device *scanner = clientdata;
#endif

	/* Clear reference in the fx2000/3000 device array */
	int scannerIndex=get_scanner_index(udev->descriptor.idProduct);
	fxdriver_presence[scannerIndex]=FALSE;
	dev_fxdriver[scannerIndex]=NULL;

#ifdef CONFIG_SYSFS
#ifdef KERNEL2613
	//class_device_unregister(&fx2000dev);
	class_device_unregister(udev->class_dev);
	class_device_destroy(sysfs_class,MKDEV(fxdriver_major,scannerIndex));
#else
	class_simple_device_remove(MKDEV(fxdriver_major,scannerIndex));
#endif
#endif

	/* Free kernel memory */
	if (scanner!=NULL)
		kfree(scanner);

#ifdef FXDRIVER_DEBUG
	FXDRIVER_DEBUG("USB disconnected\n");
	print_dev_fxdriver();
#endif

  return;
}

#ifdef KERNEL26
static int fxdriver_probe(struct usb_interface *intf,
			  const struct usb_device_id *id)
#else
static void* fxdriver_probe(struct usb_device *udev,unsigned int ifnum,
			  const struct usb_device_id *id)
#endif
{
	/*
	 * The probe procedure is pretty standard. Device matching has already
	 * been performed based on the id_table structure (defined later)
	 */

#ifdef KERNEL26
	struct usb_device *udev=interface_to_usbdev(intf);
#endif
	struct fxdriver_device *scanner;

	/* We don't handle multiple configurations */
	if (udev->descriptor.bNumConfigurations != 1)
#ifdef KERNEL26
		return -ENODEV;
#else
		return NULL;
#endif

#ifdef FXDRIVER_DEBUG
#ifdef KERNEL26
	/* Probed device info */
	FXDRIVER_DEBUG("probe: vendor id 0x%x, device id 0x%x ifnum:%d\n",
				udev->descriptor.idVendor, udev->descriptor.idProduct, intf->minor);
#endif
#endif

	/* allocate and zero a new data structure for the new device */
	scanner = kmalloc(sizeof(struct fxdriver_device), GFP_KERNEL);
	if (!scanner)
#ifdef KERNEL26
	    return -ENOMEM; /* failure */
#else
	    return NULL; /* failure */
#endif
	memset(scanner, 0,     sizeof(*scanner));
	/* Save data to new structure */
	scanner->udev=udev;
	scanner->deviceOpened=FALSE;

	FXDRIVER_DEBUG("PROBE: %d,%d\n",scanner->type,get_scanner_index(udev->descriptor.idProduct));
	dev_fxdriver[get_scanner_index(udev->descriptor.idProduct)]=scanner;
	fxdriver_presence[get_scanner_index(udev->descriptor.idProduct)]=TRUE;
	int minor=get_scanner_index(udev->descriptor.idProduct);

	char temp[20];
	memset(temp,0,20);
	sprintf(temp,"%s_%d",LCASENAME,minor);
#ifdef CONFIG_SYSFS
#ifdef KERNEL2613

	fxdriverdev=class_device_create(sysfs_class,MKDEV(fxdriver_major,minor),&udev->dev,temp)
		    ;
	fxdriverdev->class=sysfs_class;
	udev->class_dev=fxdriverdev;
	fxdriverdev->class_data=udev;
#else
	class_simple_device_add(sysfs_class,MKDEV(fxdriver_major,minor),NULL,temp);
#endif
#endif

#ifdef FXDRIVER_DEBUG
	/* announce yourself */
	FXDRIVER_DEBUG("PROBE successful\n");
	print_dev_fxdriver();
#endif

#ifdef KERNEL26
	usb_set_intfdata(intf,scanner);
	return 0;
#else
	return scanner;
#endif
}

/* -------------------------------------------------------------------------------*/
/* USB file function																															*/
/* -------------------------------------------------------------------------------*/

int f_fxdriver_open(struct inode *inode, struct file *filp)
{
	int minor=MINOR(inode->i_rdev);

#ifdef FXDRIVER_DEBUG
	FXDRIVER_DEBUG ("Open device (%i)",minor);
#endif
	int dev_position=minor;
	/* Check device presence */
	if ((minor<0) || (minor>FXDRIVER_NR_DEVS))
		return -ENODEV; // Wrong Minor (device number)
	if (!fxdriver_presence[minor])
		return -ENODEV; // Device not probed (not present)

	if (dev_fxdriver[dev_position]->deviceOpened==TRUE)
		return -ENODEV; // Device olready open!!!
	/* Config data */
	filp->private_data=&fxdriver_presence[dev_position];
	dev_fxdriver[dev_position]->deviceOpened=TRUE;
#ifdef FXDRIVER_DEBUG
	FXDRIVER_DEBUG (" - OK\n");
#endif
	return 0;          /* success */
}

int f_fxdriver_release(struct inode *inode, struct file *filp)
{
	int minor=MINOR(inode->i_rdev);

#ifdef FXDRIVER_DEBUG
	FXDRIVER_DEBUG ("Close device (%i)\n",minor);
#endif

	int dev_position=minor;
	/* Check device presence */
	if (!fxdriver_presence[minor])
		return 0;	// Something WRONG! The device was disconnect before close

	/* Remove open flag (device now available) */
	dev_fxdriver[dev_position]->deviceOpened=FALSE;
	filp->private_data=NULL;

	return 0;
}

ssize_t f_fxdriver_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	struct usb_device *udev;
	struct fxdriver_device *scanner;
	BOOL minor_presence;
	int minor;

	wait_queue_head_t wait_q;
	ssize_t bytes_read;	/* Overall count of bytes_read */
	ssize_t ret;
	int partial;		/* Number of bytes successfully read */
	int this_read;		/* Max number of bytes to read */
	int result;
	int rd_expire = RD_EXPIRE;
	int i;

	minor_presence=(BOOL)*((BOOL*)filp->private_data);

	if (!minor_presence)
		return -EFAULT;	// Something wrong: device not more available!

	// Search minor number
	for (i=0; (i<FXDRIVER_NR_DEVS) && (&fxdriver_presence[i]!=filp->private_data); i++);
	if (&fxdriver_presence[i]!=filp->private_data)
		return -EFAULT;	// Not Found!!
	minor=i;
	FXDRIVER_DEBUG("Read: %d\n",minor);
	scanner=dev_fxdriver[minor];
	udev=scanner->udev;

	bytes_read = 0;
	ret = 0;

	filp->f_dentry->d_inode->i_atime = CURRENT_TIME;
	init_waitqueue_head(&wait_q);

	while (count > 0) {
		if (signal_pending(current))
		{
			ret = -EINTR;
			break;
		}
		this_read = (count >= BUFFER_IN_SIZE) ? BUFFER_IN_SIZE : count;

		result = usb_bulk_msg(udev, usb_rcvbulkpipe(udev, 5), scanner->bufferIn, this_read, &partial,(int) RD_NAK_TIMEOUT);

#ifdef FXDRIVER_DEBUG
		FXDRIVER_DEBUG("read stats(%d): result:%d this_read:%d partial:%d count:%d\n", minor, result, this_read, partial, count);
#endif

		if (result == -ETIMEDOUT) /* NAK */
			{
				if (!partial) /* No data */
				{
					if (--rd_expire <= 0) /* Give it up */
						{
#ifdef FXDRIVER_DEBUG
							FXDRIVER_DEBUG("read_scanner(%d): excessive NAK's received", minor);
#endif
							ret = result;
							break;
						}
					else /* Keep trying to read data */
						{
							//interruptible_sleep_on_timeout(&wait_q, RD_NAK_TIMEOUT);
							continue;
						}
				}
				else /* Timeout w/ some data */
				{
					goto data_recvd;
				}
			}
		if (result == -EPIPE) /* No hope */
			{
				if(usb_clear_halt(udev, 5))
				{
#ifdef FXDRIVER_DEBUG
					FXDRIVER_DEBUG("read_scanner(%d): Failure to clear endpoint halt condition (%Zd).", minor, ret);
#endif
				}
				ret = result;
				break;
			}
		else
			if ((result < 0) && (result != -ENOSR))
			{
#ifdef FXDRIVER_DEBUG
				FXDRIVER_DEBUG("read_scanner(%d): funky result:%d. Please notify the maintainer.", minor, (int)result);
#endif
				ret = -EIO;
				break;
			}

		data_recvd:
		if (partial) /* Data returned */
			{
				if (copy_to_user(buf, scanner->bufferIn, partial))
					{
						ret = -EFAULT;
						break;
					}
				count -= this_read; /* Compensate for short reads */
				bytes_read += partial; /* Keep tally of what actually was read */
				buf += partial;
			}
		else
			{
				ret = 0;
				break;
			}
		}
	return ret ? ret : bytes_read;
}

ssize_t f_fxdriver_write(struct file *filp, const char *buf, size_t count, loff_t *f_pos)
{
	struct usb_device *udev;
	struct fxdriver_device *scanner;
	BOOL minor_presence;
	int minor,i;

	unsigned long copy_size;
	unsigned long bytes_written = 0;
	unsigned int partial;

	int result = 0;
	int maxretry;
	int errn = 0;
	wait_queue_head_t wait_q;

	minor_presence=(BOOL)*((BOOL*)filp->private_data);

	if (!minor_presence)
		return -EFAULT;	// Something wrong: device not more available!

	// Search minor number (for dev_fx2000 index)
	for (i=0; (i<FXDRIVER_NR_DEVS) && (&fxdriver_presence[i]!=filp->private_data); i++);
	if (&fxdriver_presence[i]!=filp->private_data)
		return -EFAULT;	// Not Found!!
	minor=i;
	FXDRIVER_DEBUG("Write: %d\n",minor);
	scanner=dev_fxdriver[minor];
	udev=scanner->udev;

	init_waitqueue_head(&wait_q);

	do
	{
		unsigned long thistime;
		char *obuf = scanner->bufferOut;

		thistime = copy_size = (count >= BUFFER_OUT_SIZE) ? BUFFER_OUT_SIZE : count;
		if (copy_from_user(obuf, buf, copy_size))
		{
			errn = -EFAULT;
			goto error;
		}

		maxretry = 2;
		while (thistime)
		{
			result = usb_bulk_msg(udev, usb_sndbulkpipe(udev, 6), obuf, thistime, &partial, (int) WR_NAK_TIMEOUT);

#ifdef FXDRIVER_DEBUG
			FXDRIVER_DEBUG("write stat(%i): result:%d thistime:%lu partial:%u\n",minor,result, thistime, partial);
#endif

			if (result == -ETIMEDOUT) /* NAK - so hold for a while */
				{
					if (!maxretry--)
					{
						errn = -ETIME;
						goto error;
					}
					//interruptible_sleep_on_timeout(&wait_q, WR_NAK_TIMEOUT);
					continue;
				}
				else
					if (!result & partial)
					{
						obuf += partial;
						thistime -= partial;
					}
					else
						break;
		};
		if (result)
		{
#ifdef FXDRIVER_DEBUG
			FXDRIVER_DEBUG("Write Whoops - %x", result);
#endif
			errn = -EIO;
			goto error;
		}
		bytes_written += copy_size;
		count -= copy_size;
		buf += copy_size;
	} while (count > 0);
	FXDRIVER_DEBUG("Written: %d\n",bytes_written);
	return bytes_written ? bytes_written : -EIO;

error:
	return errn;
}

/* Device handled by this driver */
static struct usb_device_id fxdriver_id_table [] = {
#ifdef FX3000
	{ USB_DEVICE(0x0c43, 0x0300) },
	{ USB_DEVICE(0x0c43, 0x0301) },
	{ USB_DEVICE(0x0c43, 0x0302) },
	{ USB_DEVICE(0x0c43, 0x0303) },
	{ USB_DEVICE(0x0c43, 0x0304) },
	{ USB_DEVICE(0x0c43, 0x0305) },
	{ USB_DEVICE(0x0c43, 0x0306) },
	{ USB_DEVICE(0x0c43, 0x0307) },
	{ USB_DEVICE(0x0c43, 0x0308) },
	{ USB_DEVICE(0x0c43, 0x0309) },
#else
	{ USB_DEVICE(0x0c43, 0x0200) },
	{ USB_DEVICE(0x0c43, 0x0201) },
	{ USB_DEVICE(0x0c43, 0x0202) },
	{ USB_DEVICE(0x0c43, 0x0203) },
	{ USB_DEVICE(0x0c43, 0x0204) },
	{ USB_DEVICE(0x0c43, 0x0205) },
	{ USB_DEVICE(0x0c43, 0x0206) },
	{ USB_DEVICE(0x0c43, 0x0207) },
	{ USB_DEVICE(0x0c43, 0x0208) },
	{ USB_DEVICE(0x0c43, 0x0209) },
#endif
	{ }						/* Terminating entry */
};

#ifdef KERNEL24
#ifdef CONFIG_DEVFS_FS
devfs_handle_t fxdriver_devfs_dir;
#endif
#endif

//static char devname[4];
MODULE_DEVICE_TABLE (usb,fxdriver_id_table);

/*
 * The callbacks are registered within the USB subsystem using the
 * usb_driver data structure
 */
static struct usb_driver fxdriver_usb_driver = {
#if LINUX_VERSION_CODE>=KERNEL_VERSION(2,4,20)
	.owner=		THIS_MODULE,
#endif
#ifdef FX3000
	.name=						"Fx3000",
#else
	.name=						"Fx2000",
#endif
	.probe=						fxdriver_probe,
	.disconnect=					fxdriver_disconnect,
	.id_table=					fxdriver_id_table,
};



/*
 * Functions called at module load and unload time: only register and
 * unregister the USB callbacks
 */
static int __init fxdriver_init(void)
{
	int result, i;

#ifdef KERNEL24
#ifdef CONFIG_DEVFS_FS
	SET_MODULE_OWNER(&fxdriver_fops);
	result = devfs_register_chrdev(fxdriver_major, LCASENAME, &fxdriver_fops);
#else
	result = register_chrdev(fxdriver_major, LCASENAME, &fxdriver_fops);
#endif
#else
	result = register_chrdev(fxdriver_major, LCASENAME, &fxdriver_fops);
#endif

#ifdef CONFIG_SYSFS
#ifdef KERNEL2613
	sysfs_class = class_create(THIS_MODULE,LCASENAME);
#else
	sysfs_class = class_simple_create(THIS_MODULE,LCASENAME);
#endif
#endif
	if (result < 0) {
			FXDRIVER_DEBUG("can't get major %d\n",fxdriver_major);
			return result;
	}
	if (fxdriver_major == 0) fxdriver_major = result;

#ifdef FXDRIVER_DEBUG
		FXDRIVER_DEBUG("%s_major=%i, max device=%i\n",LCASENAME,fxdriver_major,FXDRIVER_NR_DEVS);
#endif


	for (i=0; i < FXDRIVER_NR_DEVS; i++) {
		dev_fxdriver[i]=NULL;
		fxdriver_presence[i]=FALSE;
	}


#ifndef FXDRIVER_DEBUG
	EXPORT_NO_SYMBOLS;
#endif

	fxdriver_create_proc();


	return usb_register(&fxdriver_usb_driver);

}

static void __exit fxdriver_exit(void)
{
	int i;
#ifdef KERNEL24
#ifdef CONFIG_DEVFS_FS
	devfs_unregister_chrdev(fxdriver_major, LCASENAME);
#else
	unregister_chrdev(fxdriver_major,LCASENAME);
#endif
#else
	unregister_chrdev(fxdriver_major,LCASENAME);
#endif
	usb_deregister(&fxdriver_usb_driver);

#ifdef CONFIG_SYSFS
#ifdef KERNEL2613
	class_destroy(sysfs_class);
#else
	class_simple_destroy(sysfs_class);
#endif
#endif

	/* cleanup_module is never called if registering failed */

	fxdriver_remove_proc();

	/* Not really necessary */
	for (i=0; i < FXDRIVER_NR_DEVS; i++) {
		if (dev_fxdriver[i]!=NULL)
			kfree (dev_fxdriver[i]);
	}
}

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );
MODULE_LICENSE("GPL");

module_init(fxdriver_init);
module_exit(fxdriver_exit);
